# -*- perl -*-

#
# $Id: BBBikeAdvanced.pm,v 1.25 2000/12/17 18:13:14 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1999,2000 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

use Config;
use strict;
use vars qw($top %str_draw %p_draw $os $use_clipboard
	    $edit_mode %font $advanced
	    $tmpdir %tmpfiles $datadir @popup_style
	    $center_on_str);

if (!$advanced) {
    require Carp;
    Carp::cluck("BBBikeAdvanced loaded !?");
}

sub advanced_option_menu {
    my $opbm = shift;
    $opbm->separator;
    $opbm->command(-label => 'Ptksh',
		   -command => sub {
		       my $perldir = $Config{'scriptdir'};
		       require "$perldir/ptksh";
		   });
    $opbm->command(-label => 'WidgetDump',
		   -command => sub {
		       require Tk::WidgetDump;
		       $top->WidgetDump;
		   });
    my $add_pl = "$tmpdir/add.pl";
    $opbm->command(-label => "Eval $add_pl",
		   -command => sub {
		       if (-f "$add_pl") {
			   do $add_pl;
			   warn $@ if $@;
			   return;
		       }
		       if ($top->can('getOpenFile')) {
			   my $f = $top->getOpenFile
			     (-filetypes =>
			      [
			       ['Perl scripts',  ['.pl']],
			       ['Perl modules',  '.pm'  ],
			       ['All Files',     '*',   ],
			      ]);
			   if (defined $f and -f $f) {
			       do $f;
			       warn $@ if $@;
			   }
		       } else {
			   warn "Nothing found";
		       }
		   }
		   );
    $opbm->command(-label => 'Reload modules',
		   -command => \&reload_new_modules);
    my $custom_draw = sub {
	my $linetype = shift;
	my $draw = eval '\%' . $linetype . "_draw";
	my $file = eval '\%' . $linetype . "_file";
	if (!$draw->{'xxx'}) {
	    plot($linetype, "xxx"); # nur lschen
	} else {
	    die "Tk 800 needed"
		unless $Tk::VERSION >= 800;
	    my $f = $top->getOpenFile();
	    if (defined $file) {
		$file->{'xxx'} = $f;
		# zustzliches desc-File einlesen:
		if ($f =~ /(.*)\.bbd$/) {
		    my $desc_file = "$1.desc";
		    if (-r $desc_file && -f $desc_file) {
			require Safe;
			my $compartment = new Safe;
			$compartment->share
			  (qw(%category_color %str_color %line_width
			      %line_length %category_size %outline_color
			      %str_attrib %p_attrib));
			$compartment->rdo($desc_file);
		    }
		}

		plot($linetype, "xxx");
	    }
	}
    };

    $opbm->cascade(-label => 'Zustzlich zeichnen ...');
    {
	my $cusm = $opbm->Menu;
	$opbm->entryconfigure('last', -menu => $cusm);
	$cusm->checkbutton(-label => 'Strecken zeichnen',
			   -variable => \$str_draw{'xxx'},
			   -command => sub { $custom_draw->('str') });
	$cusm->command(-label => 'Strecken auswhlen',
		       -command => sub { choose_ort('s', 'xxx') });
	$cusm->checkbutton(-label => 'Punkte zeichnen',
			   -variable => \$p_draw{'xxx'},
			   -command => sub { $custom_draw->('p') });
	$cusm->command(-label => 'Punkte auswhlen',
		       -command => sub { choose_ort('p', 'xxx') });
    }
    $opbm->command(-label => 'Datenverzeichnis ndern ...',
		   -command => \&change_datadir);

    $top->bind("<Pause>" => sub {
		   eval {
		       require Tk::WidgetDump;
		       $top->WidgetDump;
		   }; warn $@ if $@;
		   require Config;
		   my $perldir = $Config::Config{'scriptdir'};
		   require "$perldir/ptksh";
	       });

}

sub change_datadir {
    require Tk::DirTree;
    my $t = $top->Toplevel;
    $t->title("Neues Datenverzeichnis whlen");
    my $newdir = $datadir;
    my $ok = 0;
    my $f = $t->Frame->pack(-fill => "x", -side => "bottom");
    my $d = $t->Scrolled('DirTree',
			 -scrollbars => 'osoe',
			 -width => 35,
			 -height => 20,
			 -selectmode => 'browse',
			 -exportselection => 1,
			 -browsecmd => sub { $newdir = shift },
			 -command   => sub { $ok = 1 },
			)->pack(-fill => "both", -expand => 1);
    $d->chdir($newdir);
    $f->Button(Name => 'ok',
	       -command => sub { $ok = 1 })->pack(-side => 'left');
    $f->Button(Name => 'cancel',
	       -command => sub { $ok = -1 })->pack(-side => 'left');
    $f->waitVariable(\$ok);
    if ($ok == 1) {
	set_datadir($newdir);
    }
    $t->destroy;
}

use vars qw($standard_command_index $editberlin_command_index
	    $editbrb_command_index $coord_output @edit_mode_cmd
	    $coord_system @edit_mode_brb_cmd @edit_mode_b_cmd
	    @standard_mode_cmd $pp_color $c $coord_prefix $special_edit
	    $netz $point_editor %crossings);

sub set_coord_interactive {
    if (!defined $coord_output ||
	!$Karte::map{$coord_output}) {
	die "Karte-Objekt nicht definiert... Ausgabe richtig setzen!\n";
	return;
    }

    my $t = redisplay_top($top, 'set_coord_interactive',
			  -title => "Koordinaten setzen");
    return if !defined $t;

    $t->Label(-text => "Koordinaten:\n(Koordinatensystem " . 
	      $Karte::map{$coord_output}->name . ")\n(im Menpunkt \"Ausgabe\" setzen)")->pack;
    my($valx, $valy);
    my(%val2);
    my $set_sub = sub {
	my($orig) = @_;
	if ($orig == 2) {
	    $valx = $val2{'X'}->[0] + $val2{'X'}->[1]/60 + $val2{'X'}->[2]/3600;
	    $valy = $val2{'Y'}->[0] + $val2{'Y'}->[1]/60 + $val2{'Y'}->[2]/3600;
	}
	my($setx, $sety) = transpose($Karte::map{$coord_output}->map2standard($valx, $valy));
	mark_point('-x' => $setx, '-y' => $sety);
    };

    my $f1 = $t->Frame->pack;
    my $lx = $f1->Label(-text => "X:");
    my $ex = $f1->Entry(-textvariable => \$valx);
    my $ly = $f1->Label(-text => "Y:");
    my $ey = $f1->Entry(-textvariable => \$valy);
    my $get_selection_sub = sub {
	my $interactive = shift;
	my $s;
        Tk::catch {
	    $s = $f1->SelectionGet('-selection' => ($os eq 'win'
						    ? "CLIPBOARD"
						    : "PRIMARY"));
	};
	if (defined $s and $s =~ /\d/) {
	    $s =~ s/^\D+//;
	    $s =~ s/\D+$//;
	    my($x,$y) = split(/\D+/, $s);
	    if (defined $x and defined $y) {
		($valx, $valy) = ($x, $y);
		$set_sub->(1);
	    } else {
		my $msg = "Can't parse selection $s";
		if ($interactive) {
		    $f1->messageBox(-icon => "error",
				    -message => $msg);
		} else {
		    warn $msg;
		}
	    }
	} else {
	    my $msg = "No useable selection";
	    if ($interactive) {
		$f1->messageBox(-icon => "error",
				-message => $msg);
	    } else {
		warn $msg;
	    }
	}
    };
    my $selb = $f1->Button
	(-text => "Selection",
	 -command => sub { $get_selection_sub->(1) });
    my $sb = $f1->Button(-text => "Setzen",
			 -command => sub { $set_sub->(1) },
			);
    my $autocheck = 0;
    my $acb;
    my $auto_sub = sub {
	$get_selection_sub->(0);
	$set_sub->(1);
	$f1->after(100, sub {
		       $acb->invoke;
		       $acb->invoke;
		   });
    };
    $acb = $f1->Checkbutton
	(-text => "Auto-detect",
	 -variable => \$autocheck,
	 -command => sub {
	     if ($autocheck) {
		 $f1->SelectionOwn(-command => $auto_sub);
		 # Hack to reinstall SelectionOwn handler
	     } else {
		 $f1->SelectionOwn;
	     }
	 });

    $lx->grid($ex, $selb, $acb);
    $ly->grid($ey, $sb);
    $ex->focus;

    if ($coord_output eq 'polar') {
	my $f = $t->Frame->pack;
	for my $ord ('X', 'Y') {
	    my @e2;
	    push @e2, $f->Label(-text => $ord . ":");
	    for my $i (0 .. 2) {
		push @e2, $f->Entry(-textvariable => \$val2{$ord}->[$i],
				    -width => 2);
		if ($i == 0) {
		    push @e2, $f->Label(-text => "");
		} elsif ($i == 1) {
		    push @e2, $f->Label(-text => "'");
		} elsif ($i == 2) {
		    push @e2, $f->Label(-text => "\"");
		    if ($ord eq 'Y') {
			push @e2, $f->Button(-text => "Setzen",
					     -command => sub { $set_sub->(2) },
					    );
		    }
		}
	    }
	    my $first = shift @e2;
	    $first->grid(@e2);
	}
    }
    # XXX t->popup mit popup_style verwenden
}

sub set_line_coord_interactive {
    if (!defined $coord_output ||
	!$Karte::map{$coord_output}) {
	die "Karte-Objekt nicht definiert... Ausgabe richtig setzen!\n";
	return;
    }

    my $t = redisplay_top($top, 'set_line_coord_interactive',
			  -title => "Linienkoordinaten setzen");
    return if !defined $t;

    my @coords;

    my $set_sub = sub {
	my @line_coords;
	foreach (@coords) {
	    my($valx,$valy) = @$_;
	    my($setx, $sety) = transpose($Karte::map{$coord_output}->map2standard($valx, $valy));
	    push @line_coords, [$setx, $sety];
	}
	mark_street(-coords => \@line_coords,
		    -type => 's',
		    );
    };

    $t->Button(-text => "Selection setzen",
	       -command => sub {
		   @coords = ();
		   my $s = $t->SelectionGet('-selection' => ($os eq 'win'
							     ? "CLIPBOARD"
							     : "PRIMARY"));
		   foreach (split(/\s+/, $s)) {
		       push @coords, [split /,/, $_];
		   }

		   $set_sub->();
	       })->pack;
}

sub advanced_coord_menu {
    my $bpcm = shift;
    $bpcm->command(-label => 'Koordinatenliste zeigen',
		   -command => \&show_coord_list);
    $bpcm->separator;
    $bpcm->cascade(-label => 'Ausgabe');
    {
	my $ausm = $bpcm->Menu;
	$bpcm->entryconfigure('last', -menu => $ausm);
	foreach (@Karte::map, qw(canvas)) {
	    my $name = (ref $Karte::map{$_} && $Karte::map{$_}->can('name')
			? $Karte::map{$_}->name
			: $_);
	    $ausm->radiobutton(-label => $name,
			       -variable => \$coord_output,
			       -value => $_,
			       -command => \&set_coord_output_sub,
			       );
	    if ($_ eq 'canvas') {
		my $index = $ausm->index('last');
		push @edit_mode_cmd, sub { $ausm->invoke($index) };
	    }
	}
    }

    $bpcm->cascade(-label => 'Koordinatensystem');
    {
	my $csm = $bpcm->Menu;
	$bpcm->entryconfigure('last', -menu => $csm);
	foreach (@Karte::map, qw(canvas)) {
	    my $o = $Karte::map{$_};
	    my $name = (ref $o && $o->can('name')
			? $o->name
			: $_);
	    $csm->radiobutton(-label => $name,
			      -value => $_,
			      -variable => \$coord_system,
			      -command => sub { set_coord_system($o) },
			      );
	    if ($_ eq 'brbmap') {
		my $index = $csm->index('last');
		push @edit_mode_brb_cmd, sub { $csm->invoke($index) };
	    } elsif ($_ eq 'berlinmap') {
		my $index = $csm->index('last');
		push @edit_mode_b_cmd, sub { $csm->invoke($index) };
	    } elsif ($_ eq 'standard') {
		my $index = $csm->index('last');
		push @standard_mode_cmd, sub { $csm->invoke($index) };
	    }
	}
    }
    $bpcm->command(-label => "Koordinaten setzen",
		   -command => \&set_coord_interactive);
    $bpcm->command(-label => "Linienkoordinaten setzen",
		   -command => \&set_line_coord_interactive);
    $bpcm->separator;
    {
	$bpcm->checkbutton(-label => 'Kreuzungen/Kurvenpunkte zeichnen',
			   -variable => \$p_draw{'pp'});
	my $index = $bpcm->index('last');
	push(@edit_mode_cmd,
	     sub {
		 $bpcm->invoke($index) unless $p_draw{'pp'};
	     });
	push(@standard_mode_cmd,
             sub { $bpcm->invoke($index) if $p_draw{'pp'} });
      }
    my $coldef;
    foreach $coldef (['rot', '#800000'],
		     ['grn', '#008000'],
		    ) {
      $bpcm->radiobutton(-label    => 'Kurvenpunkte ' . $coldef->[0],
			 -variable => \$pp_color,
			 -value    => $coldef->[1],
			 -command  => sub {
			   $c->itemconfigure('pp',
					     -fill => $pp_color);
			 },
			);
    }
    $bpcm->checkbutton(-label => "Prfix-Ausgabe",
		       -variable => \$coord_prefix,
		      );
    $bpcm->checkbutton(-label => 'Pltze zeichnen',
		       -variable => \$p_draw{'pl'},
		       -command => sub { plot('p','pl') },
		      );
    $bpcm->separator;
    $bpcm->command(-label => 'Neu laden',
		   -command => \&reload_all,
		   -accelerator => 'Alt-r',
		  );
    $bpcm->command(-label => 'Punkt einfgen',
		   -command => \&insert_points,
		   -accelerator => 'F2',
		  );
    $bpcm->command(-label => 'Punkt ndern',
		   -command => \&change_points,
		   -accelerator => 'F3',
		  );
    $bpcm->separator;

    $bpcm->command
      (-label => "Standard-Modus",
       -command => sub { switch_standard_mode() },
       -accelerator => 'F4',
      );
    $standard_command_index = $bpcm->index('last');
    $bpcm->command
      (-label => "Edit-Modus Berlin",
       -command => sub { switch_edit_berlin_mode() },
       -accelerator => 'F5',
      );
    $editberlin_command_index = $bpcm->index('last');
    $bpcm->command
      (-label => "Edit-Modus Brandenburg",
       -command => sub { switch_edit_brb_mode() },
       -accelerator => 'F6',
      );
    $editbrb_command_index = $bpcm->index('last');
    $bpcm->separator;

    $top->bind("<F4>" => sub { $bpcm->invoke($standard_command_index) });
    $top->bind("<F5>" => sub { $bpcm->invoke($editberlin_command_index) });
    $top->bind("<F6>" => sub { $bpcm->invoke($editbrb_command_index) });

    foreach my $def ({Label => 'Radwege',
		      Type  => 'radweg'},
		     {Label => 'Ampelschaltung',
		      Type  => 'ampel'},
		     {Label => 'Label',
		      Type  => 'label'},
		     {Label => 'Vorfahrt',
		      Type  => 'vorfahrt'}) {
      $bpcm->cascade(-label => $def->{Label});
      my $m = $bpcm->Menu;
      $bpcm->entryconfigure('last', -menu => $m);
      $m->checkbutton(-label => $def->{Label} . '-Modus',
		      -variable => \$special_edit,
		      -onvalue => $def->{Type},
		      -offvalue => '',
		      -command => sub {
			require BBBikeEdit;
			eval $def->{Type} . "_edit_toggle()";
			warn $@ if $@;
		      });
      $m->command(-label => 'Undef all',
		  -command => sub {
		    require BBBikeEdit;
		    eval $def->{Type} . "_undef_all()";
		    warn $@ if $@;
		  });
      $m->command(-label => 'Speichern als...',
		  -command => sub {
		    require BBBikeEdit;
		    eval $def->{Type} . "_save_as()";
		    warn $@ if $@;
		  });
    }
    $bpcm->checkbutton
      (-label => "Point-Editor",
       -variable => \$special_edit,
       -onvalue => "point",
       -offvalue => "",
       -command => sub {
	 if ($special_edit eq 'point') {
	   require PointEdit;
	   my $p = new MasterPunkte "$FindBin::RealBin/misc/masterpoints-orig";
	   $p->read;
	   if (!$netz) { make_netz() }
	   all_crossings();
	   $point_editor = new PointEdit
	     MasterPunkte => $p,
	     Net => $netz,
	     Crossings => \%crossings,
	     Top => $top;
	 } elsif ($point_editor) {
	   $point_editor->delete;
	   undef $point_editor;
	 }
       });
}

sub advanced_bindings {
    $top->bind("<F2>" => \&insert_points);
    $top->bind("<F3>" => \&change_points);
}

use vars qw($verbose %module_time %module_check $main_check_time);

$main_check_time = -M $0;

### AutoLoad Sub
sub check_new_modules {
    my $pkg = shift;
    $pkg = 'main' if (!defined $pkg);
    #warn "checking new modules for $pkg..." if $verbose; # nervig
    my %inc;
    eval '%inc = %' . $pkg . '::INC';
    warn $@ if $@;
    while(my($k, $v) = each %inc) {
	next if $v !~ /bbbike/i; # only record BBBike-related modules
	next if exists $module_time{$v};
	$module_time{$v} = (stat($v))[9];
	warn "recorded $module_time{$v} for $k\n" if $verbose;
    }
    $module_check{$pkg}++ if defined $pkg;
    my %stash;
    eval '%stash = %' . $pkg . '::';
    warn $@ if $@;
    foreach my $sym (keys %stash) {
	if ($sym =~ /^(.*)::$/) {
	    my $subpkg = ($pkg eq 'main' 
			  ? $1
			  : $pkg . "::" . $1);
	    if (!exists $module_check{$subpkg}) {
		check_new_modules($subpkg);
	    }
	}
    }
}

### AutoLoad Sub
sub reload_new_modules {
    while(my($k, $v) = each %module_time) {
	my $now = (stat($k))[9];
	next if $v >= $now;
	print STDERR "Reloading $k...\n";
	eval { do $k };
	warn "*** $@" if $@;
	$module_time{$k} = $now;
    }
    if ($tmpdir && -M $0 < $main_check_time) {
	if (open(MAIN, $0)) {
	    my $tmpfile = "$tmpdir/bbbike_reload_$$.pl";
	    $tmpfiles{$tmpfile}++;
	    if (open(SAVEMAIN, ">$tmpfile")) {
		my $found = 0;
		while(<MAIN>) {
		    if ($found) {
			print SAVEMAIN $_;
		    } elsif (/RELOADER_START/) {
			$found++;
			print SAVEMAIN "# line $. $0\n";
		    }
		}
		close SAVEMAIN;
		print STDERR "Reloading main...\n";
		eval { do $tmpfile };
		warn "*** $@" if $@;
		## Don't delete to track errors easier...
		#unlink $tmpfile;
	    } else {
		warn "Can't write to $tmpfile: $!";
	    }
	    close MAIN;
	} else {
	    warn "Can't open $0: $!";
	}
	$main_check_time = -M $0;
    }
}

############################################################
# Selection-Kram (Koordinatenliste, buttonpoint et al.)
#

use vars qw(@inslauf_selection @ext_selection
	    $coord_prefix $coord_system_obj $coord_output_sub
	    $clicksound $noclick $koord @names
	    $coordlist_lbox $coordlist_lbox_nl $buttonpoint_flag);

# Gibt den angewhlten Punkt aus.
# Ausgegeben wird: Name (soweit vorhanden), Canvas-Koordinaten und
# die Koordinaten abhngig von $coord_output_sub (gewhnlich berlinmap).
# Auerdem werden die $coord_output_sub-Koordinaten in die Selection
# geschrieben.
### AutoLoad Sub
sub buttonpoint {
    my($x, $y) = @_;
    $c->SelectionOwn(-command => sub {
			 @inslauf_selection = ();
			 # kein reset_ext_selection, weil dann beim Anklicken
			 # auf $coordlist_lbox die Selection verschwindet
			 @ext_selection = ();
		     });
    my $prefix = ($coord_prefix ? $coord_system_obj->coordsys : '');
    if (defined $x) {
	my $coord = sprintf "$prefix%d,%d", $coord_output_sub->($x, $y);
	push(@inslauf_selection, $coord);
	$c->clipboardAppend(" $coord") if $use_clipboard;
	my $ext = prepare_selection_line
	    (-name => "?",
	     -coord1 => Route::_coord_as_string([$x,$y]),
	     -coord2 => $coord);
	push_ext_selection($ext);
	print STDERR $ext, "\n";
	&$clicksound unless $noclick;
    } else {
	my(@tags) = $c->gettags('current');
	return if !@tags || !defined $tags[0];
	if ($tags[0] eq 'p'    ||
	    $tags[0] eq 'o'    ||
	    $tags[0] eq 'pp'   ||
	    $tags[0] =~ /^lsa/ ||
	    $tags[0] =~ /^xxx/ ||
	    $tags[0] eq 'fz'   ||
	    $tags[0] =~ /^kn/
	   ) {
	    my($tag, $s);
	    $tag = $tags[1];
	    if ($tags[0] eq 'p') {
		my($cx, $cy) = $koord->get($tag);
		my($x, $y) = $coord_output_sub->($cx, $cy);
		$s = prepare_selection_line
		    (-name => substr(Strassen::strip_bezirk($names[$tag]),
				     0, 40),
		     -coord1 => Route::_coord_as_string([$cx,$cy]),
		     -coord2 => Route::_coord_as_string([$x,$y]),
		     -tag => $tag);
		push(@inslauf_selection, $tag);
		$c->clipboardAppend(" $tag") if $use_clipboard;
		push_ext_selection($s);
	    } elsif ($tags[0] eq 'pp' || $tags[0] =~ /^lsa/ ||
		     $tags[0] =~ /^xxx/) {
		my($x, $y) = $coord_output_sub->
		  (@{Strassen::to_koord1($tags[1])});
		# XXX verallgemeinern!!!
		if ($edit_mode) {
		    all_crossings();
		}
		my $crossing =
		  (exists $crossings{$tags[1]}
		   ? join("/", map { Strassen::strip_bezirk($_) }
			  @{$crossings{$tags[1]}})
		   : "?");
		$s = prepare_selection_line
		    (-name => $crossing,
		     -coord1 => $tags[1],
		     -coord2 => Route::_coord_as_string([$x,$y]));
		my $str = $prefix . Route::_coord_as_string([$x,$y]);
		push(@inslauf_selection, $str);
		$c->clipboardAppend(" $str") if $use_clipboard;
		push_ext_selection($s);
	    } elsif ($tags[0] eq 'o' ||
		     $tags[0] eq 'fz') {
		my($cx, $cy) = anti_transpose($c->coords('current'));
		my($x, $y) = $coord_output_sub->($cx, $cy);
		my $name = ($tags[0] eq 'o'
			    ? substr(Strassen::strip_bezirk($tag), 0, 40)
			    : $tags[1]);
		$s = prepare_selection_line
		  (-name => $name,
		   -coord1 => Route::_coord_as_string([$cx,$cy]),
		   -coord2 => Route::_coord_as_string([$x,$y]));
		my $str = $prefix . Route::_coord_as_string([$x,$y]);
		push(@inslauf_selection, $str);
		$c->clipboardAppend(" $str") if $use_clipboard;
		push_ext_selection($s);
	    } else {
		die "Tag $tags[0] wird fr das Aufzeichnen von Punkten nicht untersttzt";
	    }
	    $s .= "\n";
	    print STDERR $s;
	    &$clicksound unless $noclick;
	}
    }
}

### AutoLoad Sub
sub prepare_selection_line {
    my(%args) = @_;
    if ($os eq 'win') { # XXX
	if (0) { # XXX
	    $args{-coord1} . " ";
	} else {
	    sprintf("%-13s %-33s\n",
		    $args{-coord1},
		    substr($args{-name}, 0, 33));
	}
    } else { # XXX old
	sprintf("%-40s %-15s %-15s",
		$args{-name}, $args{-coord1}, $args{-coord2})
	    . (exists $args{-tag} ? " $args{-tag}" : "");
    }
}

### AutoLoad Sub
sub push_ext_selection {
    my(@a) = @_;
    push @ext_selection, @a;
    if (defined $coordlist_lbox && Tk::Exists($coordlist_lbox)) {
	if (subw_isa($coordlist_lbox, 'Tk::Text')) {
	    $coordlist_lbox->insert('end', join($coordlist_lbox_nl,
						@a) . $coordlist_lbox_nl);
	} else {
	    $coordlist_lbox->insert('end', @a);
	}
	$coordlist_lbox->see('end');
    }
}

### AutoLoad Sub
sub reset_ext_selection {
    @ext_selection = ();
    if (defined $coordlist_lbox && Tk::Exists($coordlist_lbox)) {
	if (subw_isa($coordlist_lbox, 'Tk::Text')) {
	    $coordlist_lbox->delete("1.0", 'end');
	} else {
	    $coordlist_lbox->delete(0, 'end');
	}
    }
}

### AutoLoad Sub
sub reset_selection {
    @inslauf_selection = ();
    $c->clipboardClear() if $use_clipboard;
    reset_ext_selection();
}

### AutoLoad Sub
sub show_coord_list {
    my $coordlist_top = redisplay_top($top, 'coordlist',
				      -title => 'Koordinatenliste');
    return if !defined $coordlist_top;
    if (1 || $os eq 'win') { # XXX (1) # unter Win32 funktionieren Selections anders
	require Tk::ROText;
	$coordlist_lbox = $coordlist_top->Scrolled
	    ('ROText', -font => $font{'fixed'},
	     -width => 80,
	     -scrollbars => 'osoe')->pack;
	$coordlist_lbox_nl = "";
    } else {
	$coordlist_lbox = $coordlist_top->Scrolled
	    ('Listbox', -font => $font{'fixed'},
	     -width => 80,
	     -selectmode => 'extended',
	     -scrollbars => 'osoe')->pack;
    }
    if (@ext_selection) {
	require BBBikeAdvanced;
	$coordlist_lbox->insert('end',
				(subw_isa($coordlist_lbox, 'Tk::Text')
				 ? join($coordlist_lbox_nl, @ext_selection)
				 : @ext_selection));
    }
    $coordlist_top->Button
      (Name => 'end',
       -command => sub { $coordlist_top->destroy },
      )->pack;
    $coordlist_top->Popup(@popup_style);
}

######################################################################
#
# Zustzliche Zeichenfunktionen
#

use vars qw($progress);

# Zeichnet Haltestellen-Informationen aus der Hafas-Datenbank.
# Funktioniert nur fr ltere Daten.
# Fraglich, ob diese Funktion noch bentigt wird...
### AutoLoad Sub
sub ploths {
    status_message("");

    $c->delete("p");		# evtl. alte Koordinaten lschen
    if (!$p_draw{'p'}) {
	return;
    }

    my $anzahl_eindeutig;
    eval {
	require Fahrinfo;
	my $eh = tie @names, 'Fahrinfo::Eind_haltestellen';
	if (!$koord) {
	    $koord = new Fahrinfo::Koord $eh;
	}
	$anzahl_eindeutig = $eh->{'haltestellen'}{'anzahl_eindeutig'};
    };
    if ($@) {
	status_message($@, 'err');
	return;
    }

    destroy_delayed_restack();

    IncBusy($top);
    $progress->Init(-dependents => $c,
		    -label => 'Haltestellen');
    eval {
	# mit nextdirect geht es am schnellsten
	$koord->initnextdirect;
	for my $i (0 .. $anzahl_eindeutig-1) {
	    my ($tx, $ty) = transpose($koord->nextdirect);
	    $progress->Update($i/$anzahl_eindeutig)
	      if $i % 80 == 0;
	    $c->createLine
	      ($tx, $ty, $tx, $ty,
	       -tags => ['p', $i]);
	}

	$c->itemconfigure('p', 
			  -capstyle => 'round', 
			  -width => 5,
			  -fill => 'blue',
			 );
	restack_delayed();
    };
    $progress->Finish;
    DecBusy($top);
}

######################################################################
#
# Edit/Standard-Modus
#

use vars qw($ampelstatus_label @remember_plot_str @remember_plot_p
	    %str_obj %p_obj $map_draw $search_flag
	    $wasserstadt $wasserumland %str_far_away %do_flag
	    $wichtige_orte $info_flag
	    );

# Lscht die aktiven Straen und Punkte und merkt sie sich in
# fr das sptere Wiederzeichnen in set_remember_plot.
### AutoLoad Sub
sub remove_plot {
    undef @remember_plot_str;
    my $abk;
    foreach $abk (keys %str_draw) {
	if ($str_draw{$abk}) {
	    $str_draw{$abk} = 0;
	    plot('str',$abk);
	    push @remember_plot_str, $abk;
	}
	if (defined $str_obj{$abk}) {
	    undef $str_obj{$abk};
	}
    }
    undef @remember_plot_p;
    foreach $abk (keys %p_draw) {
	next if $abk eq 'pp';
	if ($p_draw{$abk}) {
	    $p_draw{$abk} = 0;
	    plot('p',$abk);
	    push @remember_plot_p, $abk;
	}
    }
    delete_berlinmap();
    $map_draw = 0; # XXX
}

# Zeichnet die Strecken und Punkte neu, die in remove_plot() gelscht wurden.
### AutoLoad Sub
sub set_remember_plot {
    my $abk;
    $progress->InitGroup;
    foreach $abk (@remember_plot_str) {
	if (!$str_draw{$abk}) {
	    $str_draw{$abk} = 1;
	    plot('str',$abk);
	}
    }
    foreach $abk (@remember_plot_p) {
	if (!$p_draw{$abk}) {
	    $p_draw{$abk} = 1;
	    plot('p',$abk);
	}
    }
    $progress->FinishGroup;
}

# Schaltet in einen der folgenden Modi um.
### AutoLoad Sub
sub switch_mode {
    my $mode = shift;
    if ($mode eq 'std') {
	switch_standard_mode(@_);
    } elsif ($mode eq 'b') {
	switch_edit_berlin_mode(@_);
    } elsif ($mode eq 'brb') {
	switch_edit_brb_mode(@_);
    } else {
	die "Unknown mode for switch_mode: $mode";
    }
}

# Schaltet in den Standard-Modus um.
### AutoLoad Sub
sub switch_standard_mode {
    my $init = shift;
    my($oldx, $oldy) = 
      $coord_system_obj->map2standard
	(anti_transpose(get_center()));
    remove_plot() unless $init;
    foreach (@standard_mode_cmd) { $_->() }
    $search_flag = 1;
    $buttonpoint_flag = 0;
    $info_flag = 0;
    set_edit_mode(0);
    set_remember_plot() unless $init;
    $ampelstatus_label->configure(-text => '') if $ampelstatus_label;
    center_view
      ($c,
       transpose($coord_system_obj->standard2map($oldx, $oldy)),
       NoSmoothScroll => 1);
}

# Schaltet in den Edit-Mode fr Berlin um.
### AutoLoad Sub
sub switch_edit_berlin_mode {
    my $init = shift;
    my($oldx, $oldy) = 
      $coord_system_obj->map2standard
	(anti_transpose(get_center()));
    remove_plot() unless $init;
    foreach (@edit_mode_cmd) { $_->() }
    foreach (@edit_mode_b_cmd) { $_->() }
    $search_flag = 0;
    $info_flag = 0;
    $coord_prefix = 0;
    $buttonpoint_flag = 1;
    $wasserstadt = 1;
    $wasserumland = 0;
    $str_far_away{'w'} = 0;
    set_edit_mode('b');
    $do_flag{'start'} = $do_flag{'ziel'} = 0;
    set_remember_plot() unless $init;
    center_view
      ($c,
       transpose($coord_system_obj->standard2map($oldx, $oldy)),
       NoSmoothScroll => 1);
}

# Schaltet in den Edit-Mode fr das Umland (Brandenburg) um.
### AutoLoad Sub
sub switch_edit_brb_mode {
    my $init = shift;
    my($oldx, $oldy) = 
      $coord_system_obj->map2standard
	(anti_transpose(get_center()));
    remove_plot() unless $init;
    foreach (@edit_mode_cmd) { $_->() }
    foreach (@edit_mode_brb_cmd) { $_->() }
    $search_flag = 0;
    $info_flag = 0;
    $coord_prefix = 1;
    $buttonpoint_flag = 1;
    $wasserstadt = 0;
    $wasserumland = 1;
    $wichtige_orte = 0;
    set_edit_mode('brb');
    $do_flag{'start'} = $do_flag{'ziel'} = 0;
    set_remember_plot() unless $init;
    center_view
      ($c,
       transpose($coord_system_obj->standard2map($oldx, $oldy)),
       NoSmoothScroll => 1);
}

# Erzeugt einen Hash aller Kreuzungen
### AutoLoad Sub
sub all_crossings {
    if (!keys %crossings) {
	my $s = new Strassen "strassen-orig";
	%crossings = $s->all_crossings(RetType => 'hash',
				       UseCache => 1);
    }
}

use vars qw(@search_route_points);

### AutoLoad Sub
sub search_movie {
    require "miscsrc/kino-berlin.pl";
    Kino::Berlin::init("$FindBin::RealBin/miscsrc");

    my $start;
    if (!@search_route_points || !$search_route_points[0]->[0]) {
	if (defined $center_on_str) {
	    my $plz = new PLZ;
	    $plz->open;
	    my @res = $plz->look($center_on_str);
	    if (@res) {
		$start = $res[0]->[3];
	    }
	}
	die "Startpunkt nicht gesetzt" if !$start;
    } else {
	$start = $search_route_points[0]->[0];
    }

    my $entry_t = $top->Toplevel(-title => "Movie");
    my $movie;
    my $entry = $entry_t->Entry(-textvariable => \$movie)->pack;
    $entry_t->idletasks;
    $entry->focus;
    $entry_t->bind("<<CloseWin>>" => sub { $entry_t->destroy });
    $entry->bind
	("<Return>" => sub {
	     $entry_t->destroy;
	     my @res = Kino::Berlin::search_nearest_cinema($movie, $start);
	     return unless @res;
	     my $coords;
	     Kino::Berlin::tk_result
		     ($top, \@res,
		      -variable  => \$coords,
		      -selsignal => sub {
			  print $coords, "\n";
			  $start = $search_route_points[0]->[0] || $start;
			  set_route_start($start);
			  set_route_ziel($coords);
			  zoom_view();
		      },
		      -getstartcoords => sub {
			  $search_route_points[0]->[0] || $start;
		      },
		      -transient => $top,
		     );
	 });
    $entry_t->Popup(@popup_style);
}

1;

__END__
