# -*- perl -*-

#
# $Id: BBBikePalm.pm,v 1.3 2000/12/06 02:12:34 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 2000 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

# Palm-Routinen fr BBBike

# in main muss definiert sein: is_in_path

use strict;
use vars qw($tmpdir %tmpfiles @html_show_route_save $palm_doc_format);

my @prog_order = qw(txt2pdbdoc /usr/local/txt2pdbdoc/bin/txt2pdbdoc
		    pilot_makedoc
		    iSiloBSD iSiloLinux
		    );

sub can_create_and_transfer_palm_docs {
    ((is_in_path("pilot_makedoc") ||
      is_in_path("iSiloBSD")      ||
      is_in_path("iSiloLinux")    ||
      is_in_path("txt2pdbdoc")    ||
      -x "/usr/local/txt2pdbdoc/bin/txt2pdbdoc")
     &&
     is_in_path("pilot-xfer"));
}

sub create_palm_button {
    my $parent = shift;
    my $get_endpoints_sub = shift;
    $parent->Button
	(-text => 'Palm',
	 -command => sub {
	     my $f   = "$tmpdir/palmdoc-$$.";
	     my $doc = "$tmpdir/palmdoc-$$.pdb";
	     my $is_empty = sub {
		 if (!-r $f || -z $f) {
		     die "Datei $f existiert nicht bzw. ist leer";
		 }
	     };
	     if ($palm_doc_format eq 'isilo') {
	         BBBikePalm::to_top(\@prog_order, qw(iSiloBSD iSiloLinux));
	     }
	     my $route_name = "BBBike-Route";
	     if (defined $main::show_route_start and
		 defined $main::show_route_ziel) {
		 my($start, $ziel);
		 $start = Strasse::short(Strassen::strip_bezirk($main::show_route_start), 3); # Start besser abkrzen --- ist meist immer der Gleiche
		 $ziel  = Strasse::short(Strassen::strip_bezirk($main::show_route_ziel), 2);
		 $route_name = "BBBike: $start-$ziel";
	     }
	     foreach my $check_sys (@prog_order) {
		 my $full_path = is_in_path($check_sys);
		 if (defined $full_path) {
		     if ($check_sys =~ /isilo/i) {
			 $f .= "html";
			 open(PALM, ">$f") or
			     die "$f kann nicht geschrieben werden: $!";
			 print PALM join("\n", @html_show_route_save), "\n";
			 close PALM;
			 $is_empty->();
			 system($full_path, "-y", "-i$route_name", $f, $doc);
		     } else {
			 $f .= "txt";
			 open(PALM, ">$f") or
			     die "$f kann nicht geschrieben werden: $!";
			 print PALM BBBikePalm::strip_html_tags
			     (join("\n", @html_show_route_save)), "\n";
			 close PALM;
			 $is_empty->();
			 if ($check_sys =~ /pilot_makedoc/) {
			     system($full_path, $f, $doc, $route_name);
			 } elsif ($check_sys =~ /txt2pdbdoc/) {
			     system($full_path, $route_name, $f, $doc);
			 }
		     }
		     last;
		 }
	     }
	     if (!-r $doc) {
		 die "Doc-Datei $doc konnte nicht erzeugt werden";
	     }
	     system("pilot-xfer -i $doc &");
	     $tmpfiles{$doc}++;
	     BBBikePalm::hot_sync_message($top);
	 }
	);
}

sub BBBikePalm::strip_html_tags {
    my $text = shift;
    $text =~ s/<[^>]+>//gs;
    $text;
}

sub BBBikePalm::to_top {
    my($listref, @to_top) = @_;
    my(%to_top) = map { ($_ => 1) } @to_top;
    my @new_list = @to_top;
    foreach (@$listref) {
	push @new_list, $_ if !exists $to_top{$_};
    }
    @$listref = ();
    foreach (@new_list) {
	push @$listref, $_;
    }
}

sub BBBikePalm::hot_sync_message {
    my $top = shift;
    $top->messageBox(-title => "Hotsync",
		     -text => "Hotsync-Button drcken!");
}

1;

__END__
