# -*- perl -*-

#
# $Id: NSD.pm,v 1.1 2000/07/26 20:27:14 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 2000 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

# http://www.nws.noaa.gov/pub/stninfo/nsd_cccc.txt
# Also distributed with tkGeoMap.
# A list of weather stations.

package GIS::NSD;

use strict;
use vars qw($VERSION);
$VERSION = "0.01";

######################################################################

package GIS::NSD::Record;

if (eval 'use fields qw(abbrev x1 x2 city state country x3
                        lat1 long1 lat2 long2 height1 height2 x4
                       ); $] >= 5.006') {
    eval <<'EOF';
sub new {
    my GIS::NSD::Record $self = shift;
    unless (ref $self) {
        $self = fields::new($self);
    }
    $self;
}
EOF
    warn $@ if $@;
} else {
    eval <<'EOF';
sub new { bless {}, $_[0] }
EOF
    warn $@ if $@;
}

=head2 _position_to_decimal()

Recalculate positions to decimal degrees.

=cut

sub _position_to_decimal {
    my $self = shift;
    foreach my $pos (qw/lat1 lat2 long1 long2/) {
	next if !defined $self->{$pos} || $self->{$pos} eq '';
	if ($self->{$pos} =~ /^(\d+)-(\d+)\s*(-(\d*)\s*)?(.)$/) {
	    $self->{$pos} = $1 + $2/60;
	    if (defined $4) {
		$self->{$pos} += $3/3600; # oder /60*100?
	    }
	    if ($5 =~ /^[SW]$/) {
		$self->{$pos} = -$self->{$pos};
	    }
	} else {
	    warn "Can't parse $self->{$pos}";
	}
    }
}

######################################################################

package GIS::NSD;

sub new {
    my($class, %args) = @_;

    my $self = {};
    bless $self, $class;

    if ($args{-file}) {
	$self->{File} = $args{-file};
    }

    $self;
}

sub read_file {
    my $self = shift;
    open(F, $self->{File})
	or die "Can't open $self->{File}: $!";
    $self->{Data} = [];
    while(<F>) {
	chomp;
	my $row = new GIS::NSD::Record;
	@{$row}{qw/abbrev x1 x2 city state country x3
		   lat1 long1 lat2 long2 height1 height2 x4/} =
		       split /;/;
	$row->_position_to_decimal;
	push @{ $self->{Data} }, $row;
    }
    close F;
    $self->{Data};
}

return 1 if caller;

package main;

my $nsd = new GIS::NSD
    -file => "/usr/local/src/tkGeoMap/data/places/nsd_cccc.txt";
$nsd->read_file;
for (@{$nsd->{Data}}) {
    use Data::Dumper;
    print STDERR "Line " . __LINE__ . ", File: " . __FILE__ . "\n" 
	. Data::Dumper->Dumpxs([$_],[]);
}

__END__
