# -*- perl -*-

#
# $Id: Polar.pm,v 1.4 2000/01/13 22:11:53 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

# Die X0 etc.-Parameter basieren auf Dezimalgraden, die Ausgabe auf
# Grad/Minuten

package Karte::Polar;
use Karte;
use strict;
use vars qw(@ISA $obj);

@ISA = qw(Karte);

sub new {
    my $class = shift;
    my $self =
      {
       Name     => 'Polare Koordinaten',
       Token    => 'polar',
       Coordsys => 'P',
       
       X0 => -790086.391400205,
       X1 => 67845.9393462968,
       X2 => -2073.78902640448,
       Y0 => -5824916.26148241,
       Y1 => 1021.55512610816,
       Y2 => 110888.031499432,

#       Scrollregion => [0, 0, 11778, 9724],
#       Scrollregion => [0, -3000, 11778, 9724],
      };
    bless $self, $class;
}

$obj = new Karte::Polar;

sub convert_to_route {
    my $file = shift;
    my @res;
    open(GISTRACK, $file) or die "Can't open $file: $!";
    while(<GISTRACK>) {
	chomp;
	if (m|^TRK\s+(.)(\d+)\s+([\d\.]+)\s+(.)(\d+)\s+([\d\.]+)\s+(\d+)[/-](\d+)[/-](\d+)[- ](\d+):(\d+):(\d+)\s+(\d+)$|) {
	    my($ns, $breite_grad, $breite_min,
	       $ew, $laenge_grad, $laenge_min,
	       $year, $mon, $day, $hour, $min, $sec, $xxx) =  ($1, $2, $3,
							       $4, $5, $6,
							       $7, $8, $9,
							       $10, $11, $12,
							       $13);
	    $breite_grad *= -1 if ($ns eq 'S');
	    $laenge_grad *= -1 if ($ew eq 'W');
	    my($breite, $laenge) = ($breite_grad + $breite_min/60,
				    $laenge_grad + $laenge_min/60);
	    push @res, [$obj->map2standard($laenge, $breite)];
	}
    }
    close GISTRACK;
    @res;
}

sub save_as_route {
    my $file = shift;
    require Data::Dumper;
    my @res = convert_to_route($file);
    $Data::Dumper::Indent = 0;
    print Data::Dumper->Dumpxs([\@res], ['realcoords_ref']);
}

1;
