# -*- perl -*-

#
# $Id: PLZ.pm,v 1.22 2000/11/28 01:46:14 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998, 2000 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package PLZ;
use strict;
use vars qw(@plzfile $OLD_AGREP);
use locale;
use BBBikeUtil;

use constant FMT_NORMAL  => 0;
use constant FMT_REDUCED => 1;
use constant FMT_COORDS  => 2;

@plzfile = 
  ((map { ("$_/Berlin.coords.data", "$_/data/Berlin.coords.data") } @INC),
   (map { ("$_/Berlin.small.data",  "$_/data/Berlin.small.data",
	   "$_/berlinsm.dat",       "$_/data/berlinsm.dat",
	   "$_/Berlin.data",        "$_/data/Berlin.data") } @INC),
   "/usr/www/soc/plz/Berlin.data",
   "/home/pub/lib/plz/Berlin.data",
  );
$OLD_AGREP = 0 unless defined $OLD_AGREP;

sub new {
    my($class, $file) = @_;
    my $self = {};
    if (!defined $file) {
	foreach (@plzfile) {
	    if (-r $_) {
		$file = $_;
		if (!open(DATA, $file)) {
		    $file = "$file.gz";
		    next if !-f $file || !open(DATA, "zcat $file|");
		    $self->{IsGzip} = 1;
		} else {
		    $self->{IsGzip} = 0;
		}
		my($line) = <DATA>;
		$line =~ s/[\015\012]//g;
		my(@l) = split(/\|/, $line);
		if (@l == 3) {
		    $self->{DataFmt}  = FMT_REDUCED;
		    $self->{FieldPLZ} = 2;
		} elsif (@l == 4) {
		    $self->{DataFmt}  = FMT_COORDS;
		    $self->{FieldPLZ} = 2;
		} else {
		    $self->{DataFmt} = FMT_NORMAL;
		    $self->{FieldPLZ} = 4; # XXX ja?
		}
		close DATA;
		last;
	    }
	}
    }
    return undef if !defined $file || !-r $file;
    $self->{File} = $file;
    bless $self, $class;
}

sub open {
    my($self, %args) = @_;
    my $file = $args{File} || $self->{File};
    if ($file ne $self->{Data}) { # XXX hh???
	my @data;
	open(PLZ, $file)
	  or die "Die Datei $file kann nicht geffnet werden: $!";

	my $code = <<'EOF';
	while(<PLZ>) {
	    chomp;
	    my(@l) = split(/\|/, $_);
EOF
	my $push_code;
	if ($self->{DataFmt} == FMT_REDUCED) {
	    $push_code = q{push @data, [@l[0 .. 2]]};
	} elsif ($self->{DataFmt} == FMT_COORDS) {
	    $push_code = q{push @data, [@l[0 .. 3]]};
	} else {
	    $push_code = q{push @data, [@l[0, 1, $self->{FieldPLZ}]]};
	}
	$code .= $push_code . <<'EOF';
	}
EOF
        eval $code;
	close PLZ;
	$self->{Data} = \@data;
	$self->{File} = $file;
	undef $self->{NameHash};
	undef $self->{PlzHash};
    }
}

sub make_plz_re {
    my($self, $plz) = @_;
    if ($self->{DataFmt} == FMT_REDUCED ||
	$self->{DataFmt} == FMT_COORDS) {
	'^[^|]*|[^|]*|' . $plz;
    } else {
	'^[^|]*|[^|]*|[^|]*|[^|]*|' . $plz . '|';
    }
}

# XXX make gzip-aware
# Argumente: (Beschreibung fehlt XXX)
#  Agrep
#  Noextern
#  Citypart (optionale Einschrnkung auf einen Bezirk)
#  MultiCitypart
#  MultiZIP
# Ausgabe: Array von Referenzen [strasse, bezirk, plz]
sub look {
    my($self, $str, %args) = @_;
    my $file = $args{File} || $self->{File};
    my @res;

    my $grep_type = ($args{Agrep} ? 'agrep' : 'grep');
    my @push_inx;
    if      ($self->{DataFmt} == FMT_NORMAL) {
	@push_inx = (0, 1, $self->{FieldPLZ});
    } elsif ($self->{DataFmt} == FMT_REDUCED) {
	@push_inx = (0, 1, 2);
    } else {
	@push_inx = (0, 1, 2, 3);
    }
    if ($grep_type eq 'agrep') {
	if ($OLD_AGREP ||
	    (!$args{Noextern} && !is_in_path('agrep'))) {
	    $args{Noextern} = 1;
	}
	if ($args{Noextern}) {
	    eval q{local $SIG{'__DIE__'};
		   require String::Approx;
		   String::Approx->VERSION(2.7);
	       };
	    if ($@) {
		$grep_type = 'grep';
	    }
	}
    }
    if ($grep_type eq 'grep') {
	if (!$args{Noextern} && !is_in_path('grep')) {
	    $args{Noextern} = 1;
	}
    }

    my %res;
    my $push_sub = sub {
	my(@to_push) = (split(/\|/, $_[0]))[@push_inx];
	if (($args{MultiCitypart}||!exists $res{$to_push[0]}->{$to_push[1]}) &&
	    ($args{MultiZIP}     ||!exists $res{$to_push[0]}->{$to_push[2]})) {
	    push @res, [@to_push];
	    last if defined $args{Max} and $args{Max} < $#res;
	    $res{$to_push[0]}->{$to_push[1]}++;
	    $res{$to_push[0]}->{$to_push[2]}++;
	}
    };

    if (!$args{Noextern}) {
	unless ($args{Noquote}) {
	    if ($grep_type eq 'grep') {
		$str =~ s/([\\.*\[\]])/\\$1/g; # quote metacharacters
	    } else {
		$str =~ s/([\$\^\*\[\]\^\|\(\)\!\`])/\\$1/g;
	    }
	    $str = "^$str";
	}

	# limitation of agrep:
	if (length($str) > 32) {
	    $str = substr($str, 0, 32);
	}

	my(@grep_args) = ('-i', $str, $file);
	if ($grep_type eq 'agrep' && $args{Agrep}) {
	    unshift @grep_args, "-$args{Agrep}";
	}
	CORE::open(PLZ, "-|") or exec $grep_type, @grep_args;
	my %res;
	while(<PLZ>) {
	    chomp;
	    $push_sub->($_);
#	    my(@to_push) = (split(/\|/, $_))[@push_inx];
#  	    if (!exists $res{$to_push[0]}->{$to_push[1]} &&
#  		!exists $res{$to_push[0]}->{$to_push[2]}) {
#  		push(@res, [@to_push]);
#  		last if defined $args{Max} and $args{Max} < $#res;
#  		$res{$to_push[0]}->{$to_push[1]}++;
#  		$res{$to_push[0]}->{$to_push[2]}++;
#  	    }
	}
	close PLZ;
    } else {
	CORE::open(PLZ, $file)
	  or die "Die Datei $file kann nicht geffnet werden: $!";
	if ($grep_type eq 'agrep') {
	    chomp(my @data = <PLZ>);
	    close PLZ;
	    my %res;
	    foreach (String::Approx::amatch($str,
					    ['i', $args{Agrep}], @data)) {
		$push_sub->($_);
#  		my(@to_push) = (split(/\|/, $_))[@push_inx];
#  		if (!exists $res{$to_push[0]}->{$to_push[1]} &&
#  		    !exists $res{$to_push[0]}->{$to_push[2]}) {
#  		    push(@res, [@to_push]);
#  		    last if defined $args{Max} and $args{Max} < $#res;
#  		    $res{$to_push[0]}->{$to_push[1]}++;
#  		    $res{$to_push[0]}->{$to_push[2]}++;
#  		}
	    }
	} else {
	    $str = "^$str" unless $args{Noquote};
	    $str =~ s/\|/\\|/g;
	    my %res;
	    while(<PLZ>) {
		chomp;
		if (/$str/i) {
		    $push_sub->($_);
		}
#  		my(@to_push) = (split(/\|/, $_))[@push_inx];
#  		if (/$str/i &&
#  		    !exists $res{$to_push[0]}->{$to_push[1]} &&
#  		    !exists $res{$to_push[0]}->{$to_push[2]}) {
#  		    push(@res, [@to_push]);
#  		    last if defined $args{Max} and $args{Max} < $#res;
#  		    $res{$to_push[0]}->{$to_push[1]}++;
#  		    $res{$to_push[0]}->{$to_push[2]}++;
#  		}
	    }
	    close PLZ;
	}
    }
    if (defined $args{Citypart}) {
	my @new_res;
	foreach (@res) {
	    if ($_->[1] =~ /^\Q$args{Citypart}\E/i ||
		$_->[$self->{FieldPLZ}] =~ /^\Q$args{Citypart}\E/) {
		push @new_res, $_;
	    }
	}
	@res = @new_res;
    }
    @res;
}

# Match-Reihenfolge:
# * nicht modifiziert ohne Agrep
# * "strasse" nach "str." umgewandelt ohne Agrep
# * inkrementell bis $args{Agrep} mit Agrep abwechselnd nicht modifiziert und
#   mit s/strasse/str./
# Argumente in %args:
#   Agrep: 0, wenn grep verwendet werden soll
#          >0, wenn mit Fehlern gesucht werden, dann gibt der Wert die
#              maximale Anzahl der erlaubten Fehler an
#          'default', wenn der Standardwert von 3 Fehlern genommen werden soll
#                     Bei lngeren Wrtern wird die Maximalanzahl bis 5 erhht.
# Sonstige Argumente werden nach look() durchgereicht.
# Ausgabe:
#   erstes Element: siehe look() (als Arrayreferenz)
#   zweites Element: Anzahl der Fehler fr das Ergebnis
sub look_loop {
    my($self, $str, %args) = @_;
    my $max_agrep;
    if (defined $args{Agrep} && $args{Agrep} eq 'default') {
	$max_agrep = 3;
	if    (length($str) > 15) { $max_agrep = 4 }
	elsif (length($str) > 25) { $max_agrep = 5 }
	delete $args{Agrep};
    } else {
	$max_agrep = delete $args{Agrep} || 0;
    }
    my $strip_strasse = sub {
	my $str = shift;
	if ($str =~ /stra(ss|)e/i) {
	    $str =~ s/(s)tra(ss|)e/$1tr./i;
	    $str;
	} else {
	    undef;
	}
    };
    my $agrep = 0;
    my @matchref;
    @matchref = $self->look($str, %args);
    if (!@matchref) {
	if (my $str = $strip_strasse->($str)) {
	    @matchref = $self->look($str, %args);
	}
	if (!@matchref) {
	    $agrep = 1;
	    while ($agrep <= $max_agrep) {
		@matchref = $self->look($str, %args, Agrep => $agrep);
		if (!@matchref && (my $str = $strip_strasse->($str))) {
		    @matchref = $self->look($str, %args, Agrep => $agrep);
		}
		last if @matchref;
		$agrep++;
	    }
	}
    }
    (\@matchref, $agrep);
}

# Sortiert die Straen eines look_loop-Ergebnisses.
# Argumente und Rckgabewerte sind vom gleichen Format wie bei look_loop.
sub look_loop_best {
    my($self, $str, @args) = @_;
    my($matchref, $agrep) = $self->look_loop($str, @args);
    if (@$matchref) {
	my @rating;
	for(my $i=0; $i<=$#$matchref; $i++) {
	    my $item = $matchref->[$i];
	    if ($item->[0] eq $str) {
		push @rating, [ 100, $item ];
	    } elsif ($item->[0] =~ /^$str/i) {
		push @rating, [ 40 + 40-length($item->[0]), $item ];
	    } else {
		push @rating, [ 40-length($item->[0]), $item ];
	    }
	}
	([map  { $_->[1] } sort { $b->[0] <=> $a->[0] } @rating], $agrep);
    } else {
	($matchref, $agrep);
    }
}

sub make_name_hash {
    my $self = shift;
    my %hash;
    my $i = 0;
    foreach (@{$self->{Data}}) {
	$hash{$_->[0]} = $i;
	$i++;
    }
    $self->{NameHash} = \%hash;
}

sub make_plz_hash {
    my $self = shift;
    my %hash;
    my $i = 0;
    foreach (@{$self->{Data}}) {
	$hash{$_->[2]} = $i;
	$i++;
    }
    $self->{PlzHash} = \%hash;
}

return 1 if caller();

######################################################################
#
# standalone program
#
package main;
require Getopt::Long;

my $agrep = "default";
my $extern = 1;
my $citypart;
my $multi_citypart = 0;
my $multi_zip = 0;

if (!Getopt::Long::GetOptions
    ("agrep=i" => \$agrep,
     "extern!" => \$extern,
     "citypart=s" => \$citypart,
     "multicitypart!" => \$multi_citypart,
     "multizip!" => \$multi_zip,
     )
   ) {
    die "Usage: $0 [options]";
}

my $street = shift || die "Street?";

my $plz = PLZ->new;
$plz->open;

my @args;
push @args, "Agrep", $agrep;
if (!$extern) {
    push @args, "Noextern", 1;
}
if (defined $citypart and $citypart ne "") {
    push @args, "Citypart", $citypart;
}
if ($multi_citypart) {
    push @args, "MultiCitypart", 1;
}
if ($multi_zip) {
    push @args, "MultiZIP", 1;
}

my($res_ref, $errors) = $plz->look_loop($street, @args);
foreach my $res (@$res_ref) {
    printf "%-40s %-20s %-10s\n", @$res;
}
print "*** Errors: $errors\n";

