# -*- perl -*-

#
# $Id: Strassen.pm,v 2.46 2000/11/30 14:55:48 eserte Exp $
# Author: Slaven Rezic (eserte@cs.tu-berlin.de)
#

BEGIN {
    if (!defined &main::mymstat) {
	eval q{
	    package main; # AUTOLOAD: ignore
	    sub mymstat { 1 }
	};
	warn $@ if $@;
    }
}

use strict;
use FindBin;
use lib ("$FindBin::RealBin/lib"); #, "/home/e/eserte/lib/perl");

######################################################################
package Strassen::Util;
use BBBikeUtil;
#use AutoLoader 'AUTOLOAD';
use vars qw($VERBOSE $tmpdir
	    $cachedir $cacheable @cacheable %cache_symbol
	    $datadumper_var $use_virt_array);

if (defined $main::tmpdir) {
    $tmpdir = $main::tmpdir;
} else {
    $tmpdir = $ENV{TMPDIR} || $ENV{TEMP} || "/tmp";
}
if (!-d $tmpdir || !-w $tmpdir) { undef $tmpdir }
$cachedir = (-d "$FindBin::RealBin/cache"
	     ? "$FindBin::RealBin/cache"
	     : $tmpdir)
    if !defined $cachedir;
$cacheable = 1
    if !defined $cacheable;
# FreezeThaw nicht verwenden: es ist *viel* langsamer als Storable
# Data::Dumper ist auch langsam gegenber make_net_slow
# Werte (mit dprofpp, perl5.00503, FreeBSD 3.4):
#  FreezeThaw:    55s (!)
#  Data::Dumper:  5s
#  ohne Cache:    2s
@cacheable = qw(VirtArray Storable)
    if !@cacheable;
$use_virt_array = 1;

# Argument: [x1,y1], [x2, y2]
sub strecke {
    CORE::sqrt(sqr($_[0]->[0] - $_[1]->[0]) +
	       sqr($_[0]->[1] - $_[1]->[1]));
}

# Argument: "x1,y1", "x2,y2"
sub strecke_s {
    my($x1, $y1) = split(/,/, $_[0]);
    my($x2, $y2) = split(/,/, $_[1]);
    CORE::sqrt(sqr($x1-$x2) + sqr($y1-$y2));
}

# Argumente: Indices
sub strecke_i {
    my($self, $i1, $i2) = @_;
    my($x1,$y1) = unpack("l2", $self->{Index2Coord}[$i1]);
    my($x2,$y2) = unpack("l2", $self->{Index2Coord}[$i2]);
    CORE::sqrt(sqr($x1-$x2) + sqr($y1-$y2));
}

# gibt "l" oder "r" fr Links- oder Rechtsabbiegen aus.
### AutoLoad Sub
sub abbiegen {
    my($p0,$p1,$p2) = @_;
    my($x0,$y0, $x1,$y1, $x2,$y2) = (split(/,/, $p0),
				     split(/,/, $p1),
				     split(/,/, $p2));
    my $a1 = $x1-$x0;
    my $a2 = $y1-$y0;
    my $b1 = $x2-$x1;
    my $b2 = $y2-$y1;
    ($a1*$b2-$a2*$b1 > 0 ? 'l' : 'r');
}

# gibt fr "x,y"  [x,y] aus
sub string_to_coord {
    my $s = shift;
    split(/,/, $s);
}

# Fr einen Punkt $p und einen Nachbarpunkt $neighbor_p die Himmelsrichtung
# feststellen. Punkte in "x,y"-Syntax.
# XXX $rev_y: TRUE, wenn die y-Achse geometrisch umgekehrt/computergrafisch
# korrekt ist (unten ist das gleiche "mirrored" genannt).
### AutoLoad Sub
sub get_direction {
    my($p, $neighbor_p, $rev_y) = @_;
    my($px,$py)   = split /,/, $p;
    my($npx,$npy) = split /,/, $neighbor_p;
    my $deg = rad2deg(atan2($npy-$py, $npx-$px))+22.5;
    require POSIX;
    $deg = POSIX::floor($deg/45)*45;
    if (!$rev_y) {
	return {0 => 'w', 45 => 'nw', 90 => 'n', 135 => 'ne', 180 => 'e',
		-180 => 'e', -135 => 'se', -90 => 's', -45 => 'sw'}->{$deg};
    } else {
	return {0 => 'e', 45 => 'se', 90 => 's', 135 => 'sw', 180 => 'w',
		-180 => 'w', -135 => 'nw', -90 => 'n', -45 => 'ne'}->{$deg};
    }
}

# Den besten Nachbarpunkt aus $p_ref fr einen Punkt $p in der
# gegebenen Himmelsrichtung $dir zurckgeben. Punkte in "x,y"-Syntax.
# Werte stimmen fr Edit-Berlin-Modus. Fr den Normal-Modus mu das
# Argument $mirrored auf wahr gesetzt werden.
### AutoLoad Sub
sub best_from_direction {
    my($p, $p_ref, $dir, $mirrored) = @_;
    my($px,$py)   = split /,/, $p;
    my %angle;
    $dir = lc($dir);
    $dir =~ s/o/e/g; # deutsch => englisch
    my %dir_angle;
    if ($mirrored) {
	%dir_angle = ('w' => 0, 'nw' => 45, 'n' => 90, 'ne' => 135, 'e' => 180,
		      'e' => -180, 'se' => -135, 's' => -90, 'sw' => -45);
    } else {
	%dir_angle = ('e' => 0, 'se' => 45, 's' => 90, 'sw' => 135, 'w' => 180,
		      'w' => -180, 'nw' => -135, 'n' => -90, 'ne' => -45);
    }
    if (!exists $dir_angle{$dir}) {
	warn "Can't get angle direction for $p: <$dir>";
    }
    my $dir_angle = $dir_angle{$dir};
    foreach my $neighbor_p (@$p_ref) {
	my($npx,$npy) = split /,/, $neighbor_p;
	my $angle = rad2deg(atan2($npy-$py, $npx-$px));
	my $min_angle;
	for (-360, 0, 360) {
	    if (!defined $min_angle or
		abs($angle + $_ - $dir_angle) < $min_angle) {
		$min_angle = abs($angle + $_ - $dir_angle);
	    }
	}
	$angle{$neighbor_p} = $min_angle;
    }
    my $best_angle;
    my $best_p;
    foreach my $neighbor_p (@$p_ref) {
	if (!defined $best_angle or abs($angle{$neighbor_p}) < $best_angle) {
	    $best_angle = abs($angle{$neighbor_p});
	    $best_p = $neighbor_p;
	}
    }
    $best_p;
}

### AutoLoad Sub
sub cache_ext {
    {'VirtArray'    => '.va',
     'Storable'     => '.st',
     'FreezeThaw'   => '.ft',
     'Data::Dumper' => '.pl',
    }->{$_[0]};
}

### AutoLoad Sub
sub try_cache {
    my($filename, $write, $ref, %args) = @_;
    my $cache_type;
    my $rw_text = ($write ? 'write' : 'read');
    foreach $cache_type (@cacheable) {
	my $filename = $filename . cache_ext($cache_type);
	warn "$rw_text $filename: trying cache type $cache_type...\n"
	  if $VERBOSE;
	if ($cache_type eq 'VirtArray') {
	    # soll VirtArray verwendet werden?
	    next unless $use_virt_array;
	    # wird berhaupt ein Array gespeichert?
	    next if defined $ref && ref $ref ne 'ARRAY';
	    # Arrays mit Referenzen knnen mit VirtArray nicht gespeichert
	    next if $args{-deeparray};
	    eval q{ local $SIG{'__DIE__'};
		    require VirtArray;
		    die "VirtArray kann kein TIEARRAY"
		      unless VirtArray->can("TIEARRAY");
		};
	    if (!$@) {
		if (!$write) {
		    # ist die existierende Datei berhaupt
		    # eine VirtArray-Datei?
		    next if !VirtArray::is_valid($filename);
		}
		return ($cache_type, $filename);
	    }
	} elsif ($cache_type eq 'Storable') {
	    eval q{ local $SIG{'__DIE__'};
		    require Storable;
		    Storable->VERSION(0.509);
		};
	    if (!$@) {
		next if !$write && !-f $filename; # Datei existiert nicht
		return ($cache_type, $filename);
	    }
	} elsif ($cache_type eq 'FreezeThaw') {
	    eval q{ local $SIG{'__DIE__'};
		    require FreezeThaw;
		};
	    if (!$@) {
		next if !$write && !-f $filename; # Datei existiert nicht
		return ($cache_type, $filename);
	    }
	} elsif ($cache_type eq 'Data::Dumper') {
	    if ($write) {
		eval q{ local $SIG{'__DIE__'};
			require Data::Dumper;
			Data::Dumper->VERSION(2.10);
			# use only the fast version
		        Data::Dumper->can("Dumpxs");
		    };
		if (!$@) {
		    return ($cache_type, $filename);
		}
	    } else {
		if (-f $filename) {
		    return ($cache_type, $filename);
		}
	    }
	} else {
	    die "Unknown cache type $cache_type";
	}
    }
    undef;
}

# Gibt wahr zurck, wenn der Cache nicht lter als alle Source-Dateien
# in $srcref (Array-Referenz oder ein String) ist.
### AutoLoad Sub
sub valid_cache {
    my($cachefile, $srcref) = @_;

    my(@stat_cache) = stat $cachefile;
    if (!defined $stat_cache[9]) { # cachefile nicht vorhanden
	return undef;
    }
    my $src;
    my(@src) = (ref $srcref eq 'ARRAY' ? @$srcref : $srcref);
    foreach $src (@src) {
	my(@stat_orig)  = stat $src;
	if (defined $stat_orig[9] &&
	    $stat_cache[9] < $stat_orig[9]) {
	    # Cache ist nicht gltig
	    return undef;
	}
    }
    1;
}

### AutoLoad Sub
sub get_from_cache {
    my($cachefile, $srcref) = @_;
    return if !$cacheable || !@cacheable;

    if (!defined $cachedir) {
	$cacheable = 0;
	return undef;
    }

    $cachefile = "$cachedir/$cachefile" unless $cachefile =~ m|^/|;
    my($cache_func_found, $cachepath) = try_cache($cachefile);
    return undef if !$cache_func_found;
    warn "OK\n" if $VERBOSE;

    return undef if !valid_cache($cachepath, $srcref);

    if ($cache_func_found eq 'Storable') {
	# Cache ist gltig
	my $obj;
	eval {
	    $obj = Storable::retrieve($cachepath);
	};
	warn $@ if $@;
	$obj;
    } elsif ($cache_func_found eq 'FreezeThaw') {
	if (!open(CACHE, $cachepath)) {
	    warn "Can't read $cachepath: $!";
	    return;
	}
	local $/ = undef;
	binmode CACHE;
	my $buf = <CACHE>;
	close CACHE;

	my $obj;
	eval {
	    ($obj) = FreezeThaw::thaw($buf);
	};
	warn $@ if $@;
	$obj;
    } elsif ($cache_func_found eq 'Data::Dumper') {
	do $cachepath;
	my $x = $datadumper_var;
	undef $datadumper_var; # evtl. memory leak vermeiden
	$x;
    } elsif ($cache_func_found eq 'VirtArray') {
	my @a;
	tie @a, 'VirtArray', $cachepath;
	\@a;
    } elsif (defined $cache_func_found) {
	warn "Unknown cache function: $cache_func_found";
    }
}

### AutoLoad Sub
sub write_cache {
    my($ref, $cachefile) = @_;
    return if !$cacheable || !@cacheable;

    $cachefile = "$cachedir/$cachefile" unless $cachefile =~ m|^/|;
    my($cache_func_found, $cachepath) = try_cache($cachefile, 1, $ref);
    return undef if !$cache_func_found;
    warn "OK\n" if $VERBOSE;

    if ($cache_func_found eq 'Storable') {
	Storable::store($ref, $cachepath);
    } elsif ($cache_func_found eq 'FreezeThaw') {
	my $buf = FreezeThaw::freeze($ref);
	if (open(CACHE, ">$cachepath")) {
	    print CACHE $buf;
	    close CACHE;
	} else {
	    warn "Can't write to $cachepath: $!";
	}
    } elsif ($cache_func_found eq 'Data::Dumper') {
	$Data::Dumper::Indent = 0; # write as tight as possible
	if (open(DD, ">$cachepath")) {
	    binmode DD;
	    print DD Data::Dumper->Dumpxs([$ref], ['datadumper_var']);
	    close DD;
	    1;
	} else {
	    undef;
	}
    } elsif ($cache_func_found eq 'VirtArray') {
	VirtArray::store($ref, $cachepath);
    } elsif (defined $cache_func_found) {
	warn "Unknown cache function: $cache_func_found";
    }
}

main::mymstat("vor Strassen");

1;

######################################################################
package Strassen;
use BBBikeUtil;
#require StrassenNetz; # AUTOLOAD: activate
#require MultiStrassen; # AUTOLOAD: activate
#require Strassen::Util; # AUTOLOAD: activate
#require Strasse; # AUTOLOAD: activate
#use AutoLoader 'AUTOLOAD';
use vars qw(@datadirs $OLD_AGREP $VERBOSE $VERSION);

$VERSION = sprintf("%d.%02d", q$Revision: 2.46 $ =~ /(\d+)\.(\d+)/);

@datadirs = ("$FindBin::RealBin/data", './data');
foreach (@INC) {
    push @datadirs, "$_/data";
}
# XXX
push @datadirs, "http://www/~eserte/bbbike/root/data";

$OLD_AGREP = 0;

# static method to set the datadirs variable according to the used geography
# object
sub set_geography {
    my $geo = shift;
    @datadirs = $geo->datadir;
}

sub new {
    my($class, $filename, %arg) = @_;
    my(@filenames);
    if (defined $filename) {
	push @filenames, $filename, map { $_ . "/$filename" } @datadirs;
    }
    my $self = {};
    bless $self, $class;

    if (@filenames) {
      TRY: {
	    my $file;
	    foreach $file (@filenames) {
		if (-f $file and -r _) {
		    $self->{File} = $file;
		    last TRY;
		}
		my $gzfile = "$file.gz";
		if (-f $gzfile and -r _) {
		    $self->{File} = $gzfile;
		    $self->{IsGzipped} = 1;
		    last TRY;
		}
	    }
	    # XXX 2. versuch mit internet fr minimale Funktionsfhigkeit
	    # mit bbbike.ppl
	    foreach $file (@filenames) {
		eval {
		    use lib "/home/e/eserte/src/perl/Hyper";
		    require Hyper;
		    my $cachefile = Hyper::hypercopy($file);
		    $self->{File} = $cachefile;
		};
		last TRY if $self->{File};
	    }
	    # XXX end

	    die "Can't open ", join(", ", @filenames);
	}
	unless ($arg{NoRead}) {
	    $self->read_data;
	}
    }

    $self->{Pos}   = 0;

    $self;
}

sub read_data {
    my($self) = @_;
    my $file = $self->{File};
    if ($self->{IsGzipped}) {
	die "Can't execute zcat $file" if !open(FILE, "zcat $file |");
    } else {
	die "Can't open $file" if !open(FILE, $file);
    }
    binmode FILE;
    my @data;
    my $line;
    while (defined($line = <FILE>)) {
	next if $line =~ /^\#/ || $line =~ /^\s*$/;
	push @data, $line;
    }
    close FILE;
    $self->{Data}  = \@data;
}

### AutoLoad Sub
sub new_from_data {
    my($class, @data) = @_;
    my $self = {};
    $self->{Data} = \@data;
    $self->{Pos}  = 0;
    bless $self, $class;
}

# Erzeugt ein neues Strassen-Objekt mit Restriktionen
### AutoLoad Sub
sub new_copy_restricted {
    my($class, $old_s, %args) = @_;
    my %restrictions;
    if ($args{-restrictions}) {
	%restrictions = map { ($_ => 1) } @{ $args{-restrictions} };
    }

    my $res = $class->new;
    $old_s->init;
    while(1) {
	my $ret = $old_s->next;
	last if !@{$ret->[1]};
	next if $restrictions{$ret->[2]};
	$res->push($ret);
    }
    $res;
}

# Erzeugt aus dem Objekt eine Hash-Referenz mit erster Koordinate als Key
# und der Kategorie als Value. Ist nur fr ein-Punkt-Daten geeignet.
# init()/next() wird verwendet!
### AutoLoad Sub
sub get_hashref {
    my($self) = @_;
    my $hash = {};

    $self->init;
    while(1) {
	my $ret = $self->next;
	last if !@{$ret->[1]};
	my $cat = $ret->[0];
	$hash->{$ret->[1][0]} = $cat;
    }

    $hash;
}

# Ausgabe des Source-Files
sub file { shift->{File} }

# ID (fr Caching)
sub id {
    require File::Basename;
    File::Basename::basename(shift->{File});
}

### AutoLoad Sub
sub write {
    my($self, $filename) = @_;
    if (!defined $filename) {
	$filename = $self->file;
    }
    if (!defined $filename) {
	return 0;
    }
    if (open(COPY, ">$filename")) {
	binmode COPY;
	print COPY join("", @{$self->{Data}});
	close COPY;
	1;
    } else {
	warn "Can't write to $filename: $!" if $VERBOSE;
	0;
    }
}

sub get {
    my($self, $pos) = @_;
    my $line = $self->{Data}->[$pos];
    parse($line);
}

# Fr den angegebenen Namen wird die erste gefundene Zeile im selben Format
# wie bei get(), next() und parse() zurckgegeben.
# Achtung: da mit init() und next() gearbeitet wird, wird durch diese Methode
# eine laufende Schleife aus dem Konzept gebracht!
### AutoLoad Sub
sub get_by_name {
    my($self, $name) = @_;
    $self->init;
    while(1) {
	my $ret = $self->next;
	return undef if !@{$ret->[1]};
	return $ret if ($ret->[0] eq $name);
    }
}

# XXX Die zwei verschiedenen Aufrufarten fr das Koordinatenargument in
# set und push ist unbefriedigend.
### AutoLoad Sub
sub set {
    my($self, $index, $arg) = @_;
    $self->{Data}[$index] = arr2line($arg);
}

sub push {
    my($self, $arg) = @_;
    my $x = [$arg->[0], join(" ", @{$arg->[1]}), $arg->[2]];
    push @{$self->{Data}}, arr2line($x);
}

# wandelt eine Array-Referenz ["name", [Koordinaten], "cat"] in
# einen String zum Abspeichern um
### AutoLoad Sub
sub arr2line {
    my $arg = shift;
    "$arg->[0]\t$arg->[2] $arg->[1]\n"
}

sub parse {
    my $line = shift;
    return [undef, [], undef] if !$line;
    my $name;
    ($name, $line) = split /\t/, $line, 2;
    my @s = split /\s+/, $line;
    my $category = shift @s;
    [$name, \@s, $category];
}

### AutoLoad Sub
sub parse_rigid {
    my($name, $resref, $category) = @{parse(shift)};
    my @res;
    foreach (@$resref) {
	push(@res, $_) if $_ =~ /^\d+$/ || $_ =~ /^.?-?\d+,-?\d+$/;
    }
    [$name, \@res, $category];
}

### AutoLoad Sub
sub get_obj {
    my($self, $pos) = @_;
    Strasse->new($self->get($pos));
}

# initialisiert fr next() und gibt *keinen* Wert zurck
sub init {
    my $self = shift;
    $self->{Pos} = -1;
}

# initialisiert fr next() und gibt den ersten Wert zurck
### AutoLoad Sub
sub first {
    my $self = shift;
    $self->{Pos} = 0;
    $self->get(0);
}

sub next {
    my $self = shift;
    $self->get(++($self->{Pos}));
}

sub next_obj {
    my $self = shift;
    $self->get_obj(++($self->{Pos}));
}

# XXX wird das hier verwendet? Schnerer Ersatz fr !@{$ret->[1]} ?
sub at_end {
    my $self = shift;
    $self->{Pos} >= $#{$self->{Data}};
}

sub count {
    my $self = shift;
    scalar @{$self->{Data}};
}

# gibt die aktuelle Position zurck
sub pos { shift->{Pos} }

# Gibt die Positionen (als Array) fr einen bestimmten Namen zurck
# Achtung: eine laufende init/next-Schleife wird hiermit zurckgesetzt!
### AutoLoad Sub
sub pos_from_name {
    my($self, $name) = @_;
    my @res;
    my $found = 0;
    $self->init;
    while(1) {
	my $ret = $self->next;
	last if !@{$ret->[1]};
	if ($ret->[0] eq $name) {
	    push @res, $self->pos;
	    $found++;
	} elsif ($found) {
	    last;
	}
    }
    @res;
}

# Statische Methode.
# Wandelt die Indices aus dem Ergebnis von get() (2. Element) in
# Koordinaten um (Format des Arguments: ["x1,y1", "x2,y2", ...])
# Gibt eine Referenz auf ein Array zurck: [[x1,y1], [x2,y2] ...]
sub to_koord_slow {
    my($resref) = @_;
    my @res;
    foreach (@$resref) {
	if (/^(-?\d+),(-?\d+)$/) {
	    push(@res, [$1, $2]);
	} elsif (/(-?\d+),(-?\d+)$/) { # ignore prefix XXX
	    push(@res, [$1, $2]);
	} else {
	    warn "Unrecognized reference: $_";
	    return [];
	}
    }
    \@res;
}

# Statische Methode.
# wie to_koord, nur fr einen Punkt
# XXX Koordinaten der Form prefix(x,y) bearbeiten
sub to_koord1_slow {
    my($s) = @_;
    if ($s =~ /^(-?\d+),(-?\d+)$/) {
	[$1, $2];
    } elsif ($s =~ /^((:[^:]*:)?([A-Za-z])?)?(-?\d+),(-?\d+)$/) {
	# Ausgabe: x, y, coordsystem, bahnhof
	[$4, $5, $3, $2];
    } else {
	warn "Unrecognized string: $s...";
	[undef, undef]; # XXX
    }
}

*to_koord = \&to_koord_slow;
*to_koord1 = \&to_koord1_slow;

# Gibt alle echten Kreuzungen in einer Liste oder als Hash aus.
# Argumente:
#   RetType: hash, hashpos, array oder arraypos
#            Bei den ...pos-Varianten wird statt des Straennamens die
#            Position im Strassen-Objekt zurckgegeben.
#   UseCache: gibt an, ob vom Cache gelesen und ein Cache geschrieben werden
#             soll
#   Kurvenpunkte: bei TRUE werden auch die Kurvenpunkte zurckgegeben
#
### AutoLoad Sub
sub all_crossings {
    my($self, %args) = @_;
    my $rettype      = $args{RetType};
    my $use_cache    = $args{UseCache};
    my $min_strassen = ($args{'Kurvenpunkte'} ? 1 : 2);

    if (!defined $rettype) { $rettype = 'array' }
#XXX    require File::Basename;
    my $basename = $self->id; #File::Basename::basename($self->{File});
    my $cachefile = "all_crossings_${basename}_$rettype";
    if ($args{'Kurvenpunkte'}) {
	$cachefile .= "_kurvenp";
    }
    if ($use_cache && $rettype =~ /^hash/) {
	my $hashref = Strassen::Util::get_from_cache($cachefile,
						     [$self->{File}]);
	if (defined $hashref) {
	    warn "Using cache for $cachefile\n" if $VERBOSE;
	    return %$hashref;
	}
    }
    # RetType ...pos: Positionen statt Straennamen speichern
    my $store_pos = ($rettype =~ /pos$/);
    my %crossings;
    my %crossing_name;
    $self->init();
    while(1) {
	my $ret = $self->next();
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;
	my $store = ($store_pos ? $self->pos : $ret->[0]);
	foreach my $xy (@kreuzungen) {
	    $crossings{$xy}++;
	  TEST: {
		my $test;
		foreach $test (@{$crossing_name{$xy}}) {
		    last TEST if ($test eq $store);
		}
		push @{$crossing_name{$xy}}, $store;
	    }
	}
    }
    if ($rettype =~ /^hash/) { # Rckgabewert: "x,y" => [name1,name2 ...]
	my @to_del;
	while(my($k, $v) = each %crossings) {
	    if ($v < $min_strassen) {
		push @to_del, $k;
	    } else {
		$crossings{$k} = $crossing_name{$k};
	    }
	}
	foreach (@to_del) {
	    delete $crossings{$_};
	}
	if ($use_cache &&
	    Strassen::Util::write_cache(\%crossings, $cachefile)) {
	    warn "Wrote cache ($cachefile)\n" if $VERBOSE;
	}
	%crossings;
    } else { # Rckgabewert: [x, y, "name1/name2/..."]
	my @crossings;
	while(my($k, $v) = each %crossings) {
	    if ($v >= $min_strassen) {
		my($x, $y) = split(/,/, $k);
		push @crossings, [$x, $y, join("/", @{$crossing_name{$k}})];
	    }
	}
	@crossings;
    }
}

# Gibt die Positionsnummern aller Straen aus $str_ref als Liste aus.
# $str_ref ist eine Liste von [Straenname, Bezirk]-Elementen
# Falls eine Strae durch mehrere Bezirke fhrt, wird nur _eine_ Position
# zurckgegeben.
### AutoLoad Sub
sub union {
    my($self, $str_ref, %args) = @_;

    my $uniq = !$args{Nouniq};

    my %str;
    foreach (@$str_ref) {
	$str{$_->[0]}->{$_->[1]}++;
    }

    my %res;
    $self->init;
    my $last;
    while(1) {
	my $ret = $self->next;
	last if !@{$ret->[1]};
	my $name = $ret->[0];
	if ($uniq) {
	    if (defined $last && $last eq $name) {
		next;
	    } else {
		$last = $name;
	    }
	}
	my @bez;
	if ($name =~ /^(.*)\s+\((.*)\)$/) {
	    $name = $1;
	    @bez = split(/,\s*/, $2);
	}
	if (exists $str{$name}) {
	    if (@bez) {
		foreach my $bez (@bez) {
		    if (exists $str{$name}->{$bez}) {
			$res{$self->pos}++;
			last;
		    }
		}
	    } else {
		$res{$self->pos}++;
	    }
	}
    }
    keys %res;
}

### AutoLoad Sub
sub agrep {
    my($self, $pattern, %arg) = @_;
    agrep_file($pattern, $self->{File}, %arg);
}

# XXX make gzip-aware
# %arg:
# NoDot: keine Ausgabe von "...", wenn zu viele Matches existieren
# ErrorDef: Angabe der Reihenfolge (match begin, match errors)
# Agrep: maximale Anzahl von erlaubten Fehlern
### AutoLoad Sub
sub agrep_file {
    my($pattern, $file, %arg) = @_;

    my @paths;
    my @files;
    if (ref $file eq 'ARRAY') {
	@files = @$file;
    } else {
	push @files, $file;
    }

    foreach my $file (@files) {
	my $path;
	if (-r $file) {
	    $path = $file;
	} else {
	    foreach (@datadirs) {
		if (-r "$_/$file") {
		    $path = "$_/$file";
		    last;
		}
	    }
	}
	if (!defined $path) {
	    warn "File $file not found in @datadirs.\n";
	    return undef;
	}
	push @paths, $path;
    }

    my $grep_type;
    my @data;
    if (!$OLD_AGREP && is_in_path('agrep')) {
	$grep_type = 'agrep';
	$pattern =~ s/(.)/\\$1/g;
    } else {
	foreach my $path (@paths) {
	    open(F, $path) or die "Can't open $path: $!";
	    my @file_data;
	    chomp(@file_data = <F>);
	    push @data, @file_data;
	    close F;
	}
	eval { local $SIG{'__DIE__'};
	       require String::Approx;
	       String::Approx->VERSION(2.7);
	   };
	if (!$@) {
	    $grep_type = 'approx';
	} else {
	    $grep_type = 'perl';
	}
    }
    my @def;
    if ($arg{ErrorDef}) {
	@def = @{$arg{ErrorDef}};
    } else {
	@def = ([1, 0],
		[1, 1],
		[1, 2],
		[0, 0],
		[0, 1],
		[0, 2],
		[0, 3],
	       );
    }
    for my $def (@def) {
	my($begin, $err, @extra) = @$def;
	next if (exists $arg{Agrep} && $err > $arg{Agrep});
	my @this_res;
	my $grep_pattern = $pattern;
	if (grep($_ eq 'strasse', @extra)) {
            next if ($grep_pattern !~ s/(s)trae$/$1tr./i);
	}
	if ($grep_type eq 'agrep') {
	    my @args = '-i';
	    $grep_pattern = ($begin ? "^$grep_pattern" : $grep_pattern);
	    if ($err > 0) { push @args, "-$err" }
	    open(AGREP, "-|") or
	      exec 'agrep', @args, $grep_pattern, @paths or
		die "Can't exec program: $!";
	    chomp(@this_res = <AGREP>);
	    close AGREP;
	} elsif ($grep_type eq 'approx' && $err) {
	    next if $begin || $err > 2; # Bug bei $err == 3
	    $grep_pattern =~ s/[()]/./g; # String::Approx-Bug?
	    @this_res = String::Approx::amatch
	      ($grep_pattern, ['i', $err], @data);
	} else { # weder agrep noch String::Approx
	    $grep_pattern = ($begin ? "^$grep_pattern" : $grep_pattern);
	    if ($err == 0) { 
		@this_res = grep(/\Q$grep_pattern\E/i, @data);
	    } elsif ($err == 1) { # metacharacter erlauben
		@this_res = grep(/$grep_pattern/i, @data);
	    } else {
		next;
	    }
	}
	if (@this_res == 1) {
	    return parse($this_res[0])->[0];
	} elsif (@this_res) {
	    my(@res1, @res2, @res3);
	    my $i = 0;
	    my $last_name;
	    foreach (@this_res) {
		$i++;
		my $name = parse($_)->[0];
		if (defined $last_name && $last_name eq $name) {
		    next;
		} else {
		    $last_name = $name;
		}
		if ($name eq $pattern) {
		    return $name;
		} elsif ($name =~ /^\Q$pattern\E/i) {
		    push(@res1, $name);
		} elsif ($i < 20) {
		    push(@res2, $name);
		} elsif ($i == 20) {
		    push(@res3, "...") unless $arg{NoDot};
		}
	    }
	    @res1 = sort @res1;
	    @res2 = sort @res2;
	    return @res1, @res2, @res3;
	}
    }
    ();
}

# Sucht Strae anhand des Bezirkes.
### AutoLoad Sub
sub choose_street {
    my($str, $strasse, $bezirk, %args) = @_;
    my @pos;
    $str->init;
    while(1) {
	my $ret = $str->next;
	last if !@{$ret->[1]};
	my $check_strasse = $ret->[0];
	if ($check_strasse =~ /^$strasse/) {
	    my %bez;
	    if ($check_strasse =~ /(.*)\s+\((.*)\)/) {
		$check_strasse = $1;
		foreach (split(/\s*,\s*/, $2)) {
		    $bez{$_}++;
		}
		if (exists $bez{$bezirk}) {
		    if (wantarray) {
			push @pos, $str->pos;
		    } else {
			return $str->pos;
		    }
		}
	    } elsif ($check_strasse eq $strasse) {
		if (wantarray) {
		    push @pos, $str->pos;
		} else {
		    return $str->pos;
		}
	    }
	}
    }
    if (wantarray) {
	@pos;
    } else {
	undef;
    }
}

sub copy_orig {
    my $self = shift;
    if (! -d $Strassen::Util::tmpdir) {
	warn "$Strassen::Util::tmpdir does not exist" if $VERBOSE;
	return;
    }
    my $origdir = "$Strassen::Util::tmpdir/orig";
    if (! -d $origdir) {
	mkdir $origdir, 0700;
	if (! -d $origdir) {
	    warn "Can't create $origdir: $!" if $VERBOSE;
	    return;
	}
    }
    if (!defined $self->{File}) {
	warn "File not defined" if $VERBOSE;
	return;
    }
    if (ref $self->{File} ne 'ARRAY') {
	if (!-f $self->{File}) {
	    warn "<$self->{File}> does not exist" if $VERBOSE;
	    return;
	}
    } else {
	foreach (@{$self->{File}}) {
	    if (!-f $_) {
		warn "<$_> does not exist" if $VERBOSE;
		return;
	    }
	}
    }
    require File::Basename;
    my $dest;
    if (ref $self->{File} ne 'ARRAY') {
	$dest = "$origdir/" . File::Basename::basename($self->{File});
    } else {
	$dest = "$origdir/" . File::Basename::basename($self->{File}->[0] . "-multi");
    }
    if ($self->write($dest)) {
	$self->{OrigFile} = $dest;
	1;
    } else {
	delete $self->{OrigFile};
	0;
    }
}

# Erzeugt die Differenz aus dem aktuellen Strassen-Objekt und der
# letzten Version, die (evtl.) in $origdir abgelegt ist.
# Rckgabe: (Strassen-Objekt mit neuen Straen, zu lschenden Indices)
# Argumente: -clonefile => 1: das File-Argument wird in das neue Objekt
#            kopiert
### AutoLoad Sub
sub diff_orig {
    my($self, %args) = @_;
    require File::Basename;
    my $origdir = "$Strassen::Util::tmpdir/orig";
    # XXX nicht first_file, sondern aus allen Dateinamen (evtl. verkrzt)
    # konstruieren
    my $first_file = (ref $self->{File} eq 'ARRAY'
		      ? $self->{File}->[0] . "-multi"
		      : $self->{File});
    if (!defined $self->{OrigFile}) {
	$self->{OrigFile} =
	  "$origdir/" . File::Basename::basename($first_file);
    }
    if (! -f $self->{OrigFile}) {
	warn "<$self->{OrigFile}> does not exist" if $VERBOSE;
	delete $self->{OrigFile};
	return;
    }
    if (!is_in_path("diff")) {
	warn "diff not found in path" if $VERBOSE;
	return;
    }

    my $dest = "$origdir/" . File::Basename::basename($first_file) . ".new";
    return unless $self->write($dest);

    my $curr_line = 1;
    my(@del, @add);
    open(DIFF, "diff -u $self->{OrigFile} $dest |");
    scalar <DIFF>; scalar <DIFF>; # overread header
    while(<DIFF>) {
	chomp;
	if (/^\@\@\s*-(\d+)/) {
	    $curr_line = $1;
	} elsif (/^\+(.*)/) {
	    push @add, "$1\n";
	} elsif (/^-/) {
	    push @del, $curr_line-1; # warum -1?
	    $curr_line++;
	} elsif (!/^[ \\]/) {
	    warn "Unknown diff line: $_";
	} else {
	    $curr_line++;
	}
    }
    close DIFF;

    unlink $dest;
    my $new_s = new_from_data Strassen @add;
    if ($args{-clonefile}) {
	$new_s->{File} = $self->{File};
    }
    ($new_s, \@del);
}

# Schneidet den Teil in Klammern weg.
# Wird fr Bezirke, aber auch bei Bahnhfen (z.B. (U1)) verwendet.
### AutoLoad Sub
sub strip_bezirk {
    my $str = shift;
    if ($str !~ /^\s*\(/) {
	$str =~ s/\s*\(.*\)\s*$//;
    }
    $str;
}

# Fr Orte: trennt den Namen vom Zusatz (z.B. ("Frankfurt", "Oder")
### AutoLoad Sub
sub split_ort {
    split /\|/, $_[0], 2;
}

# Create array reference from Data property:
# [[$name, $category, ["$x1,$y1", "$x2,$y2" ...]],
#  [$name2, ...]
# ]
# Warning: this method resets any init/next loop!
### AutoLoad Sub
sub as_array {
    my $self = shift;
    my $ret = [];
    $self->init;
    while(1) {
	my $r = $self->next;
	last if !@{$r->[1]};
	my $new_item = [$r->[0], $r->[2], $r->[1]];
	push @$ret, $new_item;
    }
    $ret;
}

# Create a reverse hash pointing from a point to a list of streets
# containing this point:
# { "$x1,$y1" => [$streetname1, $streetname2 ...], ... }
# Warning: this method resets any init/next loop!
### AutoLoad Sub
sub as_reverse_hash {
    my $self = shift;
    my $rev_hash = {};
    $self->init;
    while(1) {
	my $r = $self->next;
	last if !@{$r->[1]};
	foreach my $c (@{$r->[1]}) {
	    if (exists $rev_hash->{$c}) {
		push @{ $rev_hash->{$c} }, $r->[0];
	    } else {
		$rev_hash->{$c} = [$r->[0]];
	    }
	}
    }
    $rev_hash;
}

# warning: don't call this in a Strassen-loop (init, next ...)
### AutoLoad Sub
sub make_grid {
    my($self, %args) = @_;
    my $use_cache = $args{UseCache};
    my $use_exact = $args{Exact};
    my $cachefile = "grid" . ($use_exact ? "x" : "") . "_" . $self->id;
    $self->{GridWidth}  = (defined $args{GridWidth} 
			   ? $args{GridWidth} : 1000);
    $self->{GridHeight} = (defined $args{GridHeight} 
			   ? $args{GridHeight} : $self->{GridWidth});
    if ($use_cache) {
	my $hashref = Strassen::Util::get_from_cache($cachefile,
						     [$self->file]);
	if (defined $hashref) {
	    warn "Using grid cache for $cachefile\n" if $VERBOSE;
	    $self->{Grid} = $hashref;
	    return;
	}
    }
    $self->{Grid} = {};
    my $grid_build = ($use_exact
		      ? $self->_make_grid_exact
		      : $self->_make_grid_fast);
    while(my($g, $v) = each %$grid_build) {
	$self->{Grid}{$g} = [keys %$v];
    }
    if ($use_cache &&
	Strassen::Util::write_cache($self->{Grid}, $cachefile)) {
	warn "Wrote cache ($cachefile)\n" if $VERBOSE;
    }
}

### AutoLoad Sub
sub _make_grid_fast {
    my $self = shift;
    my %grid_build;
    $self->init;
    my $strpos = 0;
    while(1) {
	my $r = $self->next;
	last if !@{$r->[1]};
	foreach my $c (@{$r->[1]}) {
	    $grid_build{join(",",$self->grid(split(/,/, $c)))}->{$strpos}++;
	}
	$strpos++;
    }
    \%grid_build;
}

### AutoLoad Sub
sub _make_grid_exact {
    my $self = shift;
    my %grid_build;
    $self->init;
    my $strpos = 0;
    while(1) {
	my $r = $self->next;
	last if !@{$r->[1]};
	my %check_environ;
	for(my $i = 0; $i < $#{$r->[1]}; $i++) {
	    my($x1, $y1) = split(',', $r->[1][$i]);
	    my($x2, $y2) = split(',', $r->[1][$i+1]);
	    $check_environ{$self->grid($x1,$y1)}++;
	    $check_environ{$self->grid($x2,$y2)}++;
	}
	# XXX to b continued...
	# XXX fr die umgebung der gitter berprfen, ob die strecke in
	# diesen gitter liegt. berprfte gitter auch check_environ
	# rausnehmen, bei erfolg deren umgebung in check_environ
	# reinnehmen und weiter untersuchen, bis keine gitter mehr
	# brig sind. zum berprfen den code aus miscsrc/gridvectortest.pl
	# verwenden.
	# XXX
#	    $grid_build{join(",",$self->grid(split(/,/, $c)))}->{$strpos}++;
	$strpos++;
    }
    \%grid_build;
}

### AutoLoad Sub
sub grid {
    my($self, $x, $y) = @_;
    (int($x/$self->{GridWidth}), int($y/$self->{GridHeight}));
}

# Gibt eine Liste mit den neuen Gitterquadranten fr die
# Koordinateneckpunte aus. Mit dem Argument KnownGrids knnen bereits
# bekannte Quadranten aus der Liste ausgeschlossen werden.
### AutoLoad Sub
sub get_new_grids {
    my($self, $x1, $y1, $x2, $y2, %args) = @_;
    if ($x2 < $x1) { ($x2, $x1) = ($x1, $x2) }
    if ($y2 < $y1) { ($y2, $y1) = ($y1, $y2) }
    my $known_grids = {};
    if (exists $args{'KnownGrids'} and ref $args{'KnownGrids'} eq 'HASH') {
	$known_grids = $args{'KnownGrids'};
    }
    my @new_grids;
    my $x = int($x1/$self->{GridWidth})*$self->{GridWidth};#XXX neg Koords?
    my $ybeg = int($y1/$self->{GridHeight})*$self->{GridHeight};
    while ($x < $x2) {
	my $y = $ybeg;
	while ($y < $y2) {
	    my($gx, $gy) = $self->grid($x, $y);
	    if (!$known_grids->{"$gx,$gy"}) {
		push @new_grids, "$gx,$gy";
		$known_grids->{"$gx,$gy"}++;
	    }
	    $y += $self->{GridHeight};
	}
	$x += $self->{GridWidth};
    }
    @new_grids;
}

main::mymstat("vor Strassen::Fast");

1;

######################################################################

package Strassen::Fast;
#use AutoLoader 'AUTOLOAD';
use vars qw(@ISA);
@ISA = qw(Strassen);

sub new {
    my($class, $filename) = @_;
    require Storable;
    my $self = retrieve $filename;
    bless $self, $class;
}

main::mymstat("vor Strasse");

1;

######################################################################

package Strasse;
#use AutoLoader 'AUTOLOAD';

sub new {
    my($class, $obj) = @_;
    bless $obj, $class;
}
sub name        { $_[0]->[0] }
sub coords      { $_[0]->[1] }
sub coords_list { @{$_[0]->[1]} }
sub category    { $_[0]->[2] }
sub is_empty    { @{$_[0]->[1]} == 0 }
sub coord_as_string { $_[0]->coords->[$_[1]] }
sub coord_as_list {
    my($self, $i) = @_;
    my $s = $self->coords->[$i];
    Strassen::to_koord1($s); # XXX abhngig von Strassen!
}

# gibt TRUE zurck, wenn ein Teil der Strae im angegebenen Rechteck ist
# (keine perfekte Lsung)
### AutoLoad Sub
sub is_in {
    my($self, $x1, $y1, $x2, $y2) = @_;
    my $len = scalar $self->coords_list;
    for(my $i=0; $i<$len; $i++) {
	my($x, $y) = @{$self->coord_as_list($i)};
	if ($x >= $x1 && $x <= $x2 &&
	    $y >= $y1 && $y <= $y2) {
	    return 1;
	}
    }
    0;
}

# statische Methode
### AutoLoad Sub
sub de_artikel {
    my($strasse) = @_;
    if ($strasse =~ /(str\.|strae\b|allee\b|chaussee\b|promenade\b)/i) {
	"in die";
    } elsif ($strasse =~ /(park\b|garten\b|ring\b)/i) {
	"in den";
    } elsif ($strasse =~ /(damm\b|weg\b|steig\b)/i) {
	"in den";
    } elsif ($strasse =~ /(platz\b)/i) {
	"auf den";
    } elsif ($strasse =~ /(ufer\b|gestell\b)/i) {
	"in das";
    } elsif ($strasse =~ /(brcke\b)/i) {
	"auf die";
    } elsif ($strasse =~ /(\balt\b)/i) {
	"in die Strae";
    } elsif ($strasse =~ /(\btor\b)/i) {
	"in das";
    } elsif ($strasse =~ /(\bstern\b)/i) { # mglichst am Ende
	"auf den";
    } else {
	"=>";
    }
}

# Den Straennamen so weit wie mglich abkrzen...
# Verschiedene Level (0 bis 3) sind mglich
sub short {
    my($strname, $level) = @_;
    if ($level > 0) {
	$strname =~ s/(s)trae/$1tr./i;
	$strname =~ s/(p)latz/$1l./i;
    }
    if ($level > 2) {
	$strname =~ s/str\.//;
	$strname =~ s/^Str\./Str/;
	$strname =~ s/([ \-]S)tr\.//;
	$strname =~ s/(p)l\./$1l/i;
	$strname =~ s/damm/d./;
	$strname =~ s/brcke/br./;
    } elsif ($level > 1) {
	$strname =~ s/(s)tr\./$1tr/i;
	$strname =~ s/(p)l\./$1l/i;
    }
    $strname;
}

main::mymstat("vor Kreuzungen");

1;

######################################################################

package Kreuzungen;
#use AutoLoader 'AUTOLOAD';

# Argumente: Hash ist der Rckgabewert von Strassen::all_crossings (bei
# Verwendung von "hash" oder "hashpos" als RetType). Array wird noch nicht
# verwendet.
# Falls "Strassen" als Argument bergeben wurde, braucht "Hash" nicht
# angegeben zu sein, da es automatisch per all_crossings erstellt wird.
sub new {
    my($class, %args) = @_;
    my $self = {};
    if (!exists $args{Hash}) {
	die "Missing arg for new Kreuzungen: either Hash or Strassen"
	  if (!exists $args{Strassen});
	my(%kreuzungen) = $args{Strassen}->all_crossings(RetType => 'hash',
							 UseCache => 1);
	$args{Hash} = \%kreuzungen;
    }
    $self->{Hash}  = $args{Hash};
    $self->{Array} = $args{Array};
    bless $self, $class;
}

# Nimmt an. dass alle Punkte aus der Strassen-Datei Kreuzungen (z.B. Orte)
# darstellen.
### AutoLoad Sub
sub new_from_strassen {
    my($class, %args) = @_;
    my $self = {};
    $self->{Hash} = {};
    my $str = $args{Strassen};
    if (!$str) {
	die "Missing arg for new_from_strassen: Strassen";
    }
    $str->init;
    while(1) {
	my $ret = $str->next;
	last if !@{ $ret->[1] };
	foreach my $p (@{ $ret->[1] }) {
	    $self->{Hash}{$p} = $ret->[0];
	}
    }
    bless $self, $class;
}

### Autoload Sub
sub add {
    my($self, %args) = @_;
    if (!exists $args{Hash}) {
	die "Missing argument Hash";
    }
    my $h = $self->{Hash};
    while(my($k,$v) = each %{ $args{Hash} }) {
	$h->{$k} = $v;
    }
}

### AutoLoad Sub
sub get { $_[0]->{Hash}{$_[1]} }

### AutoLoad Sub
sub make_grid {
    my($self, $grid_width) = @_;
    $self->{GridWidth}  = (defined $grid_width ? $grid_width : 1000);
    $self->{Grid} = {};
    while(defined(my $k = each %{$self->{Hash}})) {
	my $grid = join(",", $self->grid(split(/,/, $k)));
	push @{$self->{Grid}{$grid}}, $k;
    }
}

### AutoLoad Sub
sub grid {
    my($self, $x, $y) = @_;
    (int($x/$self->{GridWidth}), int($y/$self->{GridWidth}));
}

# Argumente:
#    $x, $y
#    IncludeDistance: Rckgabewert ist ein Feld mit den Elementen ["x,y", dist]
#    Grids (default 1): Anzahl der Grids, die vom Mittelpunkt abweichen drfen
# Ansonsten wird ein Feld mit den Elementen "x,y" zurckgegeben. In beiden
# Fllen ist das Feld nach der Entfernung von $x und $y sortiert (krzeste
# zuerst).
### AutoLoad Sub
sub nearest {
    my($self, $x, $y, %args) = @_;
    $self->make_grid if (!$self->{Grid});
    my $xy = join(",", $x, $y);
    my $grids = $args{Grids} || 1;
#   my $distance = $args{Distance};
# XXX distance wird noch ignoriert
#    if (!defined $distance) { $distance = 1000 }
    my($gridx, $gridy) = $self->grid($x, $y);
    my @res;
    for my $xx (-$grids .. $grids) {
	for my $yy (-$grids .. $grids) {
	    my $s = join(",", $gridx+$xx, $gridy+$yy);
	    if (defined $self->{Grid}{$s}) {
		push @res, @{$self->{Grid}{$s}};
	    }
	}
    }
    if ($args{IncludeDistance}) {
	foreach (@res) {
	    my $dist = Strassen::Util::strecke_s($_, $xy);
	    $_ = [$_, $dist];
	}
	sort { $a->[1] <=> $b->[1] } @res;
    } else {
	sort { Strassen::Util::strecke_s($a, $xy) 
		 <=> Strassen::Util::strecke_s($b, $xy)} @res;
    }
}

# incrementally try nearest with Grids => 1 to Grids => ...
sub nearest_loop {
    my($self, $x, $y, %args) = @_;
    my @res;
    for my $grids (1 .. 5) {
	@res = $self->nearest($x, $y, %args, Grids => $grids);
	return @res if (@res);
    }
    ();
}

# wie nearest, nur wird hier "x,y" als ein Argument bergeben
# XXX nearest_loop_coord fehlt
### AutoLoad Sub
sub nearest_coord {
    my($self, $xy, %args) = @_;
    my($x, $y) = split(/,/, $xy);
    $self->nearest($x, $y, %args);
}

main::mymstat("vor MultiBezStr");

1;

######################################################################

package MultiBezStr;
#use AutoLoader 'AUTOLOAD';

# make gzip-aware
sub new {
    my $class = shift;
    require MyFile;
    if (MyFile::openlist(*MULTI,
			 map { "$_/multi_bez_str" } @Strassen::datadirs)) {
	my $self = {};
	while(<MULTI>) {
	    chomp;
	    my($str, $rest) = split(/\t/);
	    my(@bez) = split(/,/, $rest);
	    $self->{$str} = \@bez;
	}
	bless $self, $class;
    } else {
	undef;
    }
}

sub exists {
    my($self, $str) = @_;
    exists $self->{$str};
}

sub bezirke {
    my($self, $str) = @_;
    if ($self->exists($str)) {
	@{$self->{$str}}
    } else {
	();
    }
}

sub hash { %{$_[0]} }

main::mymstat("vor MultiStrassen");

1;

######################################################################

package MultiStrassen;
#use AutoLoader 'AUTOLOAD';
use vars qw(@ISA);
@ISA = qw(Strassen);

sub new {
    my($class, @obj) = @_;

    if (@obj == 1 && $obj[0]->isa('Strassen')) {
	return $obj[0];
    }

    my $self = {};

    $self->{Data} = [];
    $self->{File} = [];
    $self->{Pos}  = 0;
#XXX DEL:    $self->{Koord} = undef; # XXX
    $self->{SourcePos} = {};

    for (@obj) {
	if (!$_->isa('Strassen')) {
	    die "Object $_ not a Strassen object";
	}
	if (defined $_->{File}) {
	    push @{$self->{File}}, $_->{File};
	}
	push @{$self->{FirstIndex}}, $#{$self->{Data}}+1;
	push @{$self->{Data}}, @{$_->{Data}};
	push @{$self->{SubObj}}, $_; # XXX Performance-Hit?
    }
    bless $self, $class;
}

# Ausgabe der Source-Files
sub file { @{shift->{File}} }

sub id {
    my $self = shift;
    require File::Basename;
    join("_", sort map { File::Basename::basename($_) } $self->file);
}

main::mymstat("vor StrassenNetz");

1;

######################################################################

package StrassenNetz;
#use AutoLoader 'AUTOLOAD';
use Route;
use BBBikeUtil qw(rad2deg schnittwinkel m2km);
use BBBikeCalc;
use vars qw($VERBOSE $acos $data_format);
BEGIN {
    eval {
	require POSIX;
	$acos = \&POSIX::acos; # C, daher wahrscheinlich etwas schneller
    };
    if ($@) {
	require Math::Trig;
	$acos = \&Math::Trig::acos;
    }
}

$VERBOSE     = 0;
$data_format = 1;

sub new {
    my($class, $strassen) = @_;
    die "argument 1 is not of type Strassen"
      if !$strassen->isa('Strassen');
    my $self = {};
    $self->{Strassen} = $strassen;
    bless $self, $class;
}

### AutoLoad Sub
sub new_from_server {
    my $class = shift;
    my $server_name = shift || 'bbb';
    # nachgucken, ob vielleicht str_server.pl luft
    my $net;
    eval {
	require IPC::ShareLite;
	require Storable;
	my %options = (
		       -key => '1211', # XXX get from var
		       -create => 'no',
		       -exclusive => 'no',
		       -destroy => 'no',
		      );
	my $share = IPC::ShareLite->new(%options) or die $!;
	warn "Shared memory anzapfen...\n" if ($VERBOSE);
	$net = Storable::thaw($share->fetch);
	use_data_format(1) if $net;
    };
    warn $@ if $@ && $VERBOSE;
    return $net if defined $net;

    eval {
	require IPC::Shareable;
	my %options = (
		       'key' => 'paint',
		       'create' => 'no',
		       'exclusive' => 'no',
		       'mode' => 0644,
		       'destroy' => 'no',
		      );
	warn "Shared memory anzapfen...\n" if ($VERBOSE);
	tie $net, 'IPC::Shareable', $server_name, \%options;
	#tie $net->{Net}, 'IPC::Shareable', $server_name."1", \%options;
	#tie $net->{Net2Name}, 'IPC::Shareable', $server_name."2", \%options;
	use_data_format(1) if $net;
    };
    warn $@ if $@ && $VERBOSE;
    $net;
}

# markiert die angegebenen Objekte als Quell-Objekte fr dieses StrassenNetz
### AutoLoad Sub
sub source {
    my($self, @source) = @_;
    push(@{$self->{Source}}, @source);
}

# markiert die angegebenen Straen-Abkrzungen als Quell-Objekte fr
# dieses StrassenNetz
### AutoLoad Sub
sub source_abk {
    my($self, @source_abk) = @_;
    push(@{$self->{SourceAbk}}, @source_abk);
}

### AutoLoad Sub
sub is_source {
    my($self, $source) = @_;
    foreach (@{$self->{Source}}) {
	return 1 if $_ eq $source;
    }
    0;
}

# gibt die zugehrigen Quelldateien aus
### AutoLoad Sub
sub sourcefiles {
    my $self = shift;
    if (exists $self->{Source} && @{$self->{Source}}) {
	my @src;
	foreach (@{$self->{Source}}) {
	    push @src, $_->{File};
	}
	@src;
    } else {
	$self->{Strassen}->file;
    }
}

### AutoLoad Sub
sub statistics {
    my $self = shift;
    my $msg = '';
    if ($self->{Strassen}) {
	$msg .= "Anzahl der Straen:    " . $self->{Strassen}->count . "\n";
    }

    if ($self->{Net2Name}) {
	my $count = 0;
	while(my($k,$v) = each %{$self->{Net2Name}}) {
	    $count += scalar keys %$v;
	}
	$msg .= "Anzahl der Kanten:     " . $count . "\n";

	my $nodes = scalar keys %{$self->{Net2Name}};
	$msg .= "Anzahl der Knoten:     " . $nodes . "\n";

	$msg .= "node branching factor: " . 
	    sprintf("%.1f", $count/$nodes) . "\n";
    }
    $msg;
}

### AutoLoad Sub
sub make_net_slow_1 {
    my($self, %args) = @_;
    my($cachefile);

    my $cacheable = $args{UseCache} && $Strassen::Util::cacheable;
    if ($cacheable) {
	require File::Basename;
	my(@src) = $self->sourcefiles;
	$cachefile = join("_", map { File::Basename::basename($_) } @src);
	my $net = Strassen::Util::get_from_cache("net_1_$cachefile", \@src);
	my $net2name = Strassen::Util::get_from_cache("net2name_1_$cachefile",
						      \@src);
	if (defined $net && defined $net2name) {
	    $self->{Net} = $net;
	    $self->{Net2Name} = $net2name;
	    if ($VERBOSE) {
		warn "Using cache for $cachefile\n";
	    }
	    return;
	}
    }

    if ($VERBOSE) {
	warn "Using slow version of make_net\n";
    }
    $self->{Net}         = {}; # Verbindungsnetz
    $self->{Net2Name}    = {}; # Zuordnung Strecke => Straenname
    $self->{Wegfuehrung} = {}; # unerlaubte Wegfhrung
    my $net      = $self->{Net};
    my $net2name = $self->{Net2Name};
    my $strassen = $self->{Strassen};
    $strassen->init;
    while(1) {
	my $ret = $strassen->next;
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;
	my @kreuz_coord = @{Strassen::to_koord(\@kreuzungen)};
	my $i;
	for($i = 0; $i < $#kreuzungen; $i++) {
	    # Integer reicht vollkommen aus, da die Angaben sowieso in m sind
	    my $entf = int(Strassen::Util::strecke($kreuz_coord[$i],
						   $kreuz_coord[$i+1]));
 	    $net->{$kreuzungen[$i]}{$kreuzungen[$i+1]} = $entf;
 	    $net->{$kreuzungen[$i+1]}{$kreuzungen[$i]} = $entf;
 	    $net2name->{$kreuzungen[$i]}{$kreuzungen[$i+1]} = $strassen->pos;
	}
    }

    if ($cacheable) {
	Strassen::Util::write_cache($net, "net_1_$cachefile");
	Strassen::Util::write_cache($net2name, "net2name_1_$cachefile");
	if ($VERBOSE) {
	    warn "Wrote cache ($cachefile)\n";
	}
    }

    $self->{UseMLDBM} = 0;
}

### AutoLoad Sub
sub make_net_slow_2 {
    my($self, %args) = @_;
    my($cachefile);

    my $cacheable = $args{UseCache} && $Strassen::Util::cacheable;
    if ($cacheable) {
	require File::Basename;
	my(@src) = $self->sourcefiles;
	$cachefile = join("_", map { File::Basename::basename($_) } @src);
	my $coord2index = Strassen::Util::get_from_cache("coord2index_2_$cachefile", \@src);
	my $index2coord = Strassen::Util::get_from_cache("index2coord_2_$cachefile", \@src);
	my $net = Strassen::Util::get_from_cache("net_2_$cachefile", \@src);
	my $index2pos = Strassen::Util::get_from_cache("index2pos_2_$cachefile",
						       \@src);
	if (defined $coord2index &&
	    defined $index2coord &&
	    defined $net         &&
	    defined $index2pos) {
	    $self->{Coord2Index} = $coord2index;
	    $self->{Index2Coord} = $index2coord;
	    $self->{Net}         = $net;
	    $self->{Index2Pos}   = $index2pos;
	    if ($VERBOSE) {
		warn "Using cache for $cachefile\n";
	    }
	    return;
	}
    }

    if ($VERBOSE) {
	warn "Using new version of make_net\n";
    }
    $self->{Net}         = []; # Verbindungsnetz
    $self->{Index2Pos}   = []; # Zuordnung Index-Paar => Pos im Straenfile
    $self->{Coord2Index} = {}; # Zuordnung Koordinate => Index
    $self->{Index2Coord} = []; # Zuordnung Index => Koordinate
    $self->{Wegfuehrung} = {}; # unerlaubte Wegfhrung

    my $net         = $self->{Net};
    my $index2pos   = $self->{Index2Pos};
    my $coord2index = $self->{Coord2Index};
    my $index2coord = $self->{Index2Coord};
    my $strassen    = $self->{Strassen};
    $strassen->init();
    my $pos = 0;
local($^W)=0; #XXX
    while(1) {
	my $ret = $strassen->next();
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;

	my @kreuz_coord = @{Strassen::to_koord(\@kreuzungen)};
	my @k_i;
	foreach my $cp (@kreuz_coord) {
	    my $c = pack("l2", @$cp);
	    if (!exists $coord2index->{$c}) {
		$coord2index->{$c} = pack("l", $pos);
		$index2coord->[$pos] = $c;
		$pos++;
	    }
	    push @k_i, $coord2index->{$c};
	}

	for (my $i; $i < $#k_i; $i++) {
	    my $entf = pack("l",
			    int(Strassen::Util::strecke($kreuz_coord[$i],
							$kreuz_coord[$i+1])));
	    my $k_i_u  = unpack("l", $k_i[$i]);
	    my $k_i1_u = unpack("l", $k_i[$i+1]);
	    $net->[$k_i_u]  .= $k_i[$i+1] . $entf;
	    $net->[$k_i1_u] .= $k_i[$i]   . $entf;
	    $index2pos->[$k_i_u]  .= $k_i[$i+1] . pack("l", $strassen->pos);
	    $index2pos->[$k_i1_u] .= $k_i[$i]   . pack("l", $strassen->pos);
	}
    }

    if ($cacheable) {
	Strassen::Util::write_cache($coord2index, "coord2index_2_$cachefile");
	Strassen::Util::write_cache($index2coord, "index2coord_2_$cachefile");
	Strassen::Util::write_cache($net, "net_2_$cachefile");
	Strassen::Util::write_cache($index2pos, "index2pos_2_$cachefile");
	if ($VERBOSE) {
	    warn "Wrote cache ($cachefile)\n";
	}
    }

    $self->{UseMLDBM} = 0;
}

*make_net = \&make_net_slow_1;

### AutoLoad Sub
sub make_sperre_1 {
    my($self, $sperre_file, %args) = @_;

    my %sperre_type;
    if (exists $args{Type}) {
	foreach (@{$args{Type}}) {
	    if ($_ eq 'einbahn') {
		$sperre_type{1} = 1;
	    } elsif ($_ eq 'sperre') {
		$sperre_type{2} = 1;
	    } elsif ($_ eq 'tragen') {
		$sperre_type{0} = 1;
	    } elsif ($_ eq 'wegfuehrung') {
		$sperre_type{3} = 1;
	    } else {
		$sperre_type{$_} = 1;
	    }
	}
    } else {
	%sperre_type = (2 => 1, 1 => 1); # Standard: einbahn und sperre
    }

    my $sperre_obj = new Strassen $sperre_file;
    my $net      = $self->{Net};
    $sperre_obj->init;
    while(1) {
	my $ret = $sperre_obj->next;
	last if !@{$ret->[1]};
	my $category = $ret->[2];
	next if !exists $sperre_type{$category};
	if ($category eq "3") {
	    # Aufzeichnen der nicht erlaubten Wegfhrung
	    $self->{Wegfuehrung}{$ret->[1][-1]} = $ret->[1];
	} else {
	    my @kreuzungen = @{$ret->[1]};
	    if (@kreuzungen == 1) {
		$self->del_net($kreuzungen[0]);
	    } else {
		my $i;
		for($i = 0; $i < $#kreuzungen; $i++) {
		    $self->del_net($kreuzungen[$i], $kreuzungen[$i+1],
				   substr($category, 0, 1));
		}
	    }
	}
    }
}

*make_sperre = \&make_sperre_1;

# Erstellt ein Netz mit der Kategorie als Value.
# Wenn obey_dir TRUE ist, wird zwischen Hin- und Rckrichtung unterschieden.
### AutoLoad Sub
sub make_net_cat {
    my($self, $obey_dir) = @_;
    # Caching ist (noch) nicht notwendig
    $self->{Net} = {};
    my $net      = $self->{Net};
    my $strassen = $self->{Strassen};
    # net2name ist (noch) nicht notwendig
    $strassen->init;
    while(1) {
	my $ret = $strassen->next;
	my @kreuzungen = @{$ret->[1]};
	last if @kreuzungen == 0;
	my $cat = $ret->[2];
	my @kreuz_coord = @{Strassen::to_koord(\@kreuzungen)};
	my $i;
	for($i = 0; $i < $#kreuzungen; $i++) {
	    $net->{$kreuzungen[$i]}{$kreuzungen[$i+1]}     = $cat;
	    if (!$obey_dir) {
		$net->{$kreuzungen[$i+1]}{$kreuzungen[$i]} = $cat;
	    }
	}
    }
}

# erstellt ein Netz mit der Steigung als Value
# Argumente:
#   sourcenet: bereits existierendes StrassenNetz-Objekt, das
#              als Vorlage dient
#   hoehe: Hash-Referenz mit den Hoehenangaben
#   -min => minimale_Steigung
### AutoLoad Sub
sub make_net_steigung {
    my($self, $sourcenet, $hoehe, %args) = @_;
    die "sourcenet must be StrassenNetz object"
      if !$sourcenet->isa('StrassenNetz');
    my $min_steigung = $args{-min} || 0.001; # 0.1% als minimale Steigung
    $self->{Net} = {};
    my $net = $self->{Net};
    while(my($p1,$v) = each %{$sourcenet->{Net}}) {
	while(my($p2) = each %$v) {
	    if (exists $hoehe->{$p1}) {
		if (exists $hoehe->{$p2}) {
		    my($x1,$y1) = split(/,/, $p1);
		    my($x2,$y2) = split(/,/, $p2);
		    my $strecke = Strassen::Util::strecke([$x1,$y1],
							  [$x2,$y2]);
		    my $hoehendiff = $hoehe->{$p2}-$hoehe->{$p1};
		    $net->{$p1}{$p2} = int(($hoehendiff/$strecke)*1000)/1000;
		}
	    }
	}
    }
}

### AutoLoad Sub
sub reset {
    my $self = shift;
    $self->del_add_net;
}

use vars qw($MLDBM_SERIALIZER);
$MLDBM_SERIALIZER = 'Storable' unless defined $MLDBM_SERIALIZER;

# XXX Abspeichern der Wegfuehrung nicht getestet
### AutoLoad Sub
sub save_net_mldbm {
    my($self, $dir) = @_;
    if (!keys %{$self->{Net}}) {
	die "Net is empty";
    }
    require MLDBM;
    MLDBM->import('DB_File', $MLDBM_SERIALIZER);
    require Fcntl;
    require File::Basename;

    my(@src) = $self->sourcefiles;
    $dir = $Strassen::Util::cachedir unless $dir;
    my $file_net = "$dir/net_" .
	join("_", map { File::Basename::basename($_) } @src);
    my $file_net2name = "$dir/net2name_" .
	join("_", map { File::Basename::basename($_) } @src);
    my $file_wegfuehrung = "$dir/wegfuehrung_" .
	join("_", map { File::Basename::basename($_) } @src);

    my %mldbm_net;
    tie %mldbm_net, 'MLDBM', $file_net, &Fcntl::O_CREAT|&Fcntl::O_RDWR, 0640
      or die $!;
    while(my($k,$v) = each %{$self->{Net}}) {
	$mldbm_net{$k} = $v;
    }
    untie %mldbm_net;

    my %mldbm_net2name;
    tie
      %mldbm_net2name, 'MLDBM', $file_net2name,
      &Fcntl::O_CREAT|&Fcntl::O_RDWR, 0640
	or die $!;
    while(my($k,$v) = each %{$self->{Net2Name}}) {
	$mldbm_net2name{$k} = $v;
    }
    untie %mldbm_net2name;

    my %mldbm_wegfuehrung;
    tie
      %mldbm_wegfuehrung, 'MLDBM', $file_wegfuehrung,
      &Fcntl::O_CREAT|&Fcntl::O_RDWR, 0640
	or die $!;
    while(my($k,$v) = each %{$self->{Wegfuehrung}}) {
	$mldbm_wegfuehrung{$k} = $v;
    }
    untie %mldbm_wegfuehrung;
}

# Ein ernstes Problem ergibt sich bei der Verwendung von MLDBM:
# Da add_net neue Punkte zum Straennetz hinzufgt, wird der "Schrott"
# dadurch immer grer. Von Zeit zu Zeit sollte also mit make_net und
# save_net_mldbm ein neues, frisches Straennetz erzeugt werden.
### AutoLoad Sub
sub load_net_mldbm {
    my($self, $dir) = @_;
    require MLDBM;
    MLDBM->import('DB_File', $MLDBM_SERIALIZER);
    require Fcntl;
    require File::Basename;

    my(@src) = $self->sourcefiles;
    $dir = $Strassen::Util::cachedir unless $dir;
    my $file_net = "$dir/net_" .
	join("_", map { File::Basename::basename($_) } @src);
    my $file_net2name = "$dir/net2name_" .
	join("_", map { File::Basename::basename($_) } @src);
    my $file_wegfuehrung = "$dir/wegfuehrung_" .
	join("_", map { File::Basename::basename($_) } @src);

    my %mldbm_net;
    tie %mldbm_net, 'MLDBM', $file_net, &Fcntl::O_RDWR, 0640
      or die "Can't open $file_net: $!";
    $self->{Net} = \%mldbm_net;

    my %mldbm_net2name;
    tie
      %mldbm_net2name, 'MLDBM', $file_net2name, &Fcntl::O_RDWR, 0640
	or die "Can't open $file_net2name: $!";
    $self->{Net2Name} = \%mldbm_net2name;

    my %mldbm_wegfuehrung;
    tie
      %mldbm_wegfuehrung, 'MLDBM', $file_wegfuehrung, &Fcntl::O_RDWR, 0640
	or die "Can't open $file_wegfuehrung: $!";
    $self->{Wegfuehrung} = \%mldbm_wegfuehrung;

    $self->{UseMLDBM} = 1;
}

# Gibt die Straen-Positionsnummer fr das angegebene Koordinaten-Paar aus.
# Der zweite Rckgabewert (rueckwaerts) gibt an, ob die Reihenfolge from-to
# in der Datenbank umgedreht ist.
# Wenn $to nicht definiert ist, werden alle Straen-Positionsnummern, die
# von $from aus gehen, ausgegeben. In diesem Fall gibt es keinen
# "rueckwaerts"-Rckgabewert.
### AutoLoad Sub
sub net2name {
    my($net, $from, $to) = @_;
    if (!defined $to) {
	my(@to) = keys %{$net->{Net}{$from}};
	my @ret;
	foreach my $to (@to) {
	    push @ret, $net->net2name($from, $to);
	}
	@ret;
    } else {
	if (exists $net->{Net2Name}{$from}{$to}) {
	    ($net->{Net2Name}{$from}{$to}, 0);
	} elsif (exists $net->{Net2Name}{$to}{$from}) {
	    ($net->{Net2Name}{$to}{$from}, 1);
	} else {
	    warn "Can't find street from $from to $to"
	      if $VERBOSE;
	    undef;
	}
    }
}

# XXX mgliche Rckgabewerte:
# - die beste Pfadbeschreibung (+ Lnge etc.)
# - die besten Pfadbeschreibungen (+ Lnge etc.)
# - die beste Pfadbeschreibung (ohne Lnge etc.)
# - die besten Pfadbeschreibungen (ohne Lnge etc.)
# - die beste Route (als Objekt)
# - die besten Routen (als Objekt)
### AutoLoad Sub
sub build_search_code {
    my($self, %args) = @_;

    # Initialisierung ...
    # $has_penalty gibt an, ob die Suche nur ber die Entfernung geht oder
    # ob eine Penalty verwendet wird, die sich aus der Entfernung modifiziert
    # mit weiteren Parametern ergibt
    my $has_penalty = (exists $args{Ampeln}    ||
		       exists $args{Qualitaet} ||
		       exists $args{Strcat}    ||
		       exists $args{Radwege}   ||
		       exists $args{Steigung}  ||
		       exists $args{Abbiegen}
		      );
    my $has_ampeln  = exists $args{Ampeln};
    my $ampel_penalty;
    if ($has_ampeln) {
	$ampel_penalty = (exists $args{Ampeln}->{Penalty}
			  ? $args{Ampeln}->{Penalty}
			  : 100);
    }
    my $has_qualitaet = exists $args{Qualitaet};
    my $has_strcat    = exists $args{Strcat};
    my $has_radwege   = exists $args{Radwege};
    my $has_steigung  = exists $args{Steigung};
    my $has_abbiegen  = exists $args{Abbiegen} and exists $args{Ampeln};

    # Ausgabe einer Statistik
    my $do_stat       = $args{Stat} || 0;

    # Optionen zum ndern des Suchalgorithmus'
    my $cut_path_nr   = 10;
    my $pure_depth    = 0;
    my $backtracking  = 0;
    my $srt_algo      = 0; # use algorithm A*
    if (exists $args{Algorithm}) {
	if ($args{Algorithm} eq 'srt') {
	    $srt_algo = 1;
	}
    }
    if (exists $args{Tune}) {
	if (exists $args{Tune}->{CutPath}) {
	    $cut_path_nr = $args{Tune}->{CutPath};
	}
	if ($args{Tune}->{PureDepth}) {
	    $pure_depth = $args{Tune}->{PureDepth};
	}
	if ($args{Tune}->{Backtracking}) {
	    $backtracking = $args{Tune}->{Backtracking};
	}
    }

    # soll eine "visuelle" Suche vorgenommen werden
    my $do_visual     = exists $args{VisualSearch} ? 1 : 0;

    # Optimierung mit einem seen-Hash, damit bereits besuchte Knoten im
    # gleichen Pfad nicht nochmals berprft werden.
    my $seen_optimierung = 1;

    my $use_2 = 0;
    if (defined $args{Use2}) {
	$use_2 = $args{Use2};
    } elsif ($data_format == 2) {
	$use_2 = 1;
    }

    # XXX use_2 ist fr A* noch nicht implementiert
    if ($use_2) {
	$srt_algo = 1;
    }

    my $len_pen    = ($has_penalty ? 'pen' : 'len');

    # Aufschlag, damit Alternativ-Routen gefunden werden knnen
    my $aufschlag_code = '';
    if ($args{Aufschlag}) {
	$aufschlag_code = '*' . $args{Aufschlag};
    }
    # Code fr die Abfrage, ob der aktuelle Path das Ziel nicht mehr in einer
    # krzeren Lnge erreichen kann.
    my $skip_path_code = '
		    if (defined $visited{$next_node} and
			$next_node_'.$len_pen.' > $visited{$next_node}'
			  . $aufschlag_code . ') {
			next;
		    }
';
    my $skip_path_code2 = '
		    if (defined $visited{$to} and 
                        $virt_'.$len_pen.' > $visited{$to}'
			  . $aufschlag_code . ') {
			next;
		    }
';
    # Code fr die Abfrage, ob die Wegfhrung des aktuellen Pfades nicht
    # erlaubt ist
    # XXX Dieser Code sollte im SRT-Algorithmus verwendet werden, wird aber
    # nicht...
    my $skip_path_code3 = '
		    if ($wegfuehrung and
                        exists $wegfuehrung->{$next_node}) {
                      CHECK_WEGFUEHRUNG: {
                          my($wegfuehrung) = $wegfuehrung->{$next_node};
                          for(my $i=0; $i<$#$wegfuehrung; $i++) {
                             last CHECK_WEGFUEHRUNG if ($path[$#path-$i] ne $wegfuehrung->[$#$wegfuehrung-1-$i];
                          }
			  next;
                        }
		    }
';


    # Commoninit
    my $code = 'sub {
    my($self, $from, $to) = @_;
    my $str = $self->{Strassen};
    my $net = $self->{Net};
    my $wegfuehrung = $self->{Wegfuehrung};
';

    # Use_2_Init
    if ($use_2) {
	$code .= '
    $from = unpack("l", $self->{Coord2Index}{pack("l2", split(/,/, $from))});
    $to   = unpack("l", $self->{Coord2Index}{pack("l2", split(/,/, $to))});
';
    }

    # Visualinit
    if ($do_visual) {
	$code .= '
    my $red_val = 100;
';
    }

    # Statinit/VisualInit
    if ($do_stat || $do_visual) {
	$code .= '
    my($start_time) = (defined &Tk::timeofday ? Tk::timeofday() : time);
    my $last_time = $start_time;
';
    }

    # Penaltycode ...
    my($penalty_code) = "";
    if ($has_penalty) {
	if (!$srt_algo) {
	    $penalty_code .= '
                    my $next_node = $min_node;
                    my $last_node = $NODES{$min_node}->[0];
';
	}
	if ($has_ampeln && !$srt_algo) {
	    # XXX not yet for srt_algo
	    # XXX Penalty anpassen, falls nach links/rechts abgebogen wird.
	    # Keine Penalty bei Besonderheiten (nur eine Richtung ist relevant,
	    # Fugngerampel...) XXX
	    $penalty_code .= '
		    if (exists $ampel_net->{$next_node}) {
			$pen += ' . $ampel_penalty . ';
		    }
';
	}
	if ($has_qualitaet) {
	    $penalty_code .= '
		    if (defined $last_node and
                        exists $qualitaet_net->{$last_node}{$next_node}) {
                        $pen *= $qualitaet_penalty->{$qualitaet_net->{$last_node}{$next_node}}; # Qualittszuschlag
		    }
';
	}
	if ($has_strcat) {
	    $penalty_code .= '
		    if (defined $last_node and
                        exists $strcat_net->{$last_node}{$next_node}) {
                        $pen *= $strcat_penalty->{$strcat_net->{$last_node}{$next_node}}; # Kategorieaufschlag
		    }
';
	}
	if ($has_radwege) {
	    $penalty_code .= '
		    if (defined $last_node and
                        exists $radwege_net->{$last_node}{$next_node}) {
                        # Radwegeaufschlag
                        $pen *= $radwege_penalty->{$radwege_net->{$last_node}{$next_node}};
		    } else {
                        $pen *= $radwege_penalty->{""};
                    }
';
	}
	if ($has_steigung) {
	    $penalty_code .= '
		    if (defined $last_node and
                        exists $steigung_net->{$last_node}{$next_node}) {
                        my $norm_steigung = int(1000*$steigung_net->{$last_node}{$next_node});
                        if (!exists $steigung_penalty->{$norm_steigung}) {
                            $steigung_penalty->{$norm_steigung} = $steigung_penalty_sub->($norm_steigung);
                        }
                        $pen *= $steigung_penalty->{$norm_steigung}; # Steigungsaufschlag
		    }
';
	}
    }

###
### XXX grausam! Wie erstellt man sauber perl-Code?
### Vielleicht ein Modul dazu schreiben?
### Geschickt in Chunks aufteilen?
### ????
###

    ######################################################################
    # A*

    # NODES: Hash von Nodes auf [Vorgnger-Node ("x,y"),
    #                            Streckenlnge bis Node,
    #                            abgeschtzte Lnge bis Ziel ber Node,
    #                            weitere Array-Elemente sind optional ...]
    if (!$srt_algo) { # A*
	$code .= '
    my %OPEN = ($from => 1);
    my %NODES = ($from => [undef, 0, 0, undef]);
    my %CLOSED;
    while (1) {
';
	if ($do_visual) {
	    $code .= '
        if (Tk::timeofday() > $last_time + $visual_delay) {
            $canvas->idletasks;
            $last_time = Tk::timeofday();
        }
        $red_val+=5 if $red_val < 255;
        my $red_col = sprintf("#%02x0000", $red_val);
';
	}

	$code .= '
        return () if (keys %OPEN == 0);

        my $min_node;
	my $min_node_f;
	foreach (keys %OPEN) {
	    if (!defined $min_node_f || $NODES{$_}->[2] < $min_node_f) {
		$min_node = $_;
		$min_node_f = $NODES{$_}->[2];
	    }
	}
        # min_node wird aus OPEN nach CLOSED bewegt
	$CLOSED{$min_node} = 1;
	delete $OPEN{$min_node};
	if ($min_node eq $to) {
	    my @path;
	    my $len = 0;
	    while (1) {
		push @path, $min_node;
		my $prev_node = $NODES{$min_node}->[0];
		if (defined $prev_node) {
		    $len += Strassen::Util::strecke_s($min_node, $prev_node);
		    $min_node = $prev_node;
		} else {
		    last;
		}
	    }
	    @path = map { [ split(/,/, $_) ] } reverse @path;
';
	if ($do_stat) {
	    $code .= '
            warn "Search time: " . ((defined &Tk::timeofday ? Tk::timeofday() : time) - $start_time) . " s \n";
';
	}
	$code .= '
            return (\@path, $len, 0, 0, 0);
        }
	my @successors = keys %{ $net->{$min_node} };
	foreach my $successor (@successors) {

'

#  #XXXXXXXXXx lschen:!
#  . '
#  {
#  my($str_i0) = $self->net2name($NODES{$min_node}->[0], $min_node);
#  my($str_i)  = $self->net2name($min_node,   $successor);
#  warn "$str_i0 $str_i\n";
#  warn $self->{Strassen}->get($str_i0)->[0];
#  warn $self->{Strassen}->get($str_i)->[0];
#  }
#  '

. '
	    # den unmittelbaren Vorgnger ausschlieen
	    next if (defined $NODES{$min_node}->[0] &&
                     $NODES{$min_node}->[0] eq $successor);

	    # erlaubte Wegfhrungen beachten
	    # die Performance-Einbue liegt anscheinend unter 1%
	    if ($wegfuehrung and
		exists $wegfuehrung->{$successor}) {
		my($wegfuehrung) = $wegfuehrung->{$successor};
		my $this_node = $min_node;
		my $same = 1;
		for(my $i=$#$wegfuehrung-1; $i>=0; $i--) {
		    if ($wegfuehrung->[$i] ne $this_node) {
		        $same = 0;
			last;
		    }
		    $this_node = $NODES{$this_node}->[0];
		    if (!defined $this_node) {
		        $same = 0;
			last;
		    }
		}
		next if $same;
	    }
';
	if ($do_visual) {
	    $code .= '
	    if ($canvas) {
		# Ausgabe fr Visual Search
		my($lx, $ly) = $transpose_sub->(split(/,/, $min_node));
		my($nx, $ny) = $transpose_sub->(split(/,/, $successor));
		$canvas->createLine($lx,$ly,$nx,$ny,
				    -tag=>"visual",
				    -fill=>"red",-width=>3);
	    }
	    $node_touches++; # das gehrt in die Stat-Abteilung
';
	}
	$code .= "
            my \$" . $len_pen .' = $net->{$min_node}{$successor};
';
	if ($has_penalty) {
	    $code .= $penalty_code;
	}
	$code .= '
	    my $g = $NODES{$min_node}->[1] + ' . "\$" . $len_pen . ';
	    my $f = $g + Strassen::Util::strecke_s($successor, $to);
	    # !exists in OPEN and !exists in CLOSED:
	    if (!exists $NODES{$successor}) {
		$NODES{$successor} = [$min_node, $g, $f];
		$OPEN{$successor} = 1;
	    } else {
		if ($f < $NODES{$successor}->[2]) {
		    $NODES{$successor} = [$min_node, $g, $f];
		    if (exists $CLOSED{$successor}) {
			$OPEN{$successor} = 1;
			delete $CLOSED{$successor};
		    }
		}
	    }
	}
    }
 } # Achtung, Einrckung fr make_autoload!
';
	return $code;
    }

    ######################################################################
    # SRT-Algo - nur von historischem Interesse...

    # Format von $path_def:
    # 0: Referenz auf Pfad
    # 1: derzeitige Lnge
    # 2: virtuelle Penalty (ab hier einschlielich nur wenn
    #                       $has_penalty wahr ist)
    # 3: wahre Penalty
    # 4: Anzahl der Ampeln auf der Strecke (falls Ampeln eingestellt wurden)

    # Algorithm_Init
    $code .= '
    my @all_paths =
      (
       [[$from], 0';
    if ($has_penalty) {
	$code .= ', undef, 0';
	if ($has_ampeln) {
	    $code .= ', 0';
	}
    }
    if ($seen_optimierung) {
	$code .= ', {}';
    }

    $code .= '],
      );
    my(@found_paths, @suspended_paths);
    my %visited = ( $from => 0 );
BIGLOOP:
    while (1) {
';
    if ($do_stat > 1) {
	$code .= '
        $loop_count[0]++;
';
    };
    $code .= '
	while (@all_paths) {
';
    if ($do_visual) {
# XXX vielleicht auch: tags von visual0 bis visualX festlegen
# red_val-Inkrement mit (255-100)/X festlegen
# nachtrglich mit itemconfigure ndern
	$code .= '
	    if (Tk::timeofday() > $last_time+$visual_delay) {
                $canvas->idletasks;
                $last_time = Tk::timeofday();
            }
            $red_val+=5 if $red_val < 255;
            my $red_col = sprintf("#%02x0000", $red_val);
';
    }
    if ($do_stat > 2) {
	$code .= '
            $loop_count[1]++;
';
    };
    $code .= '
	    my @new_all_paths;
	    foreach my $path_def (@all_paths) {
';
    if ($do_stat > 3) {
	$code .= '
                $loop_count[2]++;
';
    };
    $code .= '
		my @path = @{$path_def->[0]};
		my $curr_len = $path_def->[1];
';
    if ($has_penalty) {
	$code .= '
		my $curr_pen = $path_def->[3];
';
	if ($has_ampeln) {
	    $code .= '
		my $ampeln   = $path_def->[4];
';
	}
    }
    if ($seen_optimierung) {
	$code .= '
                my $seen = $path_def->[5];
';
    }
    $code .= '
		my $last_node = $path[$#path];
';
    if ($use_2) {
	$code .= '
                my $net_s = $net->[$last_node];
                my $net_s_len = length($net_s);
                for(my $i = 0; $i < $net_s_len; $i+=8) {
                    my $next_node = unpack("l", substr($net_s, $i, 4));
                    my $len = unpack("l", substr($net_s, $i+4, 4));
';
    } else {
	$code .= '
	        my $last_node_ref = $net->{$last_node};
		while(my($next_node, $len)
		      = each %$last_node_ref) {
';
    }
    if ($seen_optimierung) {
	$code .= '
                   next if $seen->{$next_node};
';
    }

    $code .= '
                    next if $#path > 0 && $next_node eq $path[$#path-1];
';
    if ($do_visual) {
	$code .= '
                    { my($lx, $ly) = $transpose_sub->(split(/,/, $last_node));
                      my($nx, $ny) = $transpose_sub->(split(/,/, $next_node));
                      $canvas->createLine($lx,$ly,$nx,$ny,
                                          -tag=>"visual",
                                          -fill=>$red_col,-width=>3);
                    }
';
    }
    if ($do_stat) {
	$code .= '
                    $node_touches++;
';
    }
    my $sort_index = ($has_penalty ? 3 : 1);

    if ($has_penalty) {
	$code .= '  my $pen = $len;
';
    }
    $code .= $penalty_code;
    $code .= '
		    my $next_node_'.$len_pen." = \$".$len_pen.' + $curr_'.$len_pen . ';
' . $skip_path_code;
    if ($has_ampeln) {
	# XXX Penalty anpassen, falls nach links/rechts abgebogen wird.
	# Keine Penalty bei Besonderheiten (nur eine Richtung ist relevant,
	# Fugngerampel...) XXX
	$code .= '
		    if (exists $ampel_net->{$next_node}) {
			$next_node_pen += ' . $ampel_penalty . ';
' . $skip_path_code . '
			$ampeln++;
		    }
';
    }

    if ($has_abbiegen) {
	# Dieser Code ist nicht perfekt, aber eine gute Nherung.
	# Es wird festgestellt, ob es im vorherigen Knoten einen
	# Linksabbiegevorgang ohne Ampel gab. Wenn ja, wird noch
	# festgestellt, ob es beim Abbiegen eine Kategorienderung
	# gab. Nur in diesem Fall ist ein echtes Linksabbiegen
	# wahrscheinlich, ansonsten hat die Strae mglicherweise nur
	# einen leichten Knick gemacht. Da Hauptstraenkreuzungen
	# blicherweise eine Ampel haben, geht diese Regel nur bei
	# Kreuzungen von Nebenstraen und Hauptstraen.
	# XXX Echte Geradeausstrecken feststellen.
	$code .= '
                    if (@path > 1 and
                        !exists $ampel_net->{$path[$#path]} and
                        $net->{$path[$#path]} > 2
                       ) {
                        if (Strassen::Util::abbiegen($path[$#path-1],
                                                     $path[$#path],
                                                     $next_node) eq "l") {
                            my $str_i0 = $self->net2name($path[$#path-1],
                                                         $path[$#path]);
                            my $str_i = $self->net2name($path[$#path],
                                                        $next_node);
                            my $cat0 = $str->get($str_i0)->[2];
                            my $cat = $str->get($str_i)->[2];
                            if (($category_order->{$cat0} < $category_order->{$cat}
                                 and exists $abbiegen_penalty->{$cat}) or
                                ($category_order->{$cat0} > $category_order->{$cat}
                                 and exists $abbiegen_penalty->{$cat0})) {
#warn "pen=".($abbiegen_penalty->{$cat}||$abbiegen_penalty->{$cat0})." for " . $str->get($self->net2name($path[$#path-1],$path[$#path]))->[0] . " => " . $str->get($str_i)->[0] . "\n";
                                $next_node_pen += $abbiegen_penalty->{$cat};
' . $skip_path_code . '
                            }
			}
                    }
';
    }
    $code .= ($aufschlag_code ne '' ? '
                    if (!exists $visited{$next_node} or
                        $next_node_'.$len_pen.' < $visited{$next_node}) {' 
	      : '') . '
		        $visited{$next_node} = $next_node_'.$len_pen.';
' . ($aufschlag_code ne '' ? '}' : '') . '
';
    if ($has_penalty) {
	$code .= '
		    my $next_node_len = $len + $curr_len;
';
    }
    $code .= '
		    if ($next_node eq $to) {
			my @koords = (@path, $to);
';
    if ($use_2) {
	$code .= '
	                foreach (@koords) {
                            $_ = join(",", unpack("l2", $self->{Index2Coord}[$_]));
			}
';
    }
    $code .= '
			warn "Found path, len: $next_node_len\n"
			    if $VERBOSE;
			push(@found_paths,
			     [Strassen::to_koord(\@koords),
			      $next_node_len';
    if ($has_penalty) {
	$code .= ',
			      undef,
			      $next_node_pen,
';
	if ($has_ampeln) {
	    $code .= '			      $ampeln,
';
	}
    }
    $code .= '			     ]
			    );
';
    if ($pure_depth) {
	$code .= '
                        last BIGLOOP;
';
    }
    $code .= '
			next;
		    }
		    my $virt_'.$len_pen.' = $next_node_'.$len_pen;
    if ($use_2) {
	$code .= '
		      + Strassen::Util::strecke_i($self, $next_node, $to);
';
    } else {
	$code .= '
		      + Strassen::Util::strecke_s($next_node, $to);
';
    }
    if ($seen_optimierung) {
	$code .= '
                      $seen->{$next_node} = 1;
';
    }
    $code .= $skip_path_code2 . '
                    my $new_path_ref = [[@path, $next_node],
					 $next_node_len,
					 $virt_'.$len_pen.',
';
    if ($has_penalty) {
	$code .= '					  $next_node_pen,';
	if ($has_ampeln) {
	    $code .= '					  $ampeln,';
	}
    }
    if ($seen_optimierung) {
	$code .= '
                                                          $seen,';
    }
    $code .= '					 ];
                    push @new_all_paths, $new_path_ref;
		}
	    }
';
    if ($do_stat) {
	$code .= '
	    if ($max_new_paths < @all_paths) { $max_new_paths = @all_paths }
';
    }
    if ($pure_depth || $backtracking) {
	$code .= '
	    @all_paths = (sort({ $a->[2] <=> $b->[2] } @new_all_paths),
	                 @suspended_paths);
';
    } else {
	$code .= '
	    @all_paths = sort { $a->[2] <=> $b->[2] } 
	                 (@new_all_paths, @suspended_paths);
';
    }
    $code .= '
            @suspended_paths = splice(@all_paths, ' . $cut_path_nr . ');
';
    if ($do_stat) {
	$code .= '
	    if ($max_suspended_paths < @suspended_paths) {
                $max_suspended_paths = @suspended_paths;
            }
';
    }
	    # @all_paths enthlt jetzt die besten 5 Pfade,
	    # whrend in @suspended_paths alle anderen sind
	    # Damit wird gesichert, da
	    # a) die besten Pfade nur bevorzugt bearbeitet werden
	    # b) suspended_paths, falls sie irgendwann mal besser werden,
	    #    doch wieder zum Zug kommen knnen
    $code .= '

	}
	if (@suspended_paths) {
	    @all_paths = sort { $a->[2] <=> $b->[2] } splice
	      (@suspended_paths, 0, 5);
	} else {
	    last;
	}
    }

    if (!@found_paths) {
	warn "Nothing found!\n" if $VERBOSE;
        ';
    $code .= ($args{All} ? '()' : 'undef');
    $code .= ';
    } else {
	@found_paths = sort { $a->['.$sort_index.'] <=> $b->['.$sort_index.'] } @found_paths;
        ';

    if ($do_stat) {
	$code .= '
        $visited_nodes = scalar keys %visited;
warn "XXX";
        warn "\nAlgorithm: SRT (CutPath=' . $cut_path_nr .', PureDepth=' . $pure_depth . ', BackTracking=' . $backtracking . ')\n";
        warn "Search time: " . ((defined &Tk::timeofday ? Tk::timeofday() : time) - $start_time) . " s\n";
';
    }

    $code .= ($args{All} 
	      ? ($args{OnlyPath}
		 ? 'map { $_->[0] } @found_paths'
		 : '@found_paths'
		)
	      : ($args{OnlyPath}
		 ? '$found_paths[0]->[0]'
		 : '$found_paths[0]'
		)
	     );
    $code .= ';
    }
 } # Achtung, Einrckung fr make_autoload!
';

    $code;
}

# Sucht eine Route im Netz von $from bis $to.
#
# Rckgabewert:
# wenn AsObj gesetzt ist, dann eine Liste von Route-Objekten
# ansonsten eine Liste von Array-Referenzen mit folgendem Format:
#    [\@Path, $Len, $_, $Penalty, $Ampeln]
# \@Path ist eine Liste aus Punkten "$x,$y"
# $Len ist die Gesamtlnge in Metern
# $_: ?
# $Penalty ist die Penalty (in Metern ???)
# $Ampeln ist die Anzahl der Ampeln an der Route
#
### AutoLoad Sub
sub new_search {
    my($self, $from, $to, %args) = @_;

    my $ampel_net;
    if (exists $args{Ampeln}) {
	$ampel_net = $args{Ampeln}->{Net};
    }

    my($qualitaet_net, $qualitaet_penalty);
    if (exists $args{Qualitaet}) {
	$qualitaet_net = $args{Qualitaet}->{Net}->{Net};
	$qualitaet_penalty = $args{Qualitaet}->{Penalty} || die "No penalty";
    }

    my($strcat_net, $strcat_penalty);
    if (exists $args{Strcat}) {
	$strcat_net = $args{Strcat}->{Net}->{Net};
	$strcat_penalty = $args{Strcat}->{Penalty} || die "No penalty";
    }
    my($radwege_net, $radwege_penalty);
    if (exists $args{Radwege}) {
	$radwege_net = $args{Radwege}->{Net}->{Net};
	$radwege_penalty = $args{Radwege}->{Penalty} || die "No penalty";
    }
    my($steigung_net, $steigung_penalty, $steigung_penalty_sub);
    if (exists $args{Steigung}) {
	$steigung_net = $args{Steigung}->{Net}->{Net};
	$steigung_penalty = $args{Steigung}->{Penalty} || die "No penalty";
	$steigung_penalty_sub = $args{Steigung}->{PenaltySub} ||
	  die "No penalty subroutine";
    }
    my($abbiegen_penalty, $category_order);
    if (exists $args{Abbiegen}) {
	$category_order = $args{Abbiegen}->{Order} || die "No order";
	$abbiegen_penalty = $args{Abbiegen}->{Penalty} || die "No penalty";
    }
    # fr die Statistik:
    my($max_new_paths, $max_suspended_paths, $visited_nodes, $node_touches)
	= (0, 0, 0, 0);
    my(@loop_count);
    # fr Visual Search:
    my($canvas, $transpose_sub, $visual_delay);
    if ($args{'VisualSearch'}) {
	$canvas        = $args{'VisualSearch'}->{Canvas};
	$transpose_sub = $args{'VisualSearch'}->{Transpose};
	$visual_delay  = $args{'VisualSearch'}->{Delay};
	$canvas->delete("visual");
    }

    if (!$args{'ExtraResults'} && !$args{AsObj}) {
	# nur Routenpunkte ohne Streckenlnge etc. zurckgegebn
	$args{OnlyPath} = 1;
    }

    if ($args{'Algorithm'}) {
	push @INC, "$FindBin::RealBin/diplom/code";
	require BBBikeDiplom;
	if ($args{'Algorithm'} eq 'A*') {
	    return $self->search_A_star($from, $to, %args);
	} elsif ($args{'Algorithm'} eq 'DBI-A*') {
	    $args{'DBI'} = 1;
	    return $self->search_A_star($from, $to, %args);
	} elsif ($args{'Algorithm'} ne 'srt') {
	    die "Unknown algorithm $args{'Algorithm'}";
	}
    }

    my $code = $self->build_search_code(%args);
    if ($VERBOSE) {
	# dump code with line numbers
	my $i = 0;
	foreach (split(/\n/, $code)) {
	    $i++;
	    printf STDERR "%3d %s\n", $i, $_;
	}
    }
    my $search_sub = eval $code;
    warn $@ if $@;

    my @res;
    # XXX Via, All und !OnlyPath funktioniert nicht zusammen...
    if (exists $args{Via} and @{$args{Via}}) {
	my(@route) = ($from, @{$args{Via}}, $to);
	my @path;
	my $ges_len = 0;
	for(my $i = 0; $i < $#route; $i++) {
	    my($search_res, $len)
		= &$search_sub($self, $route[$i], $route[$i+1]);
	    if (ref $search_res eq 'ARRAY') {
		my(@found_path) = @$search_res;
		if ($i > 0) {
		    shift @found_path;
		}
		push(@path, @found_path);
		$ges_len += $len;
	    }
	}
	@res = (\@path, $ges_len);
    } else {
	@res = &$search_sub($self, $from, $to);
    }

    if ($args{Stat}) {
	warn "\nMax. new paths:       $max_new_paths\n";
	warn "Max. suspended paths: $max_suspended_paths\n";
	if (ref $res[0] eq 'ARRAY') {
	    warn "Path length (nodes):  " . scalar @{$res[0]} . "\n";
	}
	warn "Visited nodes:        $visited_nodes\n";
	warn "Node touches:         $node_touches\n";
	if ($visited_nodes) {
	    warn "Penetrance P:         "
		. sprintf("%.4f", scalar(@{$res[0]})/$visited_nodes) . "\n";
	    # XXX effective branching factor
	}
	if ($args{Stat} > 1) {
	    for(my $i=1; $i<=3; $i++) {
		if (defined $loop_count[$i-1]) {
		    warn "Loop count level $i:   " . $loop_count[$i-1] . "\n";
		}
	    }
	}
    }

    if ($args{AsObj}) {
	my $new_res = new Route(Path    => $res[0],
				Len     => $res[1],
				From    => $from,
				To      => $to,
				Penalty => $_->[3],
				Ampeln  => $_->[4],
				);
	$new_res;
    } else {
	@res;
    }
}

# Findet fr die Strecke c1-c2 die Position in Strassen.
# Von c1 und c2 mu mindestens ein Punkt in Net2Name existieren.
# Als zweiter Rckgabewert wird zurckgegeben, ob die Strecke rckwrts
# zur Strecke in der Datenbasis verluft.
### AutoLoad Sub
sub nearest_street {
    my($self, $c1, $c2) = @_;
    my $rueckwaerts = 0;
    my(@str);
    @str = keys %{ $self->{Net}{$c1} };
    if (!@str) {
	($c1, $c2) = ($c2, $c1);
	$rueckwaerts = 1;
	@str = keys %{ $self->{Net}{$c1} };
	if (!@str) {
	    warn "Kann weder $c1 noch $c2 in Net2Name finden"
	      if $VERBOSE;
	    return undef;
	}
    }

    my($x1,$y1) = split(/,/, $c1);
    my($x2,$y2) = split(/,/, $c2);

    my @winkel;

    for(my $i = 0; $i <= $#str; $i++) {
	my($xn,$yn) = split(/,/, $str[$i]);
	my($w) = schnittwinkel($x2,$y2, $x1,$y1, $xn,$yn);
	push @winkel, [$w, $i];
    }

    @winkel = sort { $a->[0] <=> $b->[0] } @winkel;
    ($self->net2name($c1, $str[$winkel[0]->[1]]), $rueckwaerts);
}

# Erzeugt aus einer Route eine Liste mit den zugehrigen Straennamen.
# Die Elemente der Liste bestehen aus:
#    [Straenname, Entfernung (m), Winkel , Richtung, [FromStr, ToStr]]
# FromStr und ToStr sind Indices auf das bergebene Routen-Array
# XXX
### AutoLoad Sub
sub route_to_name_1 {
    my($self, $route_ref, $start_i) = @_;
    my @strname;
    $start_i = 0 if !defined $start_i;
    my $i;
    # XXX kann beim acos anscheinend auftreten
    local $SIG{FPE} = sub { warn "Caught SIGFPE!" };
    for($i = 0; $i < $#{$route_ref}; $i++) {
	my $xy1 = Route::_coord_as_string([$route_ref->[$i][0],
					   $route_ref->[$i][1]]);
	my $xy2 = Route::_coord_as_string([$route_ref->[$i+1][0],
					   $route_ref->[$i+1][1]]);
	my($str_i, $rueckwaerts) = $self->net2name($xy1, $xy2);
	my($str);
	if (!defined $str_i) {
	    ($str_i, $rueckwaerts) = $self->nearest_street($xy1, $xy2);
	}
	if (defined $str_i) {
	    if ($str_i =~ /^\d/) {
		$str = $self->{Strassen}->get($str_i)->[0];
		if ($str =~ /^(\w+:\s+)?(.*\s-\s.*)$/) {
		    my $str_nummer = "";
 		    if (defined $1 and $1 ne "") {
		        $str_nummer = $1;
			$str = $2;
		    }
		    my(@comp) = ($rueckwaerts ? 
				 reverse split(/\s-\s/, $str) :
				 split(/\s-\s/, $str));
		    $str = $str_nummer . "(" . join(" - ", @comp[0..$#comp-1])
			. " -) " . $comp[$#comp];
		}
	    } else {
		$str = $str_i;
	    }
	} else {
	    # Aha. Wir haben hier wahrscheinlich einen angeklickten
	    # Punkt zwischen zwei Kurvenpunkten, der nicht mehr durch
	    # add_net abgedeckt ist. Also wird einfach geraten, ob der
	    # Punkt zur vorherigen Strecke gehrt, indem der Schnittwinkel
	    # berprft wird.
	    # Der Algorithmus ist nicht perfekt, weil einige Schnittwinkel
	    # im 90-Bereich liegen, wo es sich trotzdem um die gleiche
	    # Strae handelt. Naja.
	    if ($i+1 < $#{$route_ref}) { 
		my($w) = schnittwinkel
		  (split(/,/,$xy1),
		   split(/,/,$xy2),
		   split(/,/,Route::_coord_as_string
			 ([$route_ref->[$i+2][0],
			   $route_ref->[$i+2][1]])));
		if ($w < 0.15 || $w > 3.00) {
		    # ca. 10 Abweichung von der Geraden werden toleriert
		    $str = ($#strname >= 0 ? $strname[$#strname]->[0] : '???');
		}
	    }
	    # (Garantiert) unbekannte Strae.
	    if (!defined $str) {
		$str = "...";
	    }
	}
	my $entf = $self->{Net}{$xy1}{$xy2};
	if (!defined $entf) {
	    $entf = Strassen::Util::strecke($route_ref->[$i],
					    $route_ref->[$i+1]);
	}
	my($winkel, $richtung);
	if ($i+1 < $#{$route_ref}) {
	    my $a_len = Strassen::Util::strecke($route_ref->[$i],
						$route_ref->[$i+1]);
	    my $b_len = Strassen::Util::strecke($route_ref->[$i+1],
						$route_ref->[$i+2]);
	    my $a1 = $route_ref->[$i+1][0]-$route_ref->[$i][0];
	    my $a2 = $route_ref->[$i+1][1]-$route_ref->[$i][1];
	    my $b1 = $route_ref->[$i+2][0]-$route_ref->[$i+1][0];
	    my $b2 = $route_ref->[$i+2][1]-$route_ref->[$i+1][1];
	    $richtung = ($a1*$b2-$a2*$b1 > 0 ? 'l' : 'r');
	    $winkel = ($a_len == 0 || $b_len == 0 ? 0 
		       : rad2deg(&$acos(($a1*$b1+$a2*$b2)/($a_len*$b_len))));
	    $winkel = -$winkel if $winkel < 0; # bei altem Math::Trig::acos 
	}
	if (@strname && $str eq $strname[$#strname]->[0]) {
	    $strname[$#strname][1] += $entf;
	    $strname[$#strname][2] = $winkel;
	    $strname[$#strname][3] = $richtung;
	    $strname[$#strname][4][1] = $i+$start_i;
	} else {
	    my @val = ($str, $entf, $winkel, $richtung,
		       [$i+$start_i, $i+$start_i]);
	    push @strname, [@val];
	}

    }

    @strname;
}

### AutoLoad Sub
sub route_to_name_2 {
    my($self, $route_ref, $start_i) = @_;
    my @strname;
    $start_i = 0 if !defined $start_i;
    my $i;
    # XXX kann beim acos anscheinend auftreten
    local $SIG{FPE} = sub { warn "Caught SIGFPE!" };
    for($i = 0; $i < $#{$route_ref}; $i++) {
	my $xy1 = $self->{Coord2Index}->
	  {pack("l2", $route_ref->[$i][0], $route_ref->[$i][1])};
	my $xy1_u = unpack("l", $xy1);
	my $xy2 = $self->{Coord2Index}->
	  {pack("l2", $route_ref->[$i+1][0], $route_ref->[$i+1][1])};
	my $str_i;
	{
	    my $net_s = $self->{Index2Pos}[$xy1_u];
	    my $net_s_len = length($net_s);
	    for(my $i = 0; $i < $net_s_len; $i+=8) {
		if (substr($net_s, $i, 4) eq $xy2) {
		    $str_i = unpack("l", substr($net_s, $i+4, 4));
		    last;
		}
	    }
	}
	my $str;
	if (!defined $str_i) {
	    $str_i = $self->nearest_street($xy1, $xy2);
	}
	if (defined $str_i) {
	    $str = ($str_i =~ /^\d/
		    ? $self->{Strassen}->get($str_i)->[0] 
		    : $str_i
		   );
	} else {
	    # XXX Siehe route_to_name1. Hier fehlt noch was.
	    $str = "...";
	}
# 	my $str = (defined $str_i
# 		   ? ($str_i =~ /^\d/
# 		      ? $self->{Strassen}->get($str_i)->[0] 
# 		      : $str_i
# 		     )
# 		   : "...");
	my $entf;
	{
	    my $net_s = $self->{Net}[$xy1_u];
	    my $net_s_len = length($net_s);
	    for(my $i = 0; $i < $net_s_len; $i+=8) {
		if (substr($net_s, $i, 4) eq $xy2) {
		    $entf = unpack("l", substr($net_s, $i+4, 4));
		    last;
		}
	    }
	}
	if (!defined $entf) {
	    $entf = Strassen::Util::strecke($route_ref->[$i],
					    $route_ref->[$i+1]);
	}
	my($winkel, $richtung);
	if ($i+1 < $#{$route_ref}) {
	    my $a_len = Strassen::Util::strecke($route_ref->[$i],
						$route_ref->[$i+1]);
	    my $b_len = Strassen::Util::strecke($route_ref->[$i+1],
						$route_ref->[$i+2]);
	    my $a1 = $route_ref->[$i+1][0]-$route_ref->[$i][0];
	    my $a2 = $route_ref->[$i+1][1]-$route_ref->[$i][1];
	    my $b1 = $route_ref->[$i+2][0]-$route_ref->[$i+1][0];
	    my $b2 = $route_ref->[$i+2][1]-$route_ref->[$i+1][1];
	    $richtung = ($a1*$b2-$a2*$b1 > 0 ? 'l' : 'r');
	    $winkel = ($a_len == 0 || $b_len == 0 ? 0 
		       : rad2deg(&$acos(($a1*$b1+$a2*$b2)/($a_len*$b_len))));
	    $winkel = -$winkel if $winkel < 0; # bei altem Math::Trig::acos 
	}
	if (@strname && $str eq $strname[$#strname]->[0]) {
	    $strname[$#strname][1] += $entf;
	    $strname[$#strname][2] = $winkel;
	    $strname[$#strname][3] = $richtung;
	    $strname[$#strname][4][1] = $i+$start_i;
	} else {
	    push @strname, [$str, $entf, $winkel, $richtung,
			    [$i+$start_i, $i+$start_i]];
	}

    }

    @strname;
}

*route_to_name = \&route_to_name_1;

# XXX
# Argumente:
#   Route:         Liste des Pfades, wie von new_search zurckgegeben
#   Coords:        Koordinaten im Canvas (?)
#   Km:            Strecken werden in km statt m angegeben
#   AngleAccuracy: gibt an, in welcher Genauigkeit Winkel zurckgegeben werden
sub route_info {
   my($self, %args) = @_;

   my $routeref       = $args{Route} || die "Missing argument: Route";
   my $coords         = $args{Coords};
   my $s_in_km        = $args{Km};
   my $angle_accuracy = $args{AngleAccuracy} || 10;

   my $s_sub = ($s_in_km ? sub { m2km($_[0]) } : sub { $_[0] });

   my @search_route = $self->route_to_name($routeref);
   my @route_info;
   my @route_strnames;
   my $ges_entf = 0;
   my($next_entf, $ges_entf_s, $next_winkel, $next_richtung)
	= ("", "", undef, "");
   my $last_str;
   for(my $i = 0; $i <= $#search_route; $i++) {
	my $route_info_item = {};
	my($str, $index_arr);
	my $himmelsrichtung;
	my($entf, $winkel, $richtung)
	    = ($next_entf, $next_winkel, $next_richtung);
	($str, $next_entf, $next_winkel, $next_richtung, $index_arr)
	    = @{$search_route[$i]};
	my $route_strnames_index;
	if ($str ne '...' &&
	    (!defined $last_str || $last_str ne $str)) {
	    $last_str = $str;
	    $str = Strassen::strip_bezirk($str);
	    if (ref $index_arr eq 'ARRAY' &&
		ref $coords eq 'ARRAY' && 
		defined $index_arr->[0] &&
		defined $coords->[$index_arr->[0]] &&
		defined $coords->[$index_arr->[0]+1]) {
		my($x, $y) = ($coords->[$index_arr->[0]]->[0],
			      $coords->[$index_arr->[0]]->[1]);
		push @route_strnames, [$str, $x, $y, $index_arr->[0]];
		$route_strnames_index = $#route_strnames;
	    }
	}

	if ($i < $#search_route and  ref $index_arr eq 'ARRAY') {
	    $himmelsrichtung =
		    uc(main::line_to_canvas_direction
		       (@{ $routeref->[$index_arr->[0]] },
			@{ $routeref->[$index_arr->[0]+1] }));
	}

	if ($i > 0) {
	    if (!$winkel) { $winkel = 0 }
	    $winkel = int($winkel/$angle_accuracy)*$angle_accuracy;
	    if ($winkel < 30) {
		$richtung = "";
	    } else {
		$richtung = ($winkel <= 45 ? 'H' : '') . uc($richtung);
	    }
	    $ges_entf += $entf;
	    if (defined $route_strnames_index) {
		$route_strnames[$route_strnames_index]->[4]
		    = $s_sub->($ges_entf);
	    }
	}

	$route_info_item->{Hop}       = $s_sub->($entf);
	$route_info_item->{Whole}     = $s_sub->($ges_entf);
	$route_info_item->{Way}       = $richtung;
	$route_info_item->{Angle}     = $winkel;
	$route_info_item->{Direction} = $himmelsrichtung;
	$route_info_item->{Street}    = $str;

	push @route_info, $route_info_item;
   }

   push @route_info, {Whole => $s_sub->($ges_entf+$next_entf),
		      Hop   => $s_sub->($next_entf),
		     };

   @route_info;
}


# Bei einer Speicherung als MLDBM mu der in der Manpage beschriebene
# Bug umgangen werden. Diese Funktion funktioniert fr
# zweistufige Hashes
sub store_to_hash {
    my($self, $mldbm_hash, $key1, $key2, $val) = @_;
    if ($self->{UseMLDBM}) {
	my $tmp = $mldbm_hash->{$key1};
	$tmp->{$key2} = $val;
	$mldbm_hash->{$key1} = $tmp;
    } else {
	$mldbm_hash->{$key1}{$key2} = $val;
    }
}

# Lscht den Punkt aus dem Straennetz-Graphen
# Wenn nur ein Punkt angegeben ist, dann werden alle Nachbarn entfernt.
# Wenn zwei Punkte angegeben sind, dann wird nur diese Strecke entfernt,
# und zwar nur in dieser Richtung, wenn dir == 1, oder beide Richtungen,
# wenn dir == 2
sub del_net {
    my($self, $point1, $point2, $dir) = @_;
    if (!defined $point2) {
	foreach (keys %{$self->{Net}{$point1}}) {
	    delete $self->{Net}{$point1}{$_};
	    delete $self->{Net}{$_}{$point1};
	}
    } else {
	delete $self->{Net}{$point1}{$point2};
	if ($dir != 1) { # also 2
	    delete $self->{Net}{$point2}{$point1};
	}
    }
}

# Der erste Punkt in @points ist der einzufgende, die beiden anderen
# existieren bereits. Rckwrts-Eigenschaft wird beachtet.
### AutoLoad Sub
sub add_net {
    my($self, $pos, @points) = @_;
    return unless defined $pos;
    die 'Es mssen genau 3 Punkte in @points sein!' if @points != 3;
    my($startx, $starty) = @{$points[0]};
    my $starts = Route::_coord_as_string([$startx,$starty]);
    my @ex_point;
    for (1..2) {
	$ex_point[$_] = Route::_coord_as_string($points[$_]);
    }
    my $rueckwaerts = 0;
    if (exists $self->{Net2Name}{$ex_point[2]}{$ex_point[1]}) {
        $rueckwaerts = 1;
    }

    my $i;
    for($i=1; $i<=2; $i++) {
        my $s = $ex_point[$i];
	my $entf = Strassen::Util::strecke($points[0], $points[$i]);
	if (!exists $self->{Net}{$starts}{$s}) {
	    $self->store_to_hash($self->{Net}, $starts, $s, $entf);
	    #$self->{Net}{$starts}{$s} = $entf;
	    push(@{$self->{AdditionalNet}}, [$starts, $s]);
	}
	if (!exists $self->{Net}{$s}{$starts}) {
	    $self->store_to_hash($self->{Net}, $s, $starts, $entf);
	    #$self->{Net}{$s}{$starts} = $entf;
	    push(@{$self->{AdditionalNet}}, [$s, $starts]);
	}
	# XXX $pos ist hier immer definiert...
	if (!exists $self->{Net2Name}{$starts}{$s} &&
	    defined $pos) {
  	    if (($i == 1 && $rueckwaerts) || $i == 2) {
		$self->store_to_hash($self->{Net2Name}, $starts, $s, $pos);
	        #$self->{Net2Name}{$starts}{$s} = $pos;
	    } else {
		$self->store_to_hash($self->{Net2Name}, $s, $starts, $pos);
	        #$self->{Net2Name}{$s}{$starts} = $pos;
            }
	}
    }
}

### AutoLoad Sub
sub add_faehre {
    my($self, $faehre_file, %args) = @_;
    my $faehre_obj = new Strassen $faehre_file;
    $faehre_obj->init;
    while(1) {
	my $ret = $faehre_obj->next;
	last if !@{$ret->[1]};
	my @kreuzungen = @{$ret->[1]};
	my $i;
	# XXX record to make deletion possible
	for($i = 1; $i<=$#kreuzungen; $i++) {
	    $self->{Net}{$kreuzungen[$i-1]}{$kreuzungen[$i]} = 0;
	    $self->{Net}{$kreuzungen[$i]}{$kreuzungen[$i-1]} = 0;
	    $self->{Net2Name}{$kreuzungen[$i-1]}{$kreuzungen[$i]} =
	      "Fhre " . $ret->[0];
	}
    }
}

# Argument:
# (Multi)Strassen-Objekt der Bahnhfe
# optional: -addmap     (Mapping der Umsteigebahnhfe)
#           -addmapfile (Datei mit Mapping)
### AutoLoad Sub
sub add_umsteigebahnhoefe {
    my($self, $bhf_obj, %args) = @_;

    # XXX untested
    if (exists $args{-addmapfile}) {
    TRY: {
	    foreach my $dir (@Strassen::datadirs) {
		if (open(F, "$dir/" . $args{-addmapfile})) {
		    my %map;
		    while(<F>) {
			next if /^\#/;
			chomp;
			my(@l) = split /\t/;
			$map{$l[0]} = $l[1];
		    }
		    close F;
		    if (keys %map) {
			$args{-addmap} = \%map;
		    }
		    last TRY;
		}
	    }
	}
    }

    my %bahnhoefe;
    $bhf_obj->init;
    while(1) {
	my $ret = $bhf_obj->next;
	last if !@{ $ret->[1] };
	my $name   = Strassen::strip_bezirk($ret->[0]);
	if (defined $args{-addmap} and
	    exists $args{-addmap}->{$name}) {
	    $name = $args{-addmap}->{$name};
	}
	my $coords = $ret->[1][0];
	if (exists $bahnhoefe{$name}) {
	    my($x1,$y1) = split(/,/, $coords);
	    foreach my $p (@{ $bahnhoefe{$name} }) {
		if ($coords ne $p) {
		    my($x2,$y2) = split(/,/, $p);
		    my $entf = Strassen::Util::strecke([$x1,$y1],[$x2,$y2]);
		    $self->store_to_hash($self->{Net}, $coords, $p, $entf);
		    $self->store_to_hash($self->{Net}, $p, $coords, $entf);
		}
	    }
	    push @{ $bahnhoefe{$name} }, $coords;
	} else {
	    $bahnhoefe{$name} = [$coords];
	}
    }
}

### AutoLoad Sub
sub del_add_net {
    my $self = shift;
    my $a;
    my $b;
    foreach $b (@{$self->{AdditionalNet}}) {
	delete $self->{Net}{$b->[0]}{$b->[1]};
	if (exists $self->{Net2Name}{$b->[0]}{$b->[1]}) {
	    delete $self->{Net2Name}{$b->[0]}{$b->[1]};
	}
    }
    @{$self->{Additional}} = ();
    @{$self->{AdditionalNet}} = ();
}

### AutoLoad Sub
sub reachable_1 {
    my($self, $coord) = @_;
    if (!exists $self->{Net}{$coord}) {
	warn "Die Koordinate $coord kann im Netz nicht erreicht werden\n"
	  if $VERBOSE;
	0;
    } else {
	1;
    }
}

### AutoLoad Sub
sub reachable_2 {
    my($self, $coord) = @_;
    if (!defined $self->{Net}[$self->{Coord2Index}{$coord}]) {
	warn "Die Koordinate $coord kann im Netz nicht erreicht werden\n"
	  if $VERBOSE;
	0;
    } else {
	1;
    }
}

*reachable = \&reachable_1;

# Falls die Koordinate nicht exakt im Netz existiert, wird der nchstgelegene
# Punkt gesucht und zurckgegeben, ansonsten der exakte Punkt.
# Die Koordinate ist im "x,y"-Format angegeben.
# XXX Funktioniert die Methode auch mit Data_Format 2?
### AutoLoad Sub
sub fix_coords {
    my($self, $coord) = @_;
    if (!$self->reachable($coord)) {
	$self->make_crossings();
	my(@nearest) = $self->{Crossings}->nearest_coord($coord);
	if (@nearest) {
	    $nearest[0];
	} else {
	    warn "Can't find another point near to $coord.\n";
	    undef;
	}
    } else {
	$coord;
    }
}

### AutoLoad Sub
sub make_crossings {
    my $self = shift;
    if (!defined $self->{Crossings}) {
	warn "In StrassenNetz::make_crossings...\n" if $VERBOSE;
	%{ $self->{CrossingsHash} } = $self->{Strassen}->all_crossings
	  (RetType => 'hash', UseCache => 1);
	$self->{Crossings} = new Kreuzungen Hash => $self->{CrossingsHash};
	$self->{Crossings}->make_grid;
	warn "...done\n" if $VERBOSE;
    }
}

sub null { }

### AutoLoad Sub
sub use_data_format {
    my $a = shift;
    if (defined $a) {
	$data_format = $a;
    }

    local($^W) = 0;

    *make_net      = ($data_format == 1 ? \&make_net_slow_1 : \&make_net_slow_2);
    *make_sperre   = ($data_format == 1 ? \&make_sperre_1 : \&null);
    *route_to_name = ($data_format == 1 ? \&route_to_name_1 : \&route_to_name_2);
    *reachable     = ($data_format == 1 ? \&reachable_1 : \&reachable_2);
    # XXX restliche ...
}

main::mymstat("Ende von Strassen.pm");

1;
