# -*- perl -*-

#
# $Id: Update.pm,v 1.4 1999/10/07 08:21:53 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package Update;

use strict;
use vars qw($verbose $tmpdir $proxy $progname $VERSION);

sub update {
    my(%args) = @_;
    my $root = delete $args{-root} || die "No root";
    my(@files) = @{$args{-files}};
    my $dest = delete $args{-dest} || die "No destination";
    my(%modified) = %{$args{-modified}};
    my $ua;
    eval {
	require LWP::UserAgent;
	$ua = new LWP::UserAgent;
	$ua->agent("$main::progname/$main::VERSION");
	if ($main::proxy) {
	    $ua->proxy(['http', 'ftp'], $main::proxy);
	}
    };
    if ($@ || !$ua) {
	undef $ua;
	require Http;
	Http->VERSION(0.22);
    }
    $main::progress->Init(-dependents => $main::c,
			  -label => "Updating via Internet");
    my $i = 0;
    foreach my $file (@files) {
	my $src_file  = $root . "/" . $file;
	#main::update_progress($i++/$#files, $src_file);
	$main::progress->Update($i++/$#files, $src_file);
	my $dest_file = $dest . "/" . $file . "~";
	my $real_dest_file = $dest . "/" . $file;
	my $h;
	if ($ua) {
	    $h = new HTTP::Headers;
	    if (exists $modified{$file} && -f $real_dest_file) {
		$h->if_modified_since($modified{$file});
	    }
	} else {
	    if (exists $modified{$file} && -f $real_dest_file) {
		$h = {'time' => $modified{$file}};
	    } else {
		$h = {};
	    }
	}
	if ($main::verbose) {
	    print STDERR "$src_file => $dest_file...";
	}
	my($res, $success);
	if ($ua) {
	    $res = $ua->request(new HTTP::Request('GET', $src_file, $h),
				$dest_file);
	    $success = $res->is_success;
	} else {
	    my(%res) = Http::get('url' => $src_file,
				 %$h,
				);
	    $success = ($res{'error'} == 200);
	    if ($success) {
		if (!open(OUT, ">$dest_file")) {
		    $success = 0;
		} else {
		    print OUT $res{'content'};
		    close OUT;
		}
	    }
	}
	if ($success) {
	    my $tmp = $dest_file . "~~";
	    rename $real_dest_file, $tmp;
	    rename $dest_file, $real_dest_file;
	    unlink $tmp;
	    print STDERR " loaded\n" if $main::verbose;
	} elsif ($ua) {
	    if ($res->is_error) {
		print STDERR "\n", $res->as_string;
	    } else {
		print STDERR " OK\n" if $main::verbose;
	    }
	}
	unlink $dest_file;
    }
    #main::finish_progress();
    $main::progress->Finish;
}

sub create_modified_devel {
    my $rootdir = "..";
    my $datadir = $rootdir . "/data";
    if (!-f "$rootdir/bbbike" || !-d $datadir || !-f "$rootdir/MANIFEST") {
	die "Probably wrong rootdir: $rootdir from `pwd`";
    }
    
    open(MOD, ">$datadir/.modified") or die $!;
    open(MANI, "$rootdir/MANIFEST") or die $!;
    while(<MANI>) {
	if (m|^data/(.*)|) {
	    my $file = $1;
	    next if $file =~ m|^\.|;
	    my(@stat) = stat("$datadir/$file");
	    print MOD "data/$file\t$stat[9]\n";
	}
    }
    close MANI;
    close MOD;
}

sub create_modified {
    my(%args) = @_;
    my $destdir = $args{-dest};
    my $datadir = $destdir . "/data";
    my(@files) = @{$args{-files}};
    my(%modified) = %{$args{-modified}};
    if (open(MOD, ">$datadir/.modified")) {
	foreach my $file (@files) {
	    my(@stat) = stat("$file");
	    print MOD "$file\t$stat[9]\n";
	}
	close MOD;
    } else {
	warn "Can't open .modified: $!";
    }
}

sub bbbike_data_update {
    use FindBin;
    my $rootdir = "$FindBin::RealBin";
    # sichergehen, dass nicht die Originaldateien berschrieben werden...
    die "FATAL: suspicious rootdir: $rootdir"
      if ($rootdir =~ m|/home/e/eserte/src/bbbike|);
    die "FATAL: RCS in datadir detected"
      if (-e "$rootdir/data/RCS");
    die "FATAL: Makefile in datadir detected"
      if (-e "$rootdir/data/Makefile");

    my(@files, %modified);
    if (open(MOD, "$rootdir/data/.modified")) {
	while(<MOD>) {
	    chomp;
	    my($f, $t) = split(/\t/);
	    push @files, $f;
	    $modified{$f} = $t;
	}
	close MOD;
	update(-dest => $rootdir,
	       # XXX check it
	       #-root => "http://pub.cs.tu-berlin.de/src/BBBike/update",
	       -root => "http://pub.cs.tu-berlin.de/lib/BBBike",
	       #-root => "http://www/~eserte/src/bbbike",
	       -files => \@files,
	       -modified => \%modified,
	      );
	create_modified(-dest => $rootdir,
			-files => \@files,
			-modified => \%modified);
	main::reload_all();
    }
}

1;

__END__
