#!/usr/bin/perl
# -*- perl -*-

#
# $Id: cmdbbbike,v 3.3 2000/11/30 20:07:57 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 2000 Slaven Rezic. All rights reserved.
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

# the true command line tool
# ... and now with wap and sms interface!

package BBBike::Cmd;

use FindBin;
use lib ("$FindBin::RealBin",
	 "$FindBin::RealBin/lib",
	 "$FindBin::RealBin/data",
	 "$FindBin::RealBin/..",
	 "$FindBin::RealBin/../lib",
	 "$FindBin::RealBin/../data",
	);
use strict;

$^W = 1;

my $s;
my $net;
my $plz;
my $kr;
my($multi_from, $multi_to);

# global CGI and BrowserInfo object
use vars qw($q $bi);

# start/end parameter and coords
use vars qw($from_street $from_citypart);
use vars qw($to_street   $to_citypart);
use vars qw($start_coord $ziel_coord);
# result list
use vars qw(@routeinfo @path);
# context
use vars qw(%context);
# to XS or not to XS
use vars qw($use_xs $use_cache);
$use_xs    = 1 unless defined $use_xs;
$use_cache = 1 unless defined $use_cache;

# pseudo object
sub new {
    bless {}, $_[0];
}

sub init_str {
    if (!$s) {
	require Strassen;
	$s = Strassen->new("strassen");
	if (defined $context{-vehicle} && $context{-vehicle} eq 'car') {
	    $s = Strassen->new_copy_restricted($s, -restrictions => [qw/NN/]);
	}
    }
}

sub init_plz {
    if (!$plz) {
	require PLZ;
	$plz = PLZ->new();
    }
}

sub init_net {
    if (!$net) {
	init_str();
	if ($use_xs) {
	    eval q{ use BBBikeXS };
	}
	$net = StrassenNetz->new($s);
	$net->make_net(UseCache => $use_cache);
	$net->make_sperre('gesperrt',
			  Type => ['einbahn', 'sperre',
				   'tragen', 'wegfuehrung']);
    }
}

sub init_kr {
    if (!$kr) {
	init_str();
	$kr = Kreuzungen->new(Strassen => $s);
	$kr->make_grid;
    }
}

sub get_start {
    init_plz();
    init_net();

    my(@from_res) = $plz->look_loop_best($from_street,
					 MultiZIP => 1,
					 MultiCitypart => 1,
					 Agrep => 'default',
					 Citypart => $from_citypart);
    $multi_from = @{ $from_res[0] } > 1;
    my $from_data = $from_res[0]->[0];
    ($from_street, $from_citypart) = ($from_data->[0],
				      $from_data->[1]);

    $start_coord = $from_data->[3];
    if (!$net->reachable($start_coord)) {
	init_kr();
	$start_coord = ($kr->nearest_coord($start_coord))[0];
    }
    $start_coord;
}

sub fix_ziel {
    init_net();
    if (!$net->reachable($ziel_coord)) {
	init_kr();
	$ziel_coord = ($kr->nearest_coord($ziel_coord))[0];
    }
    $ziel_coord;
}

sub get_ziel {
    my(@to_res)   = $plz->look_loop_best($to_street,
					 MultiZIP => 1,
					 MultiCitypart => 1,
					 Agrep => 'default',
					 Citypart => $to_citypart);
    $multi_to = @{ $to_res[0] } > 1;
    my $to_data = $to_res[0]->[0];
    ($to_street, $to_citypart) = ($to_data->[0],
				  $to_data->[1]);

    $ziel_coord  = $to_data->[3];
    fix_ziel();
}

sub search {
    init_plz();
    init_net();

    get_start() if !defined $start_coord;
    get_ziel()  if !defined $ziel_coord;

    @path = ();
    my($res) = $net->new_search($start_coord, $ziel_coord);
    if (defined $res) {
	@path = @$res;
    }

    @routeinfo = $net->route_info(Route => \@path,
				  Km    => 1);
}

sub cmdline_input {
    require Getopt::Long;

    if (!Getopt::Long::GetOptions
	("vehicle=s" => sub { $context{'-vehicle'} = $_[1] })) {
	die "Usage: $0 [-vehicle vehicle] start ziel";
    }

    if (@ARGV < 2) {
	print "Zu wenige Argumente: mindestens Start- und Zielstrae angeben!\n";
	exit 1;
    }
    if (@ARGV == 2) {
	($from_street, $to_street) = @ARGV;
    } elsif (@ARGV == 4) {
	($from_street, $from_citypart, $to_street, $to_citypart) = @ARGV;
    } elsif (@ARGV == 3) {
	$from_street = shift @ARGV;
	if ($ARGV[0] =~ /^\d+$/) {
	    $from_citypart = shift @ARGV;
	} elsif ($ARGV[1] =~ /^\d+$/) {
	    $to_citypart = pop @ARGV;
	}
	if (@ARGV == 1) {
	    $to_street = shift @ARGV;
	} else {
	    require Geography;
	    my $geo = Geography->new('Berlin', 'DE');
	    $ENV{LANG} = "de";
	    my(@cityparts) = keys %{ $geo->subcitypart_to_citypart };
	    foreach (@cityparts) {
		if ($_ =~ /^\Q$ARGV[0]\E/i) {
		    ($from_citypart, $to_street) = @ARGV;
		    last;
		} elsif ($_ =~ /^\Q$ARGV[1]\E/i) {
		    ($to_street, $to_citypart) = @ARGV;
		    last;
		}
	    }
	    if (!defined $to_street) {
		($to_street, $to_citypart) = @ARGV;
	    }
	}
    }
}

sub cmdline_output {
    local($^W) = 0;

    print "$from_street";
    if ($multi_from) {
	print " ($from_citypart)";
    }
    print " - ";
    print "$to_street";
    if ($multi_to) {
	print " ($to_citypart)";
    }
    print "\n";

    print
	join("\n", map { sprintf("%-40s %s", $_->{Street}, $_->{Whole}) }
	               @routeinfo), "\n";
}

*sms_input = \&cmdline_input;

sub sms_output {
    my $first = 1;
    foreach (@routeinfo) {
	if ($_->{Street}) {
	    if ($_->{Way} ne '') {
		print "*" . substr($_->{Way}, -1) . "*";
	    } elsif ($first) {
		$first = 0;
	    } else {
		print ",";
	    }
	    print Strasse::short($_->{Street}, 3);
	} else {
	    print ":" . $_->{Whole} . ".\n";
	}
    }
    # ja, nein?
    if (0 && ($multi_from || $multi_to)) {
	print "Route von $from_street ($from_citypart) nach $to_street ($to_citypart)\n";
    }
}

use vars qw($can_table);

sub wap_can_table { $can_table }

sub wap_input {
    print <<EOF;
@{[ wap_header() ]}
 <card id="input">
  <do type="options" label="Info" name="info">
   <go href="@{[ $q->script_name ]}?info=1"/>
  </do>
  <do type="reset" label="Neu" name="reset"><refresh /></do>
  <p align="center"><big>BBBike</big></p>
  <p><b>Start</b><br/>
     Strae: <input type="text" name="startname" /><br/>
     Bezirk: <input type="text" name="startbezirk" /><br/>
     Ziel<br/>
     Strae: <input type="text" name="zielname" /><br/>
     Bezirk: <input type="text" name="zielbezirk" /><br/>
  <anchor>Route zeigen
   <go href="@{[ $q->script_name ]}" method="get">
    <postfield name="startname"   value="\$startname" />
    <postfield name="startbezirk" value="\$startbezirk" />
    <postfield name="zielname"    value="\$zielname" />
    <postfield name="zielbezirk"  value="\$zielbezirk" />
   </go>
  </anchor>
  </p>
 </card>
@{[ wap_footer() ]}
EOF
}

sub wap_output {
    print <<EOF;
@{[ wap_header() ]}
 <card id="output">
  <do type="options" label="Rckweg" name="back">
   <go href="@{[ $q->script_name ]}">
    <postfield name="startname"   value="$to_street" />
    <postfield name="startbezirk" value="$to_citypart" />
    <postfield name="zielname"    value="$from_street" />
    <postfield name="zielbezirk"  value="$from_citypart" />
   </go>
  </do>
  <do type="options" label="Neue Anfrage" name="newsearch">
   <go href="@{[ $q->script_name ]}">
    <setvar name="startname"   value="" />
    <setvar name="startbezirk" value="" />
    <setvar name="zielname"    value="" />
    <setvar name="zielbezirk"  value="" />
   </go>
  </do>
  <do type="options" label="Info" name="info">
   <go href="@{[ $q->script_name ]}">
    <postfield name="info" value="1" />
   </go>
  </do>
  <do type="prev" label="Zurck" name="prev"><prev /></do>
  <p>Route von <b> $from_street ($from_citypart) </b> nach <b> $to_street ($to_citypart) </b><br/>
  @{[ $can_table ? wap_output_table() : wap_output_notable() ]}
  </p>
 </card>
@{[ wap_footer() ]}
EOF
}

sub wap_output_table {
    my $out = "<table columns=\"2\">\n";
    foreach (@routeinfo) {
	$out .= "<tr><td>";
	if (defined $_->{Street}) {
	    $out .= $_->{Street};
	}
	$out .= "</td><td>";
	if (defined $_->{Way} && $_->{Way} ne "") {
	    $out .= $_->{Way};
	}
	$out .= "</td></tr>\n";
    }
    $out .= "<tr><td>$routeinfo[-1]->{Whole}</td><td></td></tr></table>\n";
    $out;
}

sub wap_output_notable {
    my $out = "";
    foreach (@routeinfo) {
        if (defined $_->{Way} && $_->{Way} ne "") {
	    $out .= "$_->{Way} =&gt; ";
        }
	if (defined $_->{Street}) {
	    $out .= $_->{Street};
	}
	$out .= "<br/>\n";
    }
    $out .= "$routeinfo[-1]->{Whole}<br/>\n";
    $out;
}

sub wap_info {
    print <<EOF;
@{[ wap_header() ]}
 <card id="info">
  <do type="prev" label="Zurck" name="prev"><prev /></do>
  <p><b>BBBike</b><br/>
  by Slaven Rezic [rezic\@onlineoffice.de]</p>
 </card>
@{[ wap_footer() ]}
EOF
}

sub wap_cgi_object {
    require CGI;
    $q = CGI->new;
    eval {
	require BrowserInfo;
	$bi = BrowserInfo->new($q);
	$can_table = $bi->{'can_table'};
    };
    warn $@ if $@;
    $q;
}

sub wap_std_header {
    print $q->header(-type => "text/vnd.wap.wml");
}

sub wap_init {
    wap_cgi_object();
    wap_std_header();
    $q;
}

sub wap_header {
    <<EOF;
<?xml version="1.0" encoding="ISO-8859-1"?>
<!DOCTYPE wml PUBLIC "-//WAPFORUM//DTD WML 1.1//EN" "http://www.wapforum.org/DTD/wml_1.1.xml">
<wml>
EOF
}

sub wap_footer {
    <<EOF;
</wml>
EOF
}

return 1 if caller() or keys %Devel::Trace::; # XXX Tracer bug

######################################################################

package main;

my $cmd = BBBike::Cmd->new;

if ($0 =~ /wapbbbike/i) {
    my $q = $cmd->wap_init;
    if ($q->param("info")) {
	$cmd->wap_info();
    } elsif (defined $q->param("startname") &&
	     $q->param("startname") ne ""
	     &&
	     defined $q->param("zielname")  &&
	     $q->param("zielname") ne ""
	    ) {
	($BBBike::Cmd::from_street, $BBBike::Cmd::to_street) =
	    ($q->param("startname"), $q->param("zielname"));
	($BBBike::Cmd::from_citypart, $BBBike::Cmd::to_citypart) =
	    ($q->param("startbezirk"), $q->param("zielbezirk"));
	$cmd->search();
	$cmd->wap_output();
    } else {
	$cmd->wap_input();
    }

} elsif ($0 =~ /smsbbbike/i) {
    $cmd->sms_input();
    $cmd->search();
    $cmd->sms_output();

} else { # ($0 =~ /cmdbbbike/i)
    $cmd->cmdline_input();
    $cmd->search();
    $cmd->cmdline_output();
}

__END__
