# -*- perl -*-

#
# $Id: EasySound.pm,v 1.4 1999/06/29 00:09:52 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

package EasySound;
use Tk;
use strict;
use vars qw(@probe_order %probe_func $VERBOSE);

@probe_order = ('perlaudio',
		'audio', 'pcaudio', 'speaker',
		'nas', 'term', 'none');
if ($^O =~ /win32/i) {
    unshift @probe_order, 'win32';
}

$VERBOSE = 0 unless defined $VERBOSE;

my $spkrdev    = "/dev/speaker";
my $audiodev   = "/dev/audio";
my $pcaudiodev = "/dev/pcaudio";

my($sound_player, $sound_file, $sound_data, $sound_server, $sound_type);

sub init {
    my $sound_file = shift; # without extension!
    my $probe;
    warn "Sound probe...\n" if $VERBOSE;
    foreach $probe (@probe_order) {
	warn "Trying $probe...\n" if $VERBOSE;
	my $r = eval '&probe_' . $probe . qq{('$sound_file')};
	if ($r) {
	    $sound_player = eval '\&playsound_' . $probe;
	    $sound_type = $probe;
	    warn "Success with $probe. Player sub: $sound_player.\n"
	      if $VERBOSE;
	    return 1;
	}
    }
    0;
}

sub probe_perlaudio {
    my $try_file = shift;
    eval { require Audio::Data;
	   require Audio::Play;
	   require IO::File;
       };
    if ($@) {
	warn $@ if $VERBOSE > 1;
	return 0;
    }
    # XXX zuerst Audio::Play::Net abfragen?
    eval { $sound_server = new Audio::Play };
    if ($@ || !$sound_server) {
	eval { require Audio::Play::Net;
	       $sound_server = new Audio::Play::Net;
	       # XXX Fehler in Audio-0.003
	       *Audio::Play::Net::play = \&Audio::Play::Net::Play;
	       *Audio::Play::Net::flush = \&Audio::Play::Net::Flush;
	   };
	if (!$@ && !$sound_server) {
	    $@ = "sound_server ist nicht definiert.";
	}
	if ($@) {
	    warn $@ if $VERBOSE > 1;
	    return 0;
	}
    }

    # wav geht nicht mit Audio::Play
    $sound_file = Tk::findINC("$try_file.au");
    return 0 if !defined $sound_file;

    my $fh = new IO::File "<$sound_file";
    return 0 if !$fh;
    $sound_data = new Audio::Data Load => $fh;
    return 0 if !defined $sound_data;
    $fh->close;
    $sound_data->rate($sound_server->rate)
      if $sound_data->rate != $sound_server->rate;
    1;
}

sub probe_nas {
    # NYI --- use auplay
    0;
}

sub probe_dev {
    my $dev = shift;
    if (-e $dev) {
	return 0 if !open(DEV, ">$dev");
	close DEV;
	return 1;
    }
    warn "$dev does not exist" if $VERBOSE > 1;
    0;
}

sub load_data {
    my $file = shift;
    return 0 if !open(SND, $file);
    local($/) = undef;
    $sound_data = <SND>;
    close SND;
    1;
}

sub probe_audio {
    my $try_file = shift;
    if (probe_dev($audiodev)) {
	$sound_file = Tk::findINC("$try_file.au");
	if (defined $sound_file) {
	    return load_data($sound_file);
	}
    }
    0;
}

sub probe_pcaudio {
    my $try_file = shift;
    if (probe_dev($pcaudiodev)) {
	$sound_file = Tk::findINC("$try_file.au");
	if (defined $sound_file) {
	    return load_data($sound_file);
	}
    }
    0;
}

sub probe_speaker {
    probe_dev($spkrdev);
}

sub probe_win32 {
    my $try_file = shift;
    eval { require Win32::Sound };
    if ($@) {
	warn $@ if $VERBOSE > 1;
	return;
    }
    $sound_file = Tk::findINC("$try_file.wav");
    defined $sound_file;
}

sub probe_term { 1 }
sub probe_none { 1 }

sub playsound {
    &$sound_player if defined $sound_player;
}

sub playsound_perlaudio {
    $sound_server->play($sound_data);
    $sound_server->flush;
}

sub playsound_audio {
    if (open(DEV, ">$audiodev")) {
	print DEV $sound_data;
	close DEV;
    }
}

sub playsound_pcaudio {
    if (open(DEV, ">$pcaudiodev")) {
	print DEV $sound_data;
	close DEV;
    }
}

sub playsound_speaker {
    if (open(SPKR, ">$spkrdev")) {
	print SPKR "O2e64";
	close SPKR;
    }
}

sub playsound_win32 {
    Win32::Sound::Play($sound_file, &Win32::Sound::SND_ASYNC);
}

sub playsound_term {
    print STDERR "\007";
}

sub playsound_none { }

sub info {
    if (!defined $sound_type || $sound_type eq 'none') {
	"kein";
    } elsif ($sound_type eq 'perlaudio') {
	"Audio::Data, Audio::Play";
    } elsif ($sound_type eq 'audio') {
	"direkt auf $audiodev"
    } elsif ($sound_type eq 'pcaudio') {
	"direkt auf $pcaudiodev";
    } elsif ($sound_type eq 'speaker') {
	"direkt auf $spkrdev";
    } elsif ($sound_type eq 'nas') {
	"direkt auf NAS";
    } elsif ($sound_type eq 'term') {
	"term bell";
    } elsif ($sound_type eq 'win32') {
	"Win32::Sound";
    } else {
	"unbekannt ($sound_type)";
    }
}

1;
