# -*- perl -*-

#
# $Id: FlatCheckbox.pm,v 1.6 2000/07/25 19:49:41 eserte Exp $
# Author: Slaven Rezic
#
# Copyright (C) 1998 Slaven Rezic. All rights reserved.
# This package is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
# Mail: eserte@cs.tu-berlin.de
# WWW:  http://user.cs.tu-berlin.de/~eserte/
#

# XXX borderwidth ist krampfig ... check it

package Tk::FlatCheckbox;
use Tk::Derived;
use Tk::Canvas;
use strict;
use vars qw($VERSION @ISA);
@ISA = qw(Tk::Derived Tk::Canvas);
Construct Tk::Widget 'FlatCheckbox';

$VERSION = '0.04';

sub ClassInit
{
 my ($class,$mw) = @_;
 $mw->bind($class,"<1>", "Invoke");
 $mw->bind($class,"<space>", "Invoke");
 $mw->bind($class,"<Enter>", "Enter");
 $mw->bind($class,"<Leave>", "Leave");
 return $class;
}

sub Populate {
    my($w, $args) = @_;
    if ($args->{-borderwidth}) {
	$w->{Configure}{borderwidth} = delete $args->{-borderwidth};
    }
    $w->{Configure}{borderwidth} = 2;#XXXXX
    $w->ConfigSpecs
      (
       #-borderwidth => ['METHOD', 'borderWidth', 'BorderWidth', 30],#XXX
       -image        => ['METHOD', 'image', 'Image', undef],
       -variable     => ['METHOD', 'variable', 'Variable', undef],
       -command      => ['CALLBACK', 'command', 'Command', undef],
       -state        => ['PASSIVE', 'state', 'State', 'active'], #XXX
       -text         => ['PASSIVE', 'text', 'Text', undef], # dummy option
       -raiseonenter => ['PASSIVE', 'raiseOnEnter', 'RaiseOnEnter', 0],
      );
}

sub ConfigChanged {
    my($w, $args) = @_;

    # XXX hack
    if ($args->{'-raiseonenter'}) {
	$w->Tk::Canvas::configure(-borderwidth => 2);
    }
}

#XXX besser: ConfigChanged verwenden!

sub image {
    my($w, $img) = @_;
    if (@_ < 2) {
	$w->{Configure}{'image'};
    } else {
	$w->delete('image');
	return if !defined $img;
	$w->{Configure}{'image'} = $img;
	$w->configure(-width => $img->width + $w->{Configure}{borderwidth}*2,
		      -height => $img->height + $w->{Configure}{borderwidth}*2
		     );
	$w->{'_ImageX'} = 1 + $w->{Configure}{borderwidth};
	$w->{'_ImageY'} = 1 + $w->{Configure}{borderwidth};
	$w->createImage($w->{'_ImageX'}, $w->{'_ImageY'},
			-anchor => 'nw', -image => $img, -tags => 'image');
    }
}

sub variable {
    my($w, $varref) = @_;
    if (@_ < 2) {
	$w->{Configure}{'variable'};
    } else {
	eval {
	    require Tie::Watch;
	};
	if (!$@) {
	    my $watch = new Tie::Watch
	      -variable => $varref,
	      -store => sub { _store_scalar($w, @_) };
	} else {
	    warn "Can't find Tie::Watch --- tie of variables will not work";
	}
	$w->{Configure}{'variable'} = $varref;
	$w->_update_state;
    }
}

sub _store_scalar {
    my($w, $self, $newval) = @_;
    $self->Store($newval);
    $w->_update_state;
}

sub _update_state {
    my $w = shift;
    my $varref = $w->{Configure}{'variable'};
    if (!defined $varref) {
	$w->{Configure}{'state'} = 0;
    } else {
	$w->{Configure}{'state'} = !!($$varref);
    }
    $w->DrawCheck;
}

sub DrawCheck {
    my $w = shift;
    if ($w->{Configure}{'state'}) {
	my $img = $w->{Configure}{'image'};
	if (defined $img) {
	    $w->createRectangle($w->{'_ImageX'} + $img->width-5,
				$w->{'_ImageY'} + $img->height-5,
				$w->{'_ImageX'} + $img->width,
				$w->{'_ImageX'} + $img->height,
				-outline => undef, -fill => 'red',
				-tags => 'check');
	}
    } else {
	$w->delete('check');
    }
}

sub invoke {
    my $w = shift;
    $w->{Configure}{'state'} = ($w->{Configure}{'state'} ? 0 : 1);
    if ($w->cget('-variable')) {
	$ {$w->cget('-variable')} = $w->{Configure}{'state'};
    }
    $w->DrawCheck;
    if (defined $w->cget(-command)) {
	$w->Callback(-command);
    }
}

sub bind {
    shift->Tk::bind(@_);
}

# sub borderwidth {
#     my $w = shift;
#     if (@_) {
# 	$w->{Configure}{'borderwidth'} = shift;
#     }
#     $w->{Configure}{'borderwidth'};
# }

sub Invoke {
    my $w = shift;
    $w->invoke() unless($w->cget("-state") eq "disabled");
}

sub Enter {
    my $w = shift;
    if ($w->cget(-raiseonenter)) {
	$w->configure(-relief => "raised");
    } else {
	$w->{"_OrigBG_"} = $w->cget(-bg) unless defined $w->{"_OrigBG_"};
	$w->configure(-bg => Tk::ACTIVE_BG);
    }
}

sub Leave {
    my $w = shift;
    if ($w->cget(-raiseonenter)) {
	$w->configure(-relief => "flat");
    } else {
	if (defined $w->{"_OrigBG_"}) {
	    $w->configure(-bg => $w->{"_OrigBG_"});
	    delete $w->{"_OrigBG_"};
	}
    }
}

1;
